//
// Copyright (c) 2023 Ted Unangst <tedu@tedunangst.com>
//
// Permission to use, copy, modify, and distribute this software for any
// purpose with or without fee is hereby granted, provided that the above
// copyright notice and this permission notice appear in all copies.
//
// THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
// WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
// ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
// WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
// ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
// OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

package main

import (
	"fmt"
	"html"
	"net/http"
	"regexp"
	"strings"
	"time"

	"humungus.tedunangst.com/r/azorius/slices"
	"humungus.tedunangst.com/r/webs/gencache"
	"humungus.tedunangst.com/r/webs/junk"
)

var apContentType = `application/ld+json; profile="https://www.w3.org/ns/activitystreams"`
var acceptNames = []string{
	`application/ld+json`,
	`application/activity+json`,
}
var apAcceptType = apContentType
var apContext = "https://www.w3.org/ns/activitystreams"
var apPublic = "https://www.w3.org/ns/activitystreams#Public"

func isAPtype(ct string) bool {
	for _, at := range acceptNames {
		if strings.HasPrefix(ct, at) {
			return true
		}
	}
	return false
}

func isAPrequest(r *http.Request) bool {
	ct := strings.ToLower(r.Header.Get("Accept"))
	return isAPtype(ct)
}

func isMine(id string) bool {
	return strings.HasPrefix(id, serverPrefix)
}

func getString(obj junk.Junk, name string, limit int) string {
	val, _ := obj.GetString(name)
	if len(val) > limit {
		val = val[:limit]
	}
	return val
}

func getArray(obj junk.Junk, name string) []interface{} {
	if val, ok := obj.GetMap(name); ok {
		return []interface{}{val}
	}
	if str, ok := obj.GetString(name); ok {
		return []interface{}{str}
	}
	arr, _ := obj.GetArray(name)
	return arr
}

func getType(obj junk.Junk) string {
	if i, ok := obj["type"]; ok {
		if s, ok := i.(string); ok {
			return s
		}
		if a, ok := i.([]interface{}); ok && len(a) > 0 {
			if s, ok := a[0].(string); ok {
				return s
			}
		}
	}
	return ""
}

func apHandleGroupUrl(url string) *Group {
	j, err := fetchActivity(url, slowTimeout)
	if err != nil {
		dlog.Printf("unable to fetch group: %s", err)
		return nil
	}
	return apHandleGroup(originate(url), j, false)
}

func apHandleGroup(origin string, obj junk.Junk, isUpdate bool) *Group {
	id, _ := obj.GetString("id")
	typ := getType(obj)
	if typ != "Group" || id == "" {
		dlog.Printf("it's not a group! %s", obj.ToString())
		return nil
	}
	if isMine(id) {
		dlog.Printf("not importing group from myself: %s", id)
		return nil
	}
	if origin != originate(id) {
		dlog.Printf("impersonated group %s from %s", id, origin)
		return nil
	}
	group := dbFindGroup(id)
	if group != nil && !isUpdate {
		dlog.Printf("duplicate group %s", id)
		return group
	}

	name := getString(obj, "preferredUsername", 100)
	if !re_plainname.MatchString(name) {
		dlog.Printf("name is too fancy: %s %s", id, name)
		return nil
	}
	name = fmt.Sprintf("%s@%s", name, origin)
	display := getString(obj, "name", 100)
	desc := getString(obj, "summary", 5000)
	inbox, _ := obj.GetString("inbox")
	sharebox, _ := obj.GetString("endpoints", "sharedInbox")
	outbox, _ := obj.GetString("outbox")
	haschat := getString(obj, "chat", 400) != ""
	if isUpdate {
		if name != group.Name {
			dlog.Printf("impossible to change group name: %s to %s", group.Name, name)
		}
		group.Display = display
		group.Desc = desc
		group.Inbox = inbox
		group.ShareBox = sharebox
		group.Outbox = outbox
		if haschat {
			group.Flags |= flagHasChat
		} else {
			group.Flags &= ^flagHasChat
		}
	} else {
		var flags int64
		if haschat {
			flags |= flagHasChat
		}
		group = &Group{
			Name:     name,
			Display:  display,
			APid:     id,
			Desc:     desc,
			Format:   "html",
			Inbox:    inbox,
			ShareBox: sharebox,
			Outbox:   outbox,
			Flags:    flags,
		}
		mtxSaveGroup.Lock()
		defer mtxSaveGroup.Unlock()
		if dup := dbFindGroup(id); dup != nil {
			dlog.Printf("duplicate group %s: %s", name, id)
			return dup
		}
		if dup := dbFindGroup(name); dup != nil {
			dlog.Printf("duplicate group name %s: %s", name, id)
			return dup
		}
		if hastombstone(id) {
			dlog.Printf("recreate of dead group: %s", id)
			return nil
		}
	}

	group.Refreshed = time.Now()
	err := dbSaveGroup(group)
	if err != nil {
		elog.Printf("error saving group: %s", err)
		return nil
	}
	dlog.Printf("saved group: %s", group.Name)
	if isUpdate {
		return group
	}

	featuredUrl, _ := obj.GetString("featured")
	if featuredUrl != "" {
		items, err := fetchCollection(featuredUrl)
		if err != nil {
			dlog.Printf("error fetching featured: %s", err)
		}
		slices.Reverse(items)
		o := Origin{Domain: origin, Group: group}
		for _, item := range items {
			o.Domain = origin
			j, ok := item.(junk.Junk)
			if !ok {
				url, ok := item.(string)
				if !ok {
					dlog.Printf("unexpected something in featured: %s: %v", outbox, item)
					continue
				}
				j, err = fetchActivity(url, fastTimeout)
				if err != nil {
					continue
				}
				o.Domain = originate(url)
			}
			o.Json = j
			post := apHandlePage(&o, j)
			if post != nil {
				group.Featured = append(group.Featured, post.APid)
			}
		}
		if len(group.Featured) > 0 {
			err = dbSaveGroup(group)
			if err != nil {
				elog.Printf("error saving group: %s", err)
			}
		}
	}

	group = findgroup(id)

	keyobj, ok := obj.GetMap("publicKey")
	if ok {
		savepubkey(origin, keyobj, group.APid)
	}
	notifNewGroup(group)
	return group
}

func apPullOutbox(group *Group, outbox string) {
	items, err := fetchCollection(outbox)
	if err != nil {
		dlog.Printf("error fetching outbox: %s", err)
		return
	}
	if len(items) > 10 {
		items = items[:10]
	}
	slices.Reverse(items)
	origin := Origin{Domain: originate(outbox), Group: group}
	for _, item := range items {
		j, ok := item.(junk.Junk)
		if !ok {
			dlog.Printf("unexpected something in outbox: %s: %v", outbox, item)
			continue
		}
		origin.Json = j
		typ := getType(j)
		if typ == "Announce" {
			apHandleAnnounce(&origin, j, true)
		} else {
			dlog.Printf("skipping %s in outbox %s", typ, outbox)
		}
	}
}

func fetchCollection(url string) ([]interface{}, error) {
	j, err := fetchActivity(url, fastTimeout)
	if err != nil {
		return nil, err
	}
	typ := getType(j)
	if typ != "OrderedCollection" {
		return nil, fmt.Errorf("not a collection")
	}
	first, _ := j.GetString("first")
	if first != "" {
		j, err = fetchActivity(first, fastTimeout)
		if err != nil {
			return nil, err
		}
	}
	items, _ := j.GetArray("orderedItems")
	return items, nil
}

func apHandleAnnounce(origin *Origin, obj junk.Junk, asUpdate bool) {
	typ := getType(obj)
	actor, _ := obj.GetString("actor")
	if typ != "Announce" || actor == "" {
		dlog.Printf("it's not an announce! %s", obj.ToString())
		return
	}
	if !origin.Match(actor) {
		dlog.Printf("forged announce: %s", origin)
		return
	}
	target, _ := obj.GetString("object")
	if target != "" {
		post, comment := postForAPid(origin.Group, target)
		if origin.Group != nil {
			dlog.Printf("we're done here")
			return
		}
		if post != nil {
			person := findperson(actor)
			if person != nil && !post.IsShared(person.ID) {
				err := dbSaveShare(person.ID, post.ID)
				if err != nil {
					elog.Printf("error saving share: %s", err)
					return
				}
				dlog.Printf("saved share: %s", target)
				return
			}
		}
		if comment != nil {
			dlog.Printf("ignoring shared comment")
			return
		}
		dlog.Printf("missing share: %s", origin)
		return
	}
	ao, _ := obj.GetMap("object")
	if len(ao) == 0 {
		dlog.Printf("lost the announce object: %s", obj.ToString())
		return
	}

	obj = ao
	typ = getType(obj)
	switch typ {
	case "Create":
		apHandleCreate(origin, obj, asUpdate)
	case "Update":
		apHandleUpdate(origin, obj)
	case "Page":
		post := apHandlePage(origin, obj)
		person := findperson(actor)
		if post != nil && person != nil && !post.IsShared(person.ID) {
			err := dbSaveShare(person.ID, post.ID)
			if err != nil {
				elog.Printf("error saving share: %s", err)
				return
			}
			dlog.Printf("saved share: %s", post.APid)
		}
	case "Note":
		apHandleNote(origin, obj)
	case "ChatMessage":
		apHandleChatMessage(origin, obj)
	case "Delete":
		apHandleDelete(origin, obj)
	case "Add":
		apHandleAdd(origin.Group, origin.Domain, obj)
	case "Remove":
		apHandleRemove(origin.Group, origin.Domain, obj)
	case "Lock":
		target, _ = obj.GetString("object")
		dlog.Printf("ignore Lock: %s", target)
	case "Block":
		target, _ = obj.GetString("object")
		dlog.Printf("ignore Block: %s", target)
	default:
		dlog.Printf("can't import announce of %s: %s", typ, origin)
	}
}

func apHandleDelete(origin *Origin, j junk.Junk) {
	target, ok := j.GetString("object")
	if !ok {
		obj, ok := j.GetMap("object")
		if ok {
			target, _ = obj.GetString("id")
		}
	}
	if target == "" {
		dlog.Printf("delete with no object: %s", origin)
		return
	}
	actor, _ := j.GetString("actor")
	if isMine(target) {
		dlog.Printf("not deleting my own %s: %s", target, origin)
		return
	}
	if hastombstone(target) {
		return
	}
	post := dbFindPost(target)
	if post != nil {
		poster := findperson(post.PersonName)
		group := findgroup(post.GroupName)
		ok := false
		if origin.Group != nil {
			ok = origin.Group.Name == post.GroupName
		} else {
			ok = actor == poster.APid
		}
		if !ok {
			dlog.Printf("unauthorized delete: %s", origin)
			return
		}
		dlog.Printf("delete of %s by %s via %s", target, actor, origin.Domain)
		mtxSavePost.Lock()
		defer mtxSavePost.Unlock()
		err := dbDeletePost(post)
		if err != nil {
			elog.Printf("error deleting post: %s", err)
		}
		if group.IsLocal() {
			apSendDeletePost(actor, post)
		}
		return
	}
	comment := dbFindComment(target)
	if comment != nil {
		poster := findperson(comment.PersonName)
		post := findpost(comment.PostXID)
		group := findgroup(post.GroupName)
		ok := false
		if origin.Group != nil {
			ok = origin.Group.Name == post.GroupName
		} else {
			ok = actor == poster.APid
		}
		if !ok {
			dlog.Printf("unauthorized delete: %s", origin)
			return
		}
		dlog.Printf("delete of %s by %s via %s", target, actor, origin.Domain)
		mtxSaveComment.Lock()
		defer mtxSaveComment.Unlock()
		err := dbDeleteComment(comment)
		if err != nil {
			elog.Printf("error deleting comment: %s", err)
		}
		if group.IsLocal() {
			apSendCommentDelete(actor, group, comment)
		}
		return
	}
	person := dbFindPerson(target)
	if person != nil {
		if actor != person.APid {
			dlog.Printf("unauthorized delete: %s", origin)
			return
		}
		dlog.Printf("self delete of %s", person.APid)
		mtxSavePerson.Lock()
		defer mtxSavePerson.Unlock()
		err := dbDeletePerson(person)
		if err != nil {
			elog.Printf("error deleting person: %s", err)
		}
		return
	}
	dlog.Printf("can't find target for delete: %s", target)
}

func apHandleCreate(origin *Origin, obj junk.Junk, asUpdate bool) {
	actor, _ := obj.GetString("actor")
	typ := getType(obj)
	if typ != "Create" || actor == "" {
		dlog.Printf("it's not a create! %s", obj.ToString())
		return
	}
	co, _ := obj.GetMap("object")
	if len(co) == 0 {
		dlog.Printf("no created object: %s", obj.ToString())
		return
	}
	obj = co
	typ = getType(obj)
	switch typ {
	case "Page":
		if asUpdate {
			apUpdatePage(origin, obj)
		} else {
			apHandlePage(origin, obj)
		}
	case "Note":
		apHandleNote(origin, obj)
	case "ChatMessage":
		apHandleChatMessage(origin, obj)
	default:
		dlog.Printf("can't import create of %s: %s", typ, obj.ToString())
	}
}

func apHandleAdd(group *Group, origin string, obj junk.Junk) {
	if group == nil {
		dlog.Printf("no group for Add: %s", obj.ToString())
		return
	}
	if origin != originate(group.APid) {
		dlog.Printf("apparently forged Add: %s", obj.ToString())
		return
	}
	target, _ := obj.GetString("target")
	if target != group.APid+"/featured" {
		dlog.Printf("confused about target for Add: %s", obj.ToString())
		return
	}
	what, _ := obj.GetString("object")
	post := findpost(what)
	if post == nil {
		dlog.Printf("don't have featured post: %s", obj.ToString())
	}

	group = dbFindGroup(group.Name)
	featured := []string{what}
	for _, f := range group.Featured {
		if f != what {
			featured = append(featured, f)
		}
	}
	group.Featured = featured
	err := dbSaveGroup(group)
	if err != nil {
		elog.Printf("error saving group: %s", err)
		return
	}
	dlog.Printf("group %s features %s", group.Name, what)
}

func apHandleRemove(group *Group, origin string, obj junk.Junk) {
	if group == nil {
		dlog.Printf("no group for Remove: %s", obj.ToString())
		return
	}
	if origin != originate(group.APid) {
		dlog.Printf("apparently forged Remove: %s", obj.ToString())
		return
	}
	target, _ := obj.GetString("target")
	if target != group.APid+"/featured" {
		dlog.Printf("confused about target for Remove: %s", obj.ToString())
		return
	}
	what, _ := obj.GetString("object")

	group = dbFindGroup(group.Name)
	featured := []string{}
	for _, f := range group.Featured {
		if f != what {
			featured = append(featured, f)
		}
	}
	group.Featured = featured
	err := dbSaveGroup(group)
	if err != nil {
		elog.Printf("error saving group: %s", err)
		return
	}
	dlog.Printf("group %s drops %s", group.Name, what)
}

func apHandleUpdate(origin *Origin, obj junk.Junk) {
	actor, _ := obj.GetString("actor")
	typ := getType(obj)
	if typ != "Update" || actor == "" {
		dlog.Printf("it's not an update! %s", obj.ToString())
		return
	}
	uo, _ := obj.GetMap("object")
	if len(uo) == 0 {
		dlog.Printf("no updated object: %s", obj.ToString())
		return
	}
	obj = uo
	typ = getType(obj)
	switch typ {
	case "Page":
		apUpdatePage(origin, obj)
	case "Note":
		apUpdateNote(origin, obj)
	case "Group":
		apHandleGroup(origin.Domain, obj, true)
	default:
		dlog.Printf("can't import update of %s: %s", typ, obj.ToString())
	}
}

func apHandleNote(origin *Origin, obj junk.Junk) *Comment {
	id, _ := obj.GetString("id")
	typ := getType(obj)
	if typ != "Note" || id == "" {
		dlog.Printf("it's not a note! %s", obj.ToString())
		return nil
	}
	if isMine(id) {
		dlog.Printf("not importing note from myself: %s", id)
		return nil
	}
	if !origin.Match(id) {
		dlog.Printf("untrusted note %s from %s", id, origin.Domain)
		if origin.Group == nil {
			dlog.Printf("skipping forged note create: %s", id)
			return nil
		}
	}
	if dup := findcomment(id); dup != nil {
		dlog.Printf("duplicate comment %s", id)
		return dup
	}

	attrTo, _ := obj.GetString("attributedTo")
	if !sameOrigin(id, attrTo) {
		dlog.Printf("attribution mismatch: %s <> %s", id, attrTo)
		return nil
	}
	person := personForAPid(attrTo)
	if person == nil {
		dlog.Printf("unable to get person for comment: %s", id)
		return nil
	}
	addrs := collectAddrs(nil, obj, "to")
	addrs = collectAddrs(addrs, obj, "cc")
	for _, addr := range addrs {
		if isMine(addr) {
			dlog.Printf("addressed to me: %s", obj.ToString())
		}
	}
	if !slices.Contains(addrs, apPublic) {
		dlog.Printf("dropping comment that's not public: %s", obj.ToString())
		return nil
	}
	subject := getString(obj, "summary", 200)
	content := getString(obj, "content", 5000)
	dt, ok := obj.GetString("published")
	if !ok {
		dt = time.Now().Format(time.RFC3339)
	}
	replyto, _ := obj.GetString("inReplyTo")
	context, _ := obj.GetString("context")
	if context == "" {
		context, _ = obj.GetString("conversation")
	}
	var post *Post
	var parent *Comment
	var parentid int64
	post = findpost(replyto)
	if post == nil {
		if parent = findcomment(replyto); parent != nil {
			post = findpost(parent.PostXID)
			parentid = parent.ID
		} else {
			post = findpost(context)
		}
	}
	if post == nil {
		dlog.Printf("no post for comment: %s", id)
		return nil
	}
	if origin.Group != nil && origin.Group.Name != post.GroupName {
		dlog.Printf("comment in wrong group: %s %s <> %s", id, origin.Group.Name, post.GroupName)
		return nil
	}
	jtags := getArray(obj, "tag")
	var tags []string
	for _, jtag := range jtags {
		tag, ok := jtag.(junk.Junk)
		if ok {
			name, _ := tag.GetString("name")
			typ := getType(tag)
			if typ == "Hashtag" {
				tags = append(tags, name)
			}
		}
	}
	comment := &Comment{
		PostID:     post.ID,
		PersonID:   person.ID,
		PersonName: person.Name,
		ParentID:   parentid,
		APid:       id,
		Subject:    subject,
		Content:    content,
		Format:     "html",
		Tags:       tags,
	}
	comment.Date, _ = time.Parse(time.RFC3339, dt)

	mtxSaveComment.Lock()
	defer mtxSaveComment.Unlock()
	if dup := findcomment(id); dup != nil {
		dlog.Printf("duplicate comment %s", id)
		return dup
	}
	if hastombstone(id) {
		dlog.Printf("recreate of dead comment: %s", id)
		return nil
	}
	filterComment(comment)
	err := dbSaveComment(comment)
	if err != nil {
		return nil
	}
	dlog.Printf("saved comment: %s", comment.APid)
	comment = findcomment(id)

	notifNewComment(post, parent, comment)
	group := findgroup(post.GroupName)
	if group.IsLocal() {
		apSendComment(person, group, comment)
	}
	return comment
}

func apUpdateNote(origin *Origin, obj junk.Junk) *Comment {
	id, _ := obj.GetString("id")
	typ := getType(obj)
	if typ != "Note" || id == "" {
		dlog.Printf("it's not a note! %s", obj.ToString())
		return nil
	}
	if isMine(id) {
		dlog.Printf("not updating note from myself: %s", id)
		return nil
	}
	comment := findcomment(id)
	if comment == nil {
		dlog.Printf("lost original for updated note %s", id)
		return apHandleNote(origin, obj)
	}
	post := findpost(comment.PostXID)
	group := findgroup(post.GroupName)
	if !origin.Match(id) {
		dlog.Printf("untrusted note update %s from %s", id, origin.Domain)
		if origin.Group == nil || origin.Group.Name != post.GroupName {
			dlog.Printf("skipping forged note update: %s", id)
			return nil
		}
	}

	subject := getString(obj, "summary", 200)
	content := getString(obj, "content", 5000)

	comment = dbFindComment(id)
	changed := false
	if comment.Subject != subject {
		changed = true
		comment.Subject = subject
	}
	if comment.Content != content {
		changed = true
		comment.Content = content
	}
	if !changed {
		dlog.Printf("no change to comment: %s", comment.APid)
		return comment
	}
	filterComment(comment)
	err := dbSaveComment(comment)
	if err != nil {
		return nil
	}
	dlog.Printf("updated comment: %s", comment.APid)
	comment = findcomment(id)

	if group.IsLocal() {
		person := findperson(comment.PersonName)
		apSendCommentUpdate(person, group, comment)
	}
	return comment
}

func apHandleAttachment(post *Post, permstore bool, seen map[string]bool, att junk.Junk) *Attachment {
	typ := getType(att)
	mt, _ := att.GetString("mediaType")
	if mt == "" {
		mt = "image"
	}
	if (typ == "Document" || typ == "Image") && strings.HasPrefix(mt, "image") {
		name, _ := att.GetString("name")
		desc, _ := att.GetString("summary")
		desc = html.UnescapeString(desc)
		if desc == "" {
			desc = name
		}
		url, ok := att.GetString("url")
		if !ok {
			url, ok = att.GetString("href")
		}
		if ok && seen[url] {
			ok = false
		}
		if !ok {
			return nil
		}
		seen[url] = true
		att, err := attachmentfromurl(url, permstore)
		if err != nil {
			dlog.Printf("error getting attachment: %s", err)
			return nil
		}
		if seen[att.Xid] {
			return nil
		}
		att.Desc = desc
		att.PostID = post.ID
		dlog.Printf("saving attachment %s for %s", url, post.Xid)
		err = dbSaveAttachment(att)
		if err != nil {
			elog.Printf("error saving att: %s", err)
			return nil
		}
		return att
	}
	return nil
}

func postForAPid(groupactor *Group, url string) (*Post, *Comment) {
	if isBannedPerson(url) {
		dlog.Printf("skipping banned post: %s", url)
		return nil, nil
	}
	if post := findpost(url); post != nil {
		return post, nil
	}
	if comment := findcomment(url); comment != nil {
		return nil, comment
	}
	j, err := fetchActivity(url, slowTimeout)
	if err != nil {
		dlog.Printf("error fetching post %s: %s", url, err)
		return nil, nil
	}
	origin := Origin{Domain: originate(url), Group: groupactor, Json: j}
	typ := getType(j)
	switch typ {
	case "Note":
		return nil, apHandleNote(&origin, j)
	case "Page":
		return apHandlePage(&origin, j), nil
	}
	dlog.Printf("unexpected type %s looking for post: %s", typ, url)
	return nil, nil
}

func apHandlePage(origin *Origin, obj junk.Junk) *Post {
	id, _ := obj.GetString("id")
	typ := getType(obj)
	if typ != "Page" || id == "" {
		dlog.Printf("it's not a page! %s", obj.ToString())
		return nil
	}
	if isMine(id) {
		dlog.Printf("not importing page from myself: %s", id)
		return nil
	}
	if !origin.Match(id) {
		dlog.Printf("untrusted page %s from %s", id, origin.Domain)
		if origin.Group == nil {
			dlog.Printf("skipping forged page create: %s", id)
			return nil
		}
	}
	if dup := findpost(id); dup != nil {
		dlog.Printf("duplicate post %s", id)
		return dup
	}
	addrs := collectAddrs(nil, obj, "to")
	addrs = collectAddrs(addrs, obj, "cc")
	for _, addr := range addrs {
		if isMine(addr) {
			dlog.Printf("addressed to me: %s", obj.ToString())
		}
	}
	if !slices.Contains(addrs, apPublic) {
		dlog.Printf("dropping post that's not public: %s", obj.ToString())
		return nil
	}

	group := origin.Group
	if group == nil {
		audience, _ := obj.GetString("audience")
		if audience == "" {
			dlog.Printf("missed the audience, trying harder")
			target, _ := obj.GetString("target")
			if target != "" {
				audience = strings.TrimSuffix(target, "/outbox")
			}
		}
		if audience == "" {
			for _, addr := range addrs {
				if findgroup(addr) != nil {
					audience = addr
					break
				}
			}
			if audience == "" {
				dlog.Printf("no audience in here: %s", obj.ToString())
				return nil
			}
		}
		if isBannedPerson(audience) {
			dlog.Printf("skipping banned group: %s", audience)
			return nil
		}
		group = findgroup(audience)
		if group == nil {
			dlog.Printf("importing new group for post: %s", audience)
			group = apHandleGroupUrl(audience)
		}
	}
	if group == nil {
		dlog.Printf("no group for the post: %s", origin)
		return nil
	}
	if origin.Group != nil && origin.Group.Name != group.Name {
		dlog.Printf("post in wrong group: %s %s <> %s", id, origin.Group.Name, group.Name)
		return nil
	}

	attrTo, _ := obj.GetString("attributedTo")
	if !sameOrigin(id, attrTo) {
		dlog.Printf("attribution mismatch: %s <> %s", id, attrTo)
		return nil
	}
	person := personForAPid(attrTo)
	if person == nil {
		dlog.Printf("unable to get person for post: %s", id)
		return nil
	}

	title := getString(obj, "name", 1000)
	content := getString(obj, "content", 50000)
	var link string
	atts := getArray(obj, "attachment")
	for _, jatt := range atts {
		att, ok := jatt.(junk.Junk)
		if ok {
			typ := getType(att)
			if typ == "Link" {
				link = getString(att, "href", 2000)
			}
		}
	}
	if link == "" {
		link = getString(obj, "url", 2000)
	}
	link = normalizeLink(link, originate(id))
	title = normalizeTitle(title)

	var tags []string
	jtags := getArray(obj, "tag")
	for _, jtag := range jtags {
		tag, ok := jtag.(junk.Junk)
		if ok {
			name, _ := tag.GetString("name")
			typ := getType(tag)
			if typ == "Hashtag" {
				tags = append(tags, name)
			}
		}
	}
	tags = append(tags, simpleName(group.Name))
	dt, ok := obj.GetString("published")
	if !ok {
		dt = time.Now().Format(time.RFC3339)
	}
	context, _ := obj.GetString("context")
	post := &Post{
		GroupID:    group.ID,
		GroupName:  group.Name,
		PersonID:   person.ID,
		PersonName: person.Name,
		Title:      title,
		APid:       id,
		Link:       link,
		Content:    content,
		Format:     "html",
		Context:    context,
		Tags:       tagSet(tags),
	}
	post.Date, _ = time.Parse(time.RFC3339, dt)
	if group.IsLocal() && group.ModPosts() {
		post.Pending = true
	}

	mtxSavePost.Lock()
	defer mtxSavePost.Unlock()
	if dup := findpost(id); dup != nil {
		dlog.Printf("duplicate post %s", id)
		return dup
	}
	if hastombstone(id) {
		dlog.Printf("recreate of dead post: %s", id)
		return nil
	}
	filterPost(post)
	err := dbSavePost(post)
	if err != nil {
		return nil
	}
	dlog.Printf("saved post: %s", post.APid)

	seen := make(map[string]bool)
	for _, jatt := range atts {
		att, ok := jatt.(junk.Junk)
		if ok {
			apHandleAttachment(post, group.IsLocal(), seen, att)
		}
	}
	img, ok := obj.GetMap("image")
	if ok {
		apHandleAttachment(post, group.IsLocal(), seen, img)
	}
	replies, _ := obj.GetString("replies")
	if replies != "" && !importInProgress {
		go apPullReplies(replies)
	}

	post = findpost(id)
	notifNewPost(group, post)
	if group.IsLocal() {
		apSendPost(person, post)
	}
	return post
}

func apPullReplies(replies string) {
	items, err := fetchCollection(replies)
	if err != nil {
		dlog.Printf("error fetching replies: %s", err)
	}
	slices.Reverse(items)
	for _, item := range items {
		s, ok := item.(string)
		if !ok {
			continue
		}
		j, err := fetchActivity(s, fastTimeout)
		if err != nil {
			dlog.Printf("error fetching reply: %s", err)
			continue
		}
		origin := Origin{Domain: originate(s), Json: j}
		apHandleNote(&origin, j)
	}
}

func apUpdatePage(origin *Origin, obj junk.Junk) *Post {
	id, _ := obj.GetString("id")
	typ := getType(obj)
	if typ != "Page" || id == "" {
		dlog.Printf("it's not a page! %s", obj.ToString())
		return nil
	}
	if isMine(id) {
		dlog.Printf("not updating page from myself: %s", id)
		return nil
	}
	post := dbFindPost(id)
	if post == nil {
		dlog.Printf("lost original for updated post %s", id)
		return apHandlePage(origin, obj)
	}
	if !origin.Match(id) {
		dlog.Printf("untrusted page update %s from %s", id, origin.Domain)
		if origin.Group == nil || origin.Group.Name != post.GroupName {
			dlog.Printf("skipping forged page update: %s", id)
			return nil
		}
	}

	title := getString(obj, "name", 1000)
	content := getString(obj, "content", 50000)
	var link string
	atts := getArray(obj, "attachment")
	for _, jatt := range atts {
		att := jatt.(junk.Junk)
		typ := getType(att)
		if typ == "Link" {
			link = getString(att, "href", 2000)
		}
	}
	if link == "" {
		link = getString(obj, "url", 2000)
	}
	link = normalizeLink(link, originate(id))
	title = normalizeTitle(title)

	group := findgroup(post.GroupName)
	post.Title = title
	post.Link = link
	post.Content = content

	seen := make(map[string]bool)
	dbDeletePostAttachments(post.ID)
	for _, jatt := range atts {
		att, ok := jatt.(junk.Junk)
		if ok {
			apHandleAttachment(post, group.IsLocal(), seen, att)
		}
	}
	img, ok := obj.GetMap("image")
	if ok {
		apHandleAttachment(post, group.IsLocal(), seen, img)
	}

	filterPost(post)
	err := dbSavePost(post)
	if err != nil {
		return nil
	}
	dlog.Printf("updated post: %s", post.APid)

	post = findpost(id)

	if group.IsLocal() {
		person := findperson(post.PersonName)
		apSendPostUpdate(person, post)
	}
	return post
}

func personForAPid(url string) *Person {
	if isBannedPerson(url) {
		dlog.Printf("skipping banned person: %s", url)
		return nil
	}
	person := findperson(url)
	if person != nil {
		return person
	}
	j, err := fetchActivity(url, slowTimeout)
	if err != nil {
		dlog.Printf("error fetching person %s: %s", url, err)
		return nil
	}
	return apHandlePerson(originate(url), j)
}

func apHandlePerson(origin string, j junk.Junk) *Person {
	typ := getType(j)
	if typ != "Person" && typ != "Service" {
		dlog.Printf("it's not a person! %s", j.ToString())
		return nil
	}
	id, _ := j.GetString("id")
	if originate(id) != origin {
		dlog.Printf("impersonated person %s from %s", id, origin)
		return nil
	}
	if dup := findperson(id); dup != nil {
		dlog.Printf("duplicate person %s", id)
		return dup
	}
	name := getString(j, "preferredUsername", 100)
	if !re_plainname.MatchString(name) {
		dlog.Printf("name is too fancy: %s %s", id, name)
		return nil
	}
	display := getString(j, "name", 100)
	summary := getString(j, "summary", 5000)
	inbox, _ := j.GetString("inbox")
	sharebox, _ := j.GetString("endpoints", "sharedInbox")
	person := &Person{
		Name:     fmt.Sprintf("%s@%s", name, origin),
		Display:  display,
		APid:     id,
		About:    summary,
		Format:   "html",
		Inbox:    inbox,
		ShareBox: sharebox,
	}

	mtxSavePerson.Lock()
	defer mtxSavePerson.Unlock()
	if dup := findperson(id); dup != nil {
		dlog.Printf("duplicate person %s", id)
		return dup
	}
	if hastombstone(id) {
		dlog.Printf("recreate of dead person: %s", id)
		return nil
	}
	err := dbSavePerson(person)
	if err != nil {
		elog.Printf("error saving person: %s", err)
		return nil
	}
	dlog.Printf("saved person: %s", person.Name)
	person = findperson(id)

	keyobj, ok := j.GetMap("publicKey")
	if ok {
		savepubkey(origin, keyobj, person.APid)
	}
	return person
}

func apHandleChatMessage(origin *Origin, obj junk.Junk) {
	id, _ := obj.GetString("id")
	if isMine(id) {
		dlog.Printf("not importing chat from myself: %s", id)
		return
	}
	if !origin.Match(id) {
		dlog.Printf("untrusted chat %s from %s", id, origin.Domain)
		if origin.Group == nil {
			dlog.Printf("skipping forged chat: %s", id)
			return
		}
	}
	group := origin.Group
	if group == nil {
		audience, _ := obj.GetString("audience")
		if audience == "" {
			addrs := collectAddrs(nil, obj, "to")
			addrs = collectAddrs(addrs, obj, "cc")
			dlog.Printf("missed the chat audience, trying harder")
			for _, addr := range addrs {
				if findgroup(addr) != nil {
					audience = addr
					break
				}
			}
			if audience == "" {
				dlog.Printf("no audience in here: %s", obj.ToString())
				return
			}
		}
		group = findgroup(audience)
	}
	if group == nil {
		dlog.Printf("no group for chat: %s", obj.ToString())
		return
	}
	if group.IsLocal() && !group.HasChat() {
		dlog.Printf("chat not enabled: %s", obj.ToString())
		return
	}
	attrTo, _ := obj.GetString("attributedTo")
	if !sameOrigin(id, attrTo) {
		dlog.Printf("attribution mismatch: %s <> %s", id, attrTo)
		return
	}
	person := personForAPid(attrTo)
	if person == nil {
		dlog.Printf("unable to get person for chat: %s", obj.ToString())
		return
	}
	content := getString(obj, "content", 1000)
	dt, ok := obj.GetString("published")
	if !ok {
		dt = time.Now().Format(time.RFC3339)
	}
	chat := &Chat{
		GroupID:    group.ID,
		GroupName:  group.Name,
		PersonID:   person.ID,
		PersonName: person.Name,
		APid:       id,
		Content:    content,
	}
	chat.Date, _ = time.Parse(time.RFC3339, dt)
	err := dbSaveChat(chat)
	if err != nil {
		elog.Printf("error saving chat: %s", err)
		return
	}
	reformatchat(chat)

	notifNewChat(0, group, chat)
	if group.IsLocal() {
		apSendChat(person, group, chat)
	}
}

var re_urlhost = regexp.MustCompile("https://([^/ #)]+)")

func originate(u string) string {
	m := re_urlhost.FindStringSubmatch(u)
	if len(m) > 1 {
		return strings.ToLower(m[1])
	}
	return ""
}

func apSearch(w http.ResponseWriter, r *http.Request, q string) {
	if isBannedPerson(q) {
		dlog.Printf("skipping banned query: %s", q)
		errorpage(w, r, "alert alert restricted access forbidden")
		return
	}
	if hastombstone(q) {
		dlog.Printf("skipping dead query: %s", q)
		errorpage(w, r, "it's dead; it's gone; it may never exist again")
		return
	}
	{
		if group := findgroup(q); group != nil {
			http.Redirect(w, r, group.Rel(), http.StatusSeeOther)
			return
		}
		if post := findpost(q); post != nil {
			http.Redirect(w, r, post.Rel(), http.StatusSeeOther)
			return
		}
		if person := findperson(q); person != nil {
			http.Redirect(w, r, person.Rel(), http.StatusSeeOther)
			return
		}
		if comment := findcomment(q); comment != nil {
			http.Redirect(w, r, comment.Rel(), http.StatusSeeOther)
			return
		}
	}
	j, err := fetchActivity(q, slowTimeout)
	if err != nil {
		dlog.Printf("search fetch activity: %s", err)
		errorpage(w, r, "unable to find that activity")
		return
	}
	origin := Origin{Domain: originate(q), Json: j}
	typ := getType(j)
	switch typ {
	case "Group":
		group := apHandleGroup(origin.Domain, j, false)
		if group == nil {
			errorpage(w, r, "unable to import group")
		} else {
			if group.Outbox != "" {
				apPullOutbox(group, group.Outbox)
			}
			http.Redirect(w, r, group.Rel(), http.StatusSeeOther)
		}
	case "Page":
		post := apHandlePage(&origin, j)
		if post == nil {
			errorpage(w, r, "unable to import post")
		} else {
			http.Redirect(w, r, post.Rel(), http.StatusSeeOther)
		}
	case "Note":
		comment := apHandleNote(&origin, j)
		if comment == nil {
			errorpage(w, r, "unable to import comment")
		} else {
			http.Redirect(w, r, comment.Rel(), http.StatusSeeOther)
		}
	case "Person":
		person := apHandlePerson(origin.Domain, j)
		if person == nil {
			errorpage(w, r, "unable to import person")
		} else {
			http.Redirect(w, r, person.Rel(), http.StatusSeeOther)
		}
	default:
		errorpage(w, r, "unable to handle that object")
	}
}

var groupByteCache = gencache.New(gencache.Options[string, []byte]{
	Fill: func(name string) ([]byte, bool) {
		group := findgroup(name)
		j := jsonGroup(group)
		return j.ToBytes(), true
	},
	Invalidator: &groupInvalidator, Limit: 300})

func apWriteGroup(group *Group, w http.ResponseWriter) {
	data, _ := groupByteCache.Get(group.Name)
	w.Header().Set("Content-Type", apContentType)
	w.Write(data)
}

func newActivity(typ string, id string) junk.Junk {
	j := junk.New()
	j["@context"] = apContext
	j["type"] = typ
	j["id"] = id
	return j
}

func jsonGroup(group *Group) junk.Junk {
	apid := group.APid
	j := newActivity("Group", apid)
	j["inbox"] = apid + "/inbox"
	j["outbox"] = apid + "/outbox"
	ep := junk.New()
	ep["sharedInbox"] = serverPrefix + "inbox"
	j["endpoints"] = ep
	j["followers"] = apid + "/followers"
	j["following"] = apid + "/following"
	j["name"] = group.Display
	j["preferredUsername"] = group.Name
	j["summary"] = group.HTDesc
	if len(group.Featured) > 0 {
		j["featured"] = apid + "/featured"
	}
	if group.HasChat() {
		j["chat"] = apid + "/chat"
	}
	pk := junk.New()
	pk["id"] = apid + "#key"
	pk["owner"] = apid
	pk["publicKeyPem"] = dbFindPubkey(apid + "#key")
	j["publicKey"] = pk
	return j
}

var personByteCache = gencache.New(gencache.Options[string, []byte]{
	Fill: func(name string) ([]byte, bool) {
		person := findperson(name)
		j := jsonPerson(person)
		return j.ToBytes(), true
	},
	Invalidator: &personInvalidator, Limit: 300})

func apWritePerson(person *Person, w http.ResponseWriter) {
	data, _ := personByteCache.Get(person.Name)
	w.Header().Set("Content-Type", apContentType)
	w.Write(data)
}

func jsonPerson(person *Person) junk.Junk {
	apid := person.APid
	j := newActivity("Person", apid)
	j["inbox"] = apid + "/inbox"
	j["outbox"] = apid + "/outbox"
	ep := junk.New()
	ep["sharedInbox"] = serverPrefix + "inbox"
	j["endpoints"] = ep
	j["followers"] = apid + "/followers"
	j["following"] = apid + "/following"
	j["name"] = person.Display
	j["preferredUsername"] = person.Name
	j["summary"] = person.HTAbout
	pk := junk.New()
	pk["id"] = apid + "#key"
	pk["owner"] = apid
	pk["publicKeyPem"] = dbFindPubkey(apid + "#key")
	j["publicKey"] = pk
	return j
}

var postByteCache = gencache.New(gencache.Options[string, []byte]{
	Fill: func(xid string) ([]byte, bool) {
		post := findpost(xid)
		j := jsonPost(post)
		return j.ToBytes(), true
	},
	Invalidator: &postInvalidator, Limit: 300})

func apWritePost(post *Post, w http.ResponseWriter) {
	data, _ := postByteCache.Get(post.Xid)
	w.Header().Set("Content-Type", apContentType)
	w.Write(data)
}

func jsonActivity(action, who, where string, what interface{}) junk.Junk {
	j := newActivity(action, who+"/activity/"+generateXid())
	j["actor"] = who
	j["to"] = apPublic
	if where != "" {
		j["cc"] = []string{where}
		j["audience"] = where
	}
	j["object"] = what
	return j
}

func jsonAnnounce(who string, where string, what interface{}) junk.Junk {
	return jsonActivity("Announce", who, where, what)
}

func jsonCreate(who string, where string, what junk.Junk) junk.Junk {
	return jsonActivity("Create", who, where, what)
}

func jsonDelete(who string, where string, what string) junk.Junk {
	return jsonActivity("Delete", who, where, what)
}

func jsonUpdate(who string, where string, what junk.Junk) junk.Junk {
	j := jsonActivity("Update", who, where, what)
	what["updated"] = time.Now().Format(time.RFC3339)
	return j
}

func jsonAddFeatured(who string, where string, what string) junk.Junk {
	j := jsonActivity("Add", who, where, what)
	j["target"] = where + "/featured"
	return j
}

func jsonRemoveFeatured(who string, where string, what string) junk.Junk {
	j := jsonActivity("Remove", who, where, what)
	j["target"] = where + "/featured"
	return j
}

func (post *Post) APContent() string {
	if post.Format == "html" {
		return post.Content
	}
	return string(post.HTContent)
}

func jsonPost(post *Post) junk.Junk {
	person := findperson(post.PersonName)
	group := findgroup(post.GroupName)
	j := newActivity("Page", post.APid)
	j["attributedTo"] = person.APid
	j["to"] = apPublic
	j["cc"] = []string{group.APid}
	j["audience"] = group.APid
	j["published"] = post.Date.Format(time.RFC3339)
	j["context"] = post.Context
	j["conversation"] = j["context"]

	j["name"] = post.Title
	j["content"] = post.APContent()
	if post.IsLocal() {
		j["replies"] = post.APid + "/replies"
	}
	var tags []junk.Junk
	for _, t := range post.Tags {
		tag := junk.New()
		tag["type"] = "Hashtag"
		tag["name"] = t
		tags = append(tags, tag)
	}
	if len(tags) > 0 {
		j["tag"] = tags
	}
	var atts []junk.Junk
	if post.Link != "" {
		att := junk.New()
		att["type"] = "Link"
		att["href"] = post.Link
		atts = append(atts, att)
	}
	for _, a := range post.Attachments {
		att := junk.New()
		att["type"] = "Document"
		att["mediaType"] = a.Media
		att["name"] = a.Desc
		att["url"] = a.Url()
		atts = append(atts, att)
	}
	if len(atts) > 0 {
		j["attachment"] = atts
	}
	return j
}

func (comment *Comment) APContent() string {
	if comment.Format == "html" {
		return comment.Content
	}
	return string(comment.HTContent)
}

func jsonComment(comment *Comment) junk.Junk {
	person := findperson(comment.PersonName)
	j := newActivity("Note", comment.APid)
	j["attributedTo"] = person.APid
	j["to"] = apPublic
	var cc []string
	if post := findpost(comment.PostXID); post != nil {
		if group := findgroup(post.GroupName); group != nil {
			cc = append(cc, group.APid)
			j["audience"] = group.APid
		}
		author := findperson(post.PersonName)
		cc = append(cc, author.APid)
		j["inReplyTo"] = post.APid
		j["context"] = post.Context
		j["conversation"] = j["context"]
	}
	if comment.ParentID > 0 {
		if parent := getcomment(comment.ParentID); parent != nil {
			author := findperson(parent.PersonName)
			cc = append(cc, author.APid)
			j["inReplyTo"] = parent.APid
		}
	}
	j["cc"] = cc

	j["published"] = comment.Date.Format(time.RFC3339)
	j["content"] = comment.APContent()
	var tags []junk.Junk
	for _, t := range comment.Tags {
		tag := junk.New()
		tag["type"] = "Hashtag"
		tag["name"] = t
		tags = append(tags, tag)
	}
	if len(tags) > 0 {
		j["tag"] = tags
	}
	return j
}

func jsonChat(chat *Chat) junk.Junk {
	person := findperson(chat.PersonName)
	group := findgroup(chat.GroupName)
	j := newActivity("ChatMessage", chat.APid)
	j["attributedTo"] = person.APid
	j["to"] = group.APid
	j["audience"] = group.APid
	j["published"] = chat.Date.Format(time.RFC3339)

	j["content"] = chat.HTContent
	return j
}

var commentByteCache = gencache.New(gencache.Options[string, []byte]{
	Fill: func(xid string) ([]byte, bool) {
		comment := findcomment(xid)
		j := jsonComment(comment)
		return j.ToBytes(), true
	},
	Invalidator: &commentInvalidator, Limit: 300})

func apWriteComment(comment *Comment, w http.ResponseWriter) {
	data, _ := commentByteCache.Get(comment.Xid)
	w.Header().Set("Content-Type", apContentType)
	w.Write(data)
}

func apWriteChat(chat *Chat, w http.ResponseWriter) {
	j := jsonChat(chat)
	data := j.ToBytes()
	w.Header().Set("Content-Type", apContentType)
	w.Write(data)
}

func collectAddrs(addrs []string, j junk.Junk, field string) []string {
	if strarr, ok := j[field].([]string); ok {
		return strarr
	}
	arr := getArray(j, field)
	for _, a := range arr {
		if addr, ok := a.(string); ok {
			addrs = append(addrs, addr)
		}
	}
	return addrs
}

func sameOrigin(url1, url2 string) bool {
	return originate(url1) == originate(url2)
}
