//
// Copyright (c) 2023 Ted Unangst <tedu@tedunangst.com>
//
// Permission to use, copy, modify, and distribute this software for any
// purpose with or without fee is hereby granted, provided that the above
// copyright notice and this permission notice appear in all copies.
//
// THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
// WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
// ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
// WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
// ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
// OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

package main

import (
	"fmt"
	"html"
	"html/template"
	"net/http"
	"strings"
	"time"

	"humungus.tedunangst.com/r/webs/junk"
	"humungus.tedunangst.com/r/webs/login"
)

var allowSignup = false
var openSignups = false
var enableFedi = true
var adminUsers []int64
var siteAlert string
var HTsiteAlert template.HTML

func admincheck(next http.Handler) http.Handler {
	return http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
		u := login.GetUserInfo(r)
		user := finduser(u.Username)
		if !user.IsAdmin() {
			errorpage(w, r, "not an admin")
			return
		}
		next.ServeHTTP(w, r)
	})
}

func approvedcheck(next http.Handler) http.Handler {
	return http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
		u := login.GetUserInfo(r)
		user := finduser(u.Username)
		if !user.IsApproved() {
			errorpage(w, r, "not yet approved")
			return
		}
		next.ServeHTTP(w, r)
	})
}

func setSiteAlert() {
	alert := string(markdownShort(siteAlert))
	HTsiteAlert = template.HTML(alert)
}

func loadAdminSettings() {
	getconfig("servermsg", &serverMsg)
	getconfig("aboutmsg", &aboutMsg)
	getconfig("sitealert", &siteAlert)
	setSiteAlert()
	getconfig("loginmsg", &loginMsg)
	getconfig("signupmsg", &signupMsg)
	getconfig("allowsignup", &allowSignup)
	getconfig("opensignups", &openSignups)
	getconfig("enablefedi", &enableFedi)
	loadBanMap()
	users := allusers()
	for _, user := range users {
		if user.IsAdmin() {
			adminUsers = append(adminUsers, user.ID)
		}
	}
	loadFilters()
}

func showadmin(w http.ResponseWriter, r *http.Request) {
	templinfo := getInfo(r)
	templinfo["AboutMsg"] = aboutMsg
	templinfo["SiteAlertTxt"] = siteAlert
	templinfo["LoginMsg"] = loginMsg
	templinfo["SignupMsg"] = signupMsg
	templinfo["AllowSignup"] = allowSignup
	templinfo["OpenSignups"] = openSignups
	templinfo["EnableFedi"] = enableFedi
	templinfo["SaveCSRF"] = login.GetCSRF("saveadmin", r)
	err := readviews.Execute(w, "admin.html", templinfo)
	if err != nil {
		elog.Print(err)
	}
}

func showadmingroups(w http.ResponseWriter, r *http.Request) {
	templinfo := getInfo(r)
	templinfo["Groups"] = getgroups()
	templinfo["SaveCSRF"] = login.GetCSRF("saveadmin", r)
	err := readviews.Execute(w, "admingroups.html", templinfo)
	if err != nil {
		elog.Print(err)
	}
}

func showadminusers(w http.ResponseWriter, r *http.Request) {
	users := allusers()
	templinfo := getInfo(r)
	templinfo["Users"] = users
	templinfo["SaveCSRF"] = login.GetCSRF("saveadmin", r)
	err := readviews.Execute(w, "adminusers.html", templinfo)
	if err != nil {
		elog.Print(err)
	}
}

func showadminpeers(w http.ResponseWriter, r *http.Request) {
	banlist := getbanlist()
	templinfo := getInfo(r)
	templinfo["Banlist"] = banlist
	templinfo["SaveCSRF"] = login.GetCSRF("saveadmin", r)
	err := readviews.Execute(w, "adminpeers.html", templinfo)
	if err != nil {
		elog.Print(err)
	}
}

func showadminreports(w http.ResponseWriter, r *http.Request) {
	templinfo := getInfo(r)
	templinfo["Reports"] = getreports()
	templinfo["SaveCSRF"] = login.GetCSRF("saveadmin", r)
	err := readviews.Execute(w, "adminreports.html", templinfo)
	if err != nil {
		elog.Print(err)
	}
}

func isChecked(r *http.Request, name string, id int64) bool {
	key := fmt.Sprintf("%s%d", name, id)
	return r.FormValue(key) == "1"
}

func saveadmin(w http.ResponseWriter, r *http.Request) {
	pg := r.FormValue("pg")
	switch pg {
	case "site":
		saveadminsite(w, r)
	case "groups":
		saveadmingroups(w, r)
	case "users":
		saveadminusers(w, r)
	case "peers":
		saveadminpeers(w, r)
	case "reports":
		saveadminreports(w, r)
	case "filters":
		saveadminfilters(w, r)
	default:
		errorpage(w, r, "saving what now?")
	}
}

func saveadmingroups(w http.ResponseWriter, r *http.Request) {
	mtxSaveGroup.Lock()
	defer mtxSaveGroup.Unlock()
	groups := getgroups()
	for _, g := range groups {
		del := isChecked(r, "delete", g.ID)
		if del {
			ilog.Printf("deleting group %d %s", g.ID, g.Name)
			err := dbDeleteGroup(g)
			if err != nil {
				elog.Printf("error deleting group: %s", err)
			}
			continue
		}
		onfront := isChecked(r, "onfront", g.ID)
		if onfront != g.OnFront() {
			g.Flags ^= flagOnFront
			dbSaveGroup(g)
		}
	}
	http.Redirect(w, r, "/admin/groups", http.StatusSeeOther)
}

func saveadminusers(w http.ResponseWriter, r *http.Request) {
	mtxSavePerson.Lock()
	defer mtxSavePerson.Unlock()
	users := allusers()
	for _, u := range users {
		del := isChecked(r, "delete", u.ID)
		if del {
			ilog.Printf("deleting user %d %s", u.ID, u.Name)
			apSendDeletePerson(u.Person.APid)
			err := dbDeleteUser(u)
			if err != nil {
				elog.Printf("error deleting user: %s", err)
			}
			continue
		}
		approved := isChecked(r, "isapproved", u.ID)
		if approved != u.IsApproved() {
			u.Flags ^= flagIsApproved
			dbSaveUser(u)
		}
	}
	http.Redirect(w, r, "/admin/users", http.StatusSeeOther)
}

func saveadminpeers(w http.ResponseWriter, r *http.Request) {
	bans := r.FormValue("banlist")
	bans = strings.Replace(bans, "\r", "", -1)
	banlist := strings.Split(bans, "\n")
	ilog.Printf("saving new banlist: %d", len(banlist))
	err := dbSaveBanlist(banlist)
	if err != nil {
		elog.Printf("error saving banlist: %s", err)
	}
	loadBanMap()
	http.Redirect(w, r, "/admin/peers", http.StatusSeeOther)
}

func saveadminreports(w http.ResponseWriter, r *http.Request) {
	http.Redirect(w, r, "/admin/reports", http.StatusSeeOther)
}

func saveSetting(r *http.Request, name string, val interface{}) {
	formval := r.FormValue(name)
	if valp, ok := val.(*bool); ok {
		newval := formval == name
		if newval != *valp {
			dlog.Printf("setting bool %s: %v", name, newval)
			*valp = newval
			setconfig(name, newval)
		}
		return
	}
	if valp, ok := val.(*string); ok {
		newval := formval
		if newval != *valp {
			dlog.Printf("setting string %s: %v", name, newval)
			*valp = newval
			setconfig(name, newval)
			if valp == &siteAlert {
				setSiteAlert()
			}
		}
		return
	}
	dlog.Printf("unknown setting type: %s", name)
}

func saveadminsite(w http.ResponseWriter, r *http.Request) {
	saveSetting(r, "aboutmsg", &aboutMsg)
	saveSetting(r, "sitealert", &siteAlert)
	saveSetting(r, "loginmsg", &loginMsg)
	saveSetting(r, "signupmsg", &signupMsg)
	saveSetting(r, "allowsignup", &allowSignup)
	saveSetting(r, "opensignups", &openSignups)
	saveSetting(r, "enablefedi", &enableFedi)

	http.Redirect(w, r, "/admin", http.StatusSeeOther)
}

func (group *Group) IsMod(user *User) bool {
	if user == nil {
		return false
	}
	if user.IsAdmin() {
		return true
	}
	for _, mod := range group.Mods {
		if mod == user.Person.Name {
			return true
		}
	}
	return false
}

func (post *Post) CanEdit(user *User) bool {
	if user == nil {
		return false
	}
	if post.PersonID == user.Person.ID {
		return true
	}
	return false
}

func (post *Post) CanMod(user *User) bool {
	if user == nil {
		return false
	}
	if post.PersonID == user.Person.ID {
		return true
	}
	group := findgroup(post.GroupName)
	if group.IsMod(user) {
		return true
	}
	return false
}

func (post *Post) CanApprove(user *User) bool {
	if user == nil {
		return false
	}
	group := findgroup(post.GroupName)
	if group.IsMod(user) {
		return true
	}
	return false
}

func (comment *Comment) CanEdit(user *User) bool {
	if user == nil {
		return false
	}
	if comment.PersonID == user.Person.ID {
		return true
	}
	return false
}

func (comment *Comment) CanMod(user *User) bool {
	if user == nil {
		return false
	}
	if comment.PersonID == user.Person.ID {
		return true
	}
	post := findpost(comment.PostXID)
	if post.CanMod(user) {
		return true
	}
	return false
}

var banMap = make(map[string]bool)

func loadBanMap() {
	newmap := make(map[string]bool)
	banlist := getbanlist()
	for _, who := range banlist {
		newmap[who] = true
	}
	banMap = newmap
}

func isBannedHost(host string) bool {
	return banMap[host]
}

func isBannedPerson(apid string) bool {
	return banMap[apid] || banMap[originate(apid)]
}

func isBannedRequest(r *http.Request) bool {
	who := requestOrigin(r)
	if isBannedHost(who) {
		dlog.Printf("banned request from %s", who)
		return true
	}
	return false
}

func getbanlist() []string {
	var banlist []string
	rows, err := stmtBanlist.Query()
	if err != nil {
		elog.Printf("error getting banlist: %s", err)
		return banlist
	}
	defer rows.Close()
	for rows.Next() {
		var who string
		err = rows.Scan(&who)
		if err != nil {
			elog.Printf("error scanning banlist: %s", err)
			continue
		}
		banlist = append(banlist, who)
	}
	return banlist
}

func dbSaveBanlist(banlist []string) error {
	db := opendatabase()
	tx, err := db.Begin()
	if err != nil {
		return err
	}
	defer func() {
		if tx != nil {
			tx.Rollback()
		}
	}()
	_, err = tx.Exec("delete from peers where what = 'banhost'")
	if err != nil {
		return err
	}
	for _, who := range banlist {
		if who == "" {
			continue
		}
		_, err = tx.Exec("insert into peers (who, what, notes) values (?, ?, ?)",
			who, "banhost", "banned")
		if err != nil {
			return err
		}
	}
	err = tx.Commit()
	if err == nil {
		tx = nil
	}
	return err
}

func webreport(w http.ResponseWriter, r *http.Request) {
	what := r.FormValue("what")

	templinfo := getInfo(r)
	templinfo["What"] = what
	templinfo["SaveCSRF"] = login.GetCSRF("savepost", r)
	err := readviews.Execute(w, "report.html", templinfo)
	if err != nil {
		elog.Print(err)
	}
}

func websavereport(w http.ResponseWriter, r *http.Request) {
	what := r.FormValue("what")
	var blame, where, link string
	group := findgroup(what)
	if group == nil {
		post := findpost(what)
		if post == nil {
			comment := findcomment(what)
			if comment == nil {
				person := findperson(what)
				if person == nil {
					errorpage(w, r, "report on what?")
					return
				}
				blame = person.APid
				link = person.Rel()
			} else {
				person := findperson(comment.PersonName)
				blame = person.APid
				link = comment.Rel()
				post = findpost(comment.PostXID)
			}
		} else {
			person := findperson(post.PersonName)
			blame = person.APid
			link = post.Rel()
		}
		if post != nil {
			group = findgroup(post.GroupName)
		}
	} else {
		blame = group.APid
		link = group.Rel()
	}
	if group != nil {
		where = group.APid
	}
	why := r.FormValue("why")
	me := personforuser(r)
	report := &Report{
		Who:   me.APid,
		What:  what,
		Where: where,
		Link:  link,
		Why:   why,
		Blame: blame,
		Date:  time.Now(),
	}
	ilog.Printf("saving report from %s on %s", me.Name, what)
	err := dbSaveReport(report)
	if err != nil {
		elog.Printf("error saving report: %s", err)
	}
	notifNewReport(report)
	apSendReport(report)

	http.Redirect(w, r, "/", http.StatusSeeOther)
}

func apHandleFlag(obj junk.Junk) {
	actor, _ := obj.GetString("actor")
	target, _ := obj.GetString("object")
	if !isMine(target) {
		dlog.Printf("bad flag: %s", obj.ToString())
		return
	}
	why, _ := obj.GetString("summary")
	why = html.UnescapeString(why)
	var where, link string
	audience, _ := obj.GetString("audience")
	where = audience
	report := &Report{
		Who:   actor,
		What:  target,
		Where: where,
		Link:  link,
		Why:   why,
		Date:  time.Now(),
	}
	ilog.Printf("saving report from %s on %s", actor, target)
	err := dbSaveReport(report)
	if err != nil {
		elog.Printf("error saving report: %s", err)
	}
	notifNewReport(report)
}

func jsonReport(report *Report) junk.Junk {
	j := jsonActivity("Flag", report.Who, report.Where, report.What)
	j["to"] = report.Blame
	j["summary"] = html.EscapeString(report.Why)
	j["published"] = report.Date.UTC().Format(time.RFC3339)
	return j
}

func apSendReport(report *Report) {
	j := jsonReport(report)
	rcpts := []string{report.Blame, report.Where}
	apSendToRcpts(report.Who, rcpts, j.ToBytes())
}
