//
// Copyright (c) 2023 Ted Unangst <tedu@tedunangst.com>
//
// Permission to use, copy, modify, and distribute this software for any
// purpose with or without fee is hereby granted, provided that the above
// copyright notice and this permission notice appear in all copies.
//
// THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
// WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
// ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
// WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
// ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
// OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

package main

import (
	"bytes"
	"crypto/rand"
	"database/sql"
	_ "embed"
	"encoding/json"
	"fmt"
	"regexp"
	"sort"
	"strconv"
	"strings"
	"sync"
	"time"

	"humungus.tedunangst.com/r/webs/gencache"
)

//go:embed schema.sql
var sqlSchema string

var dbtimeformat = "2006-01-02 15:04:05"

var debugQueries = false

func jsonify(what interface{}) (string, error) {
	var buf bytes.Buffer
	e := json.NewEncoder(&buf)
	e.SetEscapeHTML(false)
	e.SetIndent("", "")
	err := e.Encode(what)
	return buf.String(), err
}

func unjsonify(s string, dest interface{}) error {
	d := json.NewDecoder(strings.NewReader(s))
	err := d.Decode(dest)
	return err
}

func xcelerate(b []byte) string {
	letters := "BCDFGHJKLMNPQRSTVWXYZbcdfghjklmnpqrstvwxyz1234567891234567891234"
	for i, c := range b {
		b[i] = letters[c&63]
	}
	s := string(b)
	return s
}

func generateXid() string {
	var b [18]byte
	rand.Read(b[:])
	return xcelerate(b[:])
}

func generateDailyXid() string {
	var b [21]byte
	rand.Read(b[:])
	now := time.Now().Unix() / 60 / 60 / 24
	b[2] = byte(now & 63)
	b[1] = byte((now / 64) & 63)
	b[0] = byte((now / 64 / 64) & 63)
	return xcelerate(b[:])
}

func generateNiceXid(title string) string {
	buf := make([]byte, len(title))
	i := 0
	for _, c := range title {
		if (c >= 'a' && c <= 'z') ||
			(c >= 'A' && c <= 'Z') ||
			(c >= '0' && c <= '9') {
			buf[i] = byte(c)
			i++
		}
		if c == ' ' {
			buf[i] = '-'
			i++
		}
	}
	if i > 48 {
		i = 48
	}
	for i > 0 && buf[i-1] == '-' {
		i--
	}
	return generateXid() + "-" + string(buf[:i])
}

type RowLike interface {
	Scan(dest ...interface{}) error
}

var stmtFindPerson, stmtFindPersonByAPid, stmtGetPerson, stmtSavePerson, stmtUpdatePerson *sql.Stmt
var stmtUpdateUser, stmtFindUser, stmtGetUser, stmtAllUsers, stmtSaveUser *sql.Stmt
var stmtSaveGroup, stmtUpdateGroup, stmtGetGroups, stmtGetPublicGroups, stmtFindGroup, stmtFindGroupByAPid *sql.Stmt
var stmtSavePost, stmtUpdatePost, stmtGetPostsForGroup, stmtGetPostsByPerson, stmtGetPostsForFront, stmtGetPostsForUser, stmtFindPost, stmtFindPostByAPid *sql.Stmt
var sqlGetPosts string
var stmtFindPostForLink *sql.Stmt
var stmtSaveComment, stmtUpdateComment, stmtGetComments, stmtGetComment, stmtFindComment, stmtFindCommentByAPid *sql.Stmt
var stmtSaveChat, stmtFindChat, stmtGetChats *sql.Stmt
var stmtSaveSub, stmtDeleteSub, stmtGetSubsForUser, stmtGetSubsToGroup, stmtGetSubsToPerson, stmtGetSubsToPost, stmtFindSub *sql.Stmt
var sqlGetAttachments string
var stmtSaveAttachment, stmtDeletePostAttachments, stmtUpdateAttachment, stmtGetAttachments *sql.Stmt
var stmtFindFile, stmtFindFileHash, stmtFindFileUrl, stmtSaveFileInfo *sql.Stmt
var stmtSavePubkey, stmtFindPubkey, stmtSaveSeckey, stmtFindSeckey *sql.Stmt
var stmtSaveFollow, stmtDeleteFollow, stmtFindFollow, stmtFindFollowers, stmtFindFollowing *sql.Stmt
var stmtSaveShare, stmtDeleteShare, stmtFindShare *sql.Stmt
var stmtAddDelivery, stmtGetDeliveries, stmtLoadDelivery, stmtZapDelivery, stmtDelinquentCheck, stmtDelinquentUpdate *sql.Stmt
var stmtFindTombstone, stmtSaveTombstone, stmtBanlist *sql.Stmt
var stmtSaveReport, stmtUpdateReport, stmtGetReports, stmtGetReportsByWhere *sql.Stmt
var stmtSaveNotif, stmtDeleteNewNotif, stmtGetNotifs, stmtClearNotifs *sql.Stmt
var stmtSaveFilter, stmtGetFilters, stmtDeleteFilter *sql.Stmt

func prepareStatements(db *sql.DB) {
	try := func(s string) *sql.Stmt {
		stmt, err := db.Prepare(s)
		if err != nil {
			elog.Fatalf("error %s: %s", err, s)
		}
		return stmt
	}

	// persons
	personcols := "personid, pname, display, papid, personmeta"
	stmtFindPerson = try("select " + personcols + " from persons where pname = ?")
	stmtFindPersonByAPid = try("select " + personcols + " from persons where papid = ?")
	stmtGetPerson = try("select " + personcols + " from persons where personid = ?")
	stmtSavePerson = try("insert into persons (" + personcols[10:] + ") values (?, ?, ?, ?)")
	stmtUpdatePerson = try("update persons set (" + personcols[10:] + ") = (?, ?, ?, ?) where personid = ?")
	usercols := "userid, username, hash, flags, email, personid"
	stmtGetUser = try("select " + usercols + " from users where userid = ?")
	stmtFindUser = try("select " + usercols + " from users where username = ?")
	stmtAllUsers = try("select " + usercols + " from users")
	stmtSaveUser = try("insert into users (" + usercols[8:] + ") values (?, ?, ?, ?, ?)")
	stmtUpdateUser = try("update users set (" + usercols[8:] + ") = (?, ?, ?, ?, ?) where userid = ?")
	// groups
	groupcols := "groupid, gname, gdisplay, gapid, flags, gmeta"
	stmtSaveGroup = try("insert into groups (" + groupcols[9:] + ") values (?, ?, ?, ?, ?)")
	stmtUpdateGroup = try("update groups set (" + groupcols[9:] + ") = (?, ?, ?, ?, ?) where groupid = ?")
	stmtGetGroups = try("select " + groupcols + " from groups")
	stmtGetPublicGroups = try("select " + groupcols + " from groups where flags & 1 = 1")
	stmtFindGroup = try("select " + groupcols + " from groups where gname = ?")
	stmtFindGroupByAPid = try("select " + groupcols + " from groups where gapid = ?")
	// posts
	postcols := "postid, pgroupid, ppersonid, postxid, postapid, dt, title, lnk, content, postmeta"
	postjoins := "posts join groups on posts.pgroupid = groups.groupid join persons on posts.ppersonid = persons.personid left outer join comments on posts.postid = comments.cpostid"
	postjoincols := postcols + ", gname, pname, count(comments.commentid)"
	stmtSavePost = try("insert into posts (" + postcols[8:] + ") values (?, ?, ?, ?, ?, ?, ?, ?, ?)")
	stmtUpdatePost = try("update posts set (" + postcols[8:] + ") = (?, ?, ?, ?, ?, ?, ?, ?, ?) where postid = ?")
	sqlGetPosts = "select " + postjoincols + " from " + postjoins + " WHERE group by postid order by postid desc limit 200"
	stmtGetPostsForFront = try(strings.ReplaceAll(sqlGetPosts, "WHERE", "where groups.flags & 1 = 1"))
	stmtGetPostsForUser = try("select " + postjoincols + " from " + postjoins + " where groupid in (select groupid from subs where userid = ? and groupid <> 0) or postid in (select postid from shares where personid in (select personid from subs where userid = ? and personid <> 0)) group by postid order by postid desc limit 200")
	stmtGetPostsForGroup = try("select " + postjoincols + " from " + postjoins + " where pgroupid = ? group by postid order by postid desc limit 200")
	stmtGetPostsByPerson = try("select " + postjoincols + " from " + postjoins + " where ppersonid = ? group by postid order by postid desc limit 200")
	stmtFindPost = try("select " + postjoincols + " from " + postjoins + " where postxid = ? group by postid")
	stmtFindPostByAPid = try("select " + postjoincols + " from " + postjoins + " where postapid = ? group by postid")
	stmtFindPostForLink = try("select postxid from posts where pgroupid = ? and lnk = ? and postid in (select postid from posts order by postid desc limit 20)")
	// comments
	commentcols := "commentid, cpostid, cpersonid, parentid, cxid, capid, cdt, subject, ccontent, cmeta"
	commentjoins := "comments join persons on comments.cpersonid = persons.personid"
	onecommentjoins := commentjoins + " join posts on comments.cpostid = posts.postid"
	commentjoincols := commentcols + ", pname, display, '.'"
	onecommentjoincols := commentcols + ", pname, display, postxid"
	stmtSaveComment = try("insert into comments (" + commentcols[11:] + ") values (?, ?, ?, ?, ?, ?, ?, ?, ?)")
	stmtUpdateComment = try("update comments set (" + commentcols[11:] + ") = (?, ?, ?, ?, ?, ?, ?, ?, ?) where commentid = ?")
	stmtGetComments = try("select " + commentjoincols + " from " + commentjoins + " where cpostid = ? limit 1000")
	stmtGetComment = try("select " + onecommentjoincols + " from " + onecommentjoins + " where commentid = ?")
	stmtFindComment = try("select " + onecommentjoincols + " from " + onecommentjoins + " where cxid = ?")
	stmtFindCommentByAPid = try("select " + onecommentjoincols + " from " + onecommentjoins + " where capid = ?")
	chatcols := "chatid, cgroupid, cpersonid, cxid, capid, cdt, ccontent"
	chatjoins := "chats join persons on chats.cpersonid = persons.personid join groups on chats.cgroupid = groups.groupid"
	chatjoincols := chatcols + ", pname, gname"
	stmtSaveChat = try("insert into chats (" + chatcols[8:] + ") values (?, ?, ?, ?, ?, ?)")
	stmtGetChats = try("select " + chatjoincols + " from " + chatjoins + " where cgroupid = ? order by chatid desc limit 200")
	stmtFindChat = try("select " + chatjoincols + " from " + chatjoins + " where cxid = ?")
	// subs
	stmtSaveSub = try("insert into subs (userid, groupid, personid, postid) values (?, ?, ?, ?)")
	stmtDeleteSub = try("delete from subs where userid = ? and groupid = ? and personid = ? and postid = ?")
	stmtFindSub = try("select subid from subs where userid = ? and groupid = ? and personid = ? and postid = ?")
	stmtGetSubsForUser = try("select groupid, personid, postid from subs where userid = ?")
	stmtGetSubsToGroup = try("select userid from subs where groupid > 0 and groupid = ?")
	stmtGetSubsToPerson = try("select userid from subs where personid > 0 and personid = ?")
	stmtGetSubsToPost = try("select userid from subs where postid > 0 and postid = ?")
	// attachments
	attcols := "attachmentid, postid, xid, media, description"
	stmtSaveAttachment = try("insert into attachments (" + attcols[14:] + ") values (?, ?, ?, ?)")
	stmtUpdateAttachment = try("update attachments set description = ? where attachmentid = ?")
	sqlGetAttachments = "select " + attcols + " from attachments"
	stmtGetAttachments = try(sqlGetAttachments + " where postid = ?")
	stmtDeletePostAttachments = try("delete from attachments where postid = ?")
	stmtFindFile = try("select xid, media from files where xid = ?")
	stmtFindFileHash = try("select xid, media from files where hash = ?")
	stmtFindFileUrl = try("select xid, media from files where url = ?")
	stmtSaveFileInfo = try("insert into files (hash, url, media, xid) values (?, ?, ?, ?)")
	// pubkeys
	stmtSavePubkey = try("insert into pubkeys (url, owner, pubkey, dt) values (?, ?, ?, ?)")
	stmtFindPubkey = try("select pubkey from pubkeys where url = ?")
	stmtSaveSeckey = try("insert into seckeys (owner, seckey) values (?, ?)")
	stmtFindSeckey = try("select seckey from seckeys where owner = ?")
	stmtSaveFollow = try("insert into follows (folxid, who, what) values (?, ?, ?)")
	stmtDeleteFollow = try("delete from follows where who = ? and what = ?")
	stmtFindFollow = try("select folxid from follows where who = ? and what = ?")
	stmtFindFollowers = try("select who from follows where what = ?")
	stmtFindFollowing = try("select what from follows where who = ?")
	stmtSaveShare = try("insert into shares (personid, postid) values (?, ?)")
	stmtDeleteShare = try("delete from shares where personid = ? and postid = ?")
	stmtFindShare = try("select shareid from shares where personid = ? and postid = ?")
	// deliveries
	stmtAddDelivery = try("insert into deliveries (dt, tries, rcpt, data) values (?, ?, ?, ?)")
	stmtGetDeliveries = try("select deliveryid, dt from deliveries")
	stmtLoadDelivery = try("select tries, rcpt, data from deliveries where deliveryid = ?")
	stmtZapDelivery = try("delete from deliveries where deliveryid = ?")

	stmtDelinquentCheck = try("select deliveryid, data from deliveries where rcpt = ?")
	stmtDelinquentUpdate = try("update deliveries set data = ? where deliveryid = ?")
	stmtBanlist = try("select who from peers where what = 'banhost'")
	reportcols := "reportid, who, what, were, dt, meta"
	stmtSaveReport = try("insert into reports (" + reportcols[10:] + ") values (?, ?, ?, ?, ?)")
	stmtUpdateReport = try("update reports set (" + reportcols[10:] + ") = (?, ?, ?, ?, ?) where reportid = ?")
	stmtGetReports = try("select " + reportcols + " from reports order by reportid desc limit 200")
	stmtGetReportsByWhere = try("select " + reportcols + " from reports where were = ? order by reportid desc limit 200")
	stmtFindTombstone = try("select tombstoneid from tombstones where what = ?")
	stmtSaveTombstone = try("insert into tombstones (what) values (?)")
	stmtSaveNotif = try("insert into notifs (userid, dt, flags, msg, url) values (?, ?, ?, ?, ?)")
	stmtDeleteNewNotif = try("delete from notifs where userid = ? and msg = ? and url = ? and flags & 1 = 1")
	stmtGetNotifs = try("select notifid, userid, dt, flags, msg, url from notifs where userid = ? order by notifid desc limit 100")
	stmtClearNotifs = try("update notifs set flags = 0 where userid = ?")
	stmtSaveFilter = try("insert into filters (userid, data) values (?, ?)")
	stmtGetFilters = try("select filterid, userid, data from filters")
	stmtDeleteFilter = try("delete from filters where userid = ? and filterid = ?")
}

func dbQuery(msg string, args ...interface{}) {
	if debugQueries {
		dlog.Printf("dbQuery: "+msg, args...)
	}
}

var tombcache = gencache.New(gencache.Options[string, bool]{
	Fill: func(what string) (bool, bool) {
		dbQuery("findtombstone %s", what)
		row := stmtFindTombstone.QueryRow(what)
		var id int64
		err := row.Scan(&id)
		if err != nil && err != sql.ErrNoRows {
			elog.Printf("error scanning tombstone: %s", err)
		}
		return id > 0, true
	},
	Invalidators: []*gencache.Invalidator[string]{
		&groupInvalidator, &personInvalidator, &postInvalidator, &commentInvalidator,
	},
	Limit: 4000,
})

func hastombstone(what string) bool {
	gone, _ := tombcache.Get(what)
	return gone
}

func dbSaveTombstone(what string) error {
	_, err := stmtSaveTombstone.Exec(what)
	return err
}

var mtxSaveGroup sync.Mutex
var mtxSavePost sync.Mutex
var mtxSaveComment sync.Mutex
var mtxSavePerson sync.Mutex

type personMeta struct {
	About    string `json:",omitempty"`
	Format   string `json:",omitempty"`
	Inbox    string `json:",omitempty"`
	ShareBox string `json:",omitempty"`
}

func txDeletePerson(tx *sql.Tx, person *Person) error {
	var err error
	try := func(s string, args ...interface{}) {
		if err == nil {
			_, err = tx.Exec(s, args...)
		}
	}
	try("delete from follows where who = ? or what = ?", person.APid, person.APid)
	try("delete from subs where personid = ?", person.ID)
	try("delete from reports where who = ?", person.APid)
	try("delete from shares where personid = ?", person.ID)
	try("delete from chats where cpersonid = ?", person.ID)
	try("delete from subs where postid in (select postid from posts where ppersonid = ?)", person.ID)
	try("delete from shares where postid in (select postid from posts where ppersonid = ?)", person.ID)
	try("delete from comments where cpostid in (select postid from posts where ppersonid = ?)", person.ID)
	try("delete from attachments where postid in (select postid from posts where ppersonid = ?)", person.ID)
	try("delete from posts where ppersonid = ?", person.ID)
	try("update comments set cpersonid = -1, capid = '.', subject = '', ccontent = '.' where cpersonid = ?", person.ID)
	try("delete from persons where personid = ?", person.ID)
	try("insert into tombstones (what) values (?)", person.APid)
	return err
}

func dbDeletePerson(person *Person) error {
	db := opendatabase()
	tx, err := db.Begin()
	if err != nil {
		return err
	}
	err = txDeletePerson(tx, person)
	if err != nil {
		tx.Rollback()
	} else {
		err = tx.Commit()
	}
	personInvalidator.Clear(person.Name)
	personInvalidator.Clear(person.APid)
	postInvalidator.Flush()
	commentInvalidator.Flush()
	return err
}

func dbSavePerson(person *Person) error {
	if person.APid == "" {
		person.APid = fmt.Sprintf("https://%s/u/%s", serverName, person.Name)
	}
	meta := personMeta{
		About:    person.About,
		Format:   person.Format,
		Inbox:    person.Inbox,
		ShareBox: person.ShareBox,
	}
	j, err := jsonify(&meta)
	if err == nil {
		if person.ID == 0 {
			var res sql.Result
			res, err = stmtSavePerson.Exec(person.Name, person.Display, person.APid, j)
			if err == nil {
				person.ID, _ = res.LastInsertId()
			}
		} else {
			_, err = stmtUpdatePerson.Exec(person.Name, person.Display, person.APid, j, person.ID)
		}
	}
	personInvalidator.Clear(person.Name)
	personInvalidator.Clear(person.APid)
	return err
}

func scanperson(row *sql.Row) *Person {
	person := new(Person)
	var j string
	err := row.Scan(&person.ID, &person.Name, &person.Display, &person.APid, &j)
	if err == sql.ErrNoRows {
		return nil
	}
	if err == nil {
		var meta personMeta
		err = unjsonify(j, &meta)
		if err == nil {
			person.About = meta.About
			person.Format = meta.Format
			person.Inbox = meta.Inbox
			person.ShareBox = meta.ShareBox
		}
	}
	if err != nil {
		elog.Printf("error scanning person: %s", err)
		return nil
	}
	return person
}

var personInvalidator gencache.Invalidator[string]
var personCache = gencache.New(gencache.Options[string, *Person]{
	Fill: func(ref string) (*Person, bool) {
		person := dbFindPerson(ref)
		return person, true
	},
	Invalidator: &personInvalidator, Limit: 600})

func findperson(ref string) *Person {
	person, _ := personCache.Get(ref)
	return person
}

func dbFindPerson(ref string) *Person {
	dbQuery("findperson %s", ref)
	var person *Person
	if strings.HasPrefix(ref, "https://") {
		row := stmtFindPersonByAPid.QueryRow(ref)
		person = scanperson(row)
	} else {
		row := stmtFindPerson.QueryRow(ref)
		person = scanperson(row)
	}
	if person != nil {
		reformatperson(person)
	}
	return person
}

func dbSaveUser(user *User) error {
	var err error
	if user.ID == 0 {
		var res sql.Result
		res, err := stmtSaveUser.Exec(user.Name, user.Hash, user.Flags, user.Email, user.Person.ID)
		if err == nil {
			user.ID, _ = res.LastInsertId()
		}
	} else {
		_, err = stmtUpdateUser.Exec(user.Name, user.Hash, user.Flags, user.Email, user.Person.ID, user.ID)
	}
	personInvalidator.Clear(user.Name)
	return err
}

func dbDeleteUser(user *User) error {
	db := opendatabase()
	tx, err := db.Begin()
	if err != nil {
		return err
	}
	try := func(s string, args ...interface{}) {
		if err == nil {
			_, err = tx.Exec(s, args...)
		}
	}
	person := user.Person
	err = txDeletePerson(tx, person)
	try("delete from filters where userid = ?", user.ID)
	try("delete from subs where userid = ?", user.ID)
	try("delete from users where userid = ?", user.ID)
	try("insert into tombstones (what) values (?)", user.Name)
	if err != nil {
		tx.Rollback()
	} else {
		err = tx.Commit()
	}
	personInvalidator.Clear(person.Name)
	personInvalidator.Clear(person.APid)
	postInvalidator.Flush()
	commentInvalidator.Flush()
	return err
}

func scanuser(row RowLike) *User {
	user := new(User)
	var personid int64
	err := row.Scan(&user.ID, &user.Name, &user.Hash, &user.Flags, &user.Email, &personid)
	if err == sql.ErrNoRows {
		return nil
	}
	if err == nil {
		user.Person = findperson(user.Name)
		if user.Person == nil {
			elog.Printf("lost person %d for user %s", personid, user.Name)
			return nil
		}
	}
	if err != nil {
		elog.Printf("error scanning user: %s", err)
		return nil
	}
	return user
}

var userCache = gencache.New(gencache.Options[string, *User]{
	Fill: func(name string) (*User, bool) {
		user := dbFindUser(name)
		return user, true
	},
	Invalidator: &personInvalidator, Limit: 600})

func finduser(name string) *User {
	user, _ := userCache.Get(name)
	return user
}

func dbFindUser(name string) *User {
	dbQuery("finduser %s", name)
	row := stmtFindUser.QueryRow(name)
	return scanuser(row)
}

func dbGetUser(userid int64) *User {
	dbQuery("getuser %d", userid)
	row := stmtGetUser.QueryRow(userid)
	return scanuser(row)
}

func allusers() []*User {
	dbQuery("allusers")
	var users []*User
	rows, err := stmtAllUsers.Query()
	if err != nil {
		elog.Printf("error finding new users: %s", err)
		return users
	}
	defer rows.Close()
	for rows.Next() {
		user := scanuser(rows)
		if user == nil {
			continue
		}
		users = append(users, user)
	}
	return users
}

type groupMeta struct {
	Desc      string    `json:",omitempty"`
	Format    string    `json:",omitempty"`
	Mods      []string  `json:",omitempty"`
	Featured  []string  `json:",omitempty"`
	Inbox     string    `json:",omitempty"`
	ShareBox  string    `json:",omitempty"`
	Outbox    string    `json:",omitempty"`
	Refreshed time.Time `json:",omitempty"`
}

func dbDeleteGroup(group *Group) error {
	db := opendatabase()
	tx, err := db.Begin()
	if err != nil {
		return err
	}
	try := func(s string, args ...interface{}) {
		if err == nil {
			_, err = tx.Exec(s, args...)
		}
	}
	try("delete from follows where who = ? or what = ?", group.APid, group.APid)
	try("delete from subs where groupid = ?", group.ID)
	try("delete from subs where postid in (select postid from posts where pgroupid = ?)", group.ID)
	try("delete from shares where postid in (select postid from posts where pgroupid = ?)", group.ID)
	try("delete from comments where cpostid in (select postid from posts where pgroupid = ?)", group.ID)
	try("delete from attachments where postid in (select postid from posts where pgroupid = ?)", group.ID)
	try("delete from posts where postid in (select postid from posts where pgroupid = ?)", group.ID)
	try("delete from groups where groupid = ?", group.ID)
	if group.IsLocal() {
		try("insert into tombstones (what) values (?)", group.Name)
	}
	try("insert into tombstones (what) values (?)", group.APid)
	if err != nil {
		tx.Rollback()
	} else {
		err = tx.Commit()
	}
	groupInvalidator.Clear(group.Name)
	groupInvalidator.Clear(group.APid)
	postInvalidator.Flush()
	commentInvalidator.Flush()
	return err
}

func dbSaveGroup(group *Group) error {
	if group.APid == "" {
		group.APid = fmt.Sprintf("https://%s/g/%s", serverName, group.Name)
	}
	meta := groupMeta{
		Desc:      group.Desc,
		Format:    group.Format,
		Mods:      group.Mods,
		Featured:  group.Featured,
		Inbox:     group.Inbox,
		ShareBox:  group.ShareBox,
		Outbox:    group.Outbox,
		Refreshed: group.Refreshed,
	}
	j, err := jsonify(&meta)
	if err == nil {
		if group.ID == 0 {
			var res sql.Result
			res, err = stmtSaveGroup.Exec(group.Name, group.Display, group.APid, group.Flags, j)
			if err == nil {
				group.ID, _ = res.LastInsertId()
			}
		} else {
			_, err = stmtUpdateGroup.Exec(group.Name, group.Display, group.APid, group.Flags, j, group.ID)
		}
	}
	groupInvalidator.Clear(group.Name)
	groupInvalidator.Clear(group.APid)
	return err
}

func scangroup(row RowLike) *Group {
	group := new(Group)
	var j string
	err := row.Scan(&group.ID, &group.Name, &group.Display, &group.APid, &group.Flags, &j)
	if err == sql.ErrNoRows {
		return nil
	}
	if err == nil {
		var meta groupMeta
		err = unjsonify(j, &meta)
		if err == nil {
			group.Desc = meta.Desc
			group.Format = meta.Format
			group.Mods = meta.Mods
			group.Featured = meta.Featured
			group.Inbox = meta.Inbox
			group.ShareBox = meta.ShareBox
			group.Outbox = meta.Outbox
			group.Refreshed = meta.Refreshed
		}
	}
	if err != nil {
		elog.Printf("error scanning group: %s", err)
		return nil
	}
	return group
}

func scangroups(rows *sql.Rows, err error) []*Group {
	var groups []*Group
	if err != nil {
		elog.Printf("error getting groups: %s", err)
		return groups
	}
	defer rows.Close()
	for rows.Next() {
		group := scangroup(rows)
		if group == nil {
			continue
		}
		groups = append(groups, group)
	}
	return groups
}

func getgroups() []*Group {
	dbQuery("getgroups")
	rows, err := stmtGetGroups.Query()
	groups := scangroups(rows, err)
	sort.Slice(groups, func(i, j int) bool {
		if groups[i].IsLocal() == groups[j].IsLocal() {
			return groups[i].Name < groups[j].Name
		}
		return groups[i].IsLocal()
	})
	return groups
}

func getpublicgroups() []*Group {
	dbQuery("getpublicgroups")
	rows, err := stmtGetPublicGroups.Query()
	groups := scangroups(rows, err)
	sort.Slice(groups, func(i, j int) bool {
		if groups[i].IsLocal() == groups[j].IsLocal() {
			return groups[i].Name < groups[j].Name
		}
		return groups[i].IsLocal()
	})
	return groups
}

var groupInvalidator gencache.Invalidator[string]
var groupCache = gencache.New(gencache.Options[string, *Group]{
	Fill: func(ref string) (*Group, bool) {
		group := dbFindGroup(ref)
		return group, true
	},
	Invalidator: &groupInvalidator, Limit: 600})

func findgroup(ref string) *Group {
	group, _ := groupCache.Get(ref)
	return group
}

func dbFindGroup(ref string) *Group {
	dbQuery("findgroup %s", ref)
	var group *Group
	if strings.HasPrefix(ref, "https://") {
		row := stmtFindGroupByAPid.QueryRow(ref)
		group = scangroup(row)
	} else {
		row := stmtFindGroup.QueryRow(ref)
		group = scangroup(row)
	}
	if group != nil {
		reformatgroup(group)
	}
	return group
}

type postMeta struct {
	Pending bool     `json:",omitempty"`
	Format  string   `json:",omitempty"`
	Context string   `json:",omitempty"`
	Tags    []string `json:",omitempty"`
}

func dbDeletePost(post *Post) error {
	person := findperson(post.PersonName)
	group := findgroup(post.GroupName)
	db := opendatabase()
	tx, err := db.Begin()
	if err != nil {
		return err
	}
	try := func(s string, args ...interface{}) {
		if err == nil {
			_, err = tx.Exec(s, args...)
		}
	}
	try("delete from subs where postid = ?", post.ID)
	try("delete from shares where postid = ?", post.ID)
	try("delete from comments where cpostid = ?", post.ID)
	try("delete from attachments where postid = ?", post.ID)
	try("delete from posts where postid = ?", post.ID)
	try("insert into tombstones (what) values (?)", post.APid)
	if err != nil {
		tx.Rollback()
	} else {
		err = tx.Commit()
	}
	if person.IsLocal() {
		personInvalidator.Clear(person.APid)
	}
	if group.IsLocal() {
		groupInvalidator.Clear(group.APid)
	}
	postInvalidator.Clear(post.Xid)
	postInvalidator.Clear(post.APid)
	commentInvalidator.Flush()
	return err
}

func dbSavePost(post *Post) error {
	person := findperson(post.PersonName)
	group := findgroup(post.GroupName)
	if post.Xid == "" {
		post.Xid = generateNiceXid(post.Title)
	}
	if post.APid == "" {
		post.APid = fmt.Sprintf("https://%s/g/%s/p/%s", serverName, post.GroupName, post.Xid)
	}
	if post.Context == "" {
		post.Context = post.APid
	}
	meta := postMeta{
		Pending: post.Pending,
		Format:  post.Format,
		Context: post.Context,
		Tags:    post.Tags,
	}
	j, err := jsonify(&meta)
	if err == nil {
		dt := post.Date.UTC().Format(dbtimeformat)
		if post.ID == 0 {
			var res sql.Result
			res, err = stmtSavePost.Exec(post.GroupID, post.PersonID, post.Xid, post.APid, dt, post.Title, post.Link, post.Content, j)
			if err == nil {
				post.ID, _ = res.LastInsertId()
			}
		} else {
			_, err = stmtUpdatePost.Exec(post.GroupID, post.PersonID, post.Xid, post.APid, dt, post.Title, post.Link, post.Content, j, post.ID)
		}
	}
	if person.IsLocal() {
		personInvalidator.Clear(person.APid)
	}
	if group.IsLocal() {
		groupInvalidator.Clear(group.APid)
	}
	postInvalidator.Clear(post.Xid)
	postInvalidator.Clear(post.APid)
	return err
}

var re_linkhost = regexp.MustCompile("https?://([^/ #)]+)")

func scanpost(row RowLike) *Post {
	post := new(Post)
	var dt, j string
	err := row.Scan(&post.ID, &post.GroupID, &post.PersonID, &post.Xid, &post.APid, &dt, &post.Title, &post.Link, &post.Content, &j, &post.GroupName, &post.PersonName, &post.NumComments)
	if err == sql.ErrNoRows {
		return nil
	}
	if err == nil {
		var meta postMeta
		err = unjsonify(j, &meta)
		if err == nil {
			post.Pending = meta.Pending
			post.Format = meta.Format
			post.Context = meta.Context
			post.Tags = meta.Tags
			post.Date, _ = time.Parse(dbtimeformat, dt)
			m := re_linkhost.FindStringSubmatch(post.Link)
			if len(m) > 1 {
				post.Site = strings.ToLower(m[1])
			}
		}
	}
	if err != nil {
		elog.Printf("error scanning post: %s", err)
		return nil
	}
	return post
}

func scanposts(rows *sql.Rows, err error, pending bool) []*Post {
	var posts []*Post
	if err != nil {
		elog.Printf("error getting posts: %s", err)
		return posts
	}
	defer rows.Close()
	for rows.Next() {
		post := scanpost(rows)
		if post == nil {
			continue
		}
		if !pending && post.Pending {
			continue
		}
		posts = append(posts, post)
	}
	return posts

}

func getpostsforgroup(group *Group, pending bool) []*Post {
	dbQuery("getpostsforgroup %d", group.ID)
	rows, err := stmtGetPostsForGroup.Query(group.ID)
	return scanposts(rows, err, pending)
}

func getfeaturedposts(group *Group) []*Post {
	var featured []*Post
	for _, f := range group.Featured {
		post := findpost(f)
		if post != nil {
			featured = append(featured, post)
		}
	}
	return featured
}

func getpostsforfront() []*Post {
	dbQuery("getpostsforfront")
	rows, err := stmtGetPostsForFront.Query()
	return scanposts(rows, err, false)
}

func getpostsforuser(userid int64) []*Post {
	dbQuery("getpostsforuser %d", userid)
	rows, err := stmtGetPostsForUser.Query(userid, userid)
	return scanposts(rows, err, false)
}

func getpostsbyperson(personid int64) []*Post {
	dbQuery("getpostsbyperson %d", personid)
	rows, err := stmtGetPostsByPerson.Query(personid)
	return scanposts(rows, err, true)
}

func searchposts(query string) []*Post {
	dbQuery("searchposts")
	terms := strings.Split(query, " ")
	var queries []string
	var params []interface{}
	for _, t := range terms {
		if t == "" {
			continue
		}
		if strings.HasPrefix(t, "title:") {
			title := t[6:]
			queries = append(queries, "title like ?")
			params = append(params, "%"+title+"%")
			continue
		}
		if strings.HasPrefix(t, "site:") {
			site := t[5:]
			queries = append(queries, "lnk like ?")
			params = append(params, "%"+site+"%")
			continue
		}
		t = "%" + t + "%"
		queries = append(queries, "(title like ? or content like ?)")
		params = append(params, t, t)
	}
	where := "where " + strings.Join(queries, " and ")
	sql := strings.ReplaceAll(sqlGetPosts, "WHERE", where)
	rows, err := opendatabase().Query(sql, params...)
	return scanposts(rows, err, false)
}

var postInvalidator gencache.Invalidator[string]
var postCache = gencache.New(gencache.Options[string, *Post]{
	Fill: func(ref string) (*Post, bool) {
		post := dbFindPost(ref)
		return post, true
	},
	Invalidator: &postInvalidator, Limit: 600})

func findpost(ref string) *Post {
	post, _ := postCache.Get(ref)
	return post
}

func dbFindPost(ref string) *Post {
	dbQuery("findpost %s", ref)
	var post *Post
	if strings.HasPrefix(ref, "https://") {
		row := stmtFindPostByAPid.QueryRow(ref)
		post = scanpost(row)
	} else {
		row := stmtFindPost.QueryRow(ref)
		post = scanpost(row)
	}
	if post != nil {
		reformatpost(post)
		post.Attachments = getattachments(post.ID)
	}
	return post
}

func dbFindPostForLink(groupid int64, link string) string {
	dbQuery("findpostforlink %d %s", groupid, link)
	row := stmtFindPostForLink.QueryRow(groupid, link)
	var xid string
	err := row.Scan(&xid)
	if err != nil && err != sql.ErrNoRows {
		elog.Printf("error scanning xid for link: %s", err)
	}
	return xid
}

type commentMeta struct {
	Format string   `json:",omitempty"`
	Tags   []string `json:",omitempty"`
}

func dbDeleteComment(comment *Comment) error {
	db := opendatabase()
	tx, err := db.Begin()
	if err != nil {
		return err
	}
	try := func(s string, args ...interface{}) {
		if err == nil {
			_, err = tx.Exec(s, args...)
		}
	}
	try("update comments set cpersonid = -1, capid = '.', subject = '', ccontent = '.' where commentid = ?", comment.ID)
	try("insert into tombstones (what) values (?)", comment.APid)
	if err != nil {
		tx.Rollback()
	} else {
		err = tx.Commit()
	}
	commentInvalidator.Clear(comment.Xid)
	commentInvalidator.Clear(comment.APid)
	return err
}

func dbSaveComment(comment *Comment) error {
	if comment.Xid == "" {
		comment.Xid = generateXid()
	}
	if comment.APid == "" {
		comment.APid = fmt.Sprintf("https://%s/c/%s", serverName, comment.Xid)
	}
	meta := commentMeta{
		Format: comment.Format,
		Tags:   comment.Tags,
	}
	j, err := jsonify(&meta)
	if err == nil {
		dt := comment.Date.UTC().Format(dbtimeformat)
		if comment.ID == 0 {
			var res sql.Result
			res, err = stmtSaveComment.Exec(comment.PostID, comment.PersonID, comment.ParentID, comment.Xid, comment.APid, dt, comment.Subject, comment.Content, j)
			if err == nil {
				comment.ID, _ = res.LastInsertId()
			}
		} else {
			_, err = stmtUpdateComment.Exec(comment.PostID, comment.PersonID, comment.ParentID, comment.Xid, comment.APid, dt, comment.Subject, comment.Content, j, comment.ID)
		}
	}
	commentInvalidator.Clear(comment.Xid)
	commentInvalidator.Clear(comment.APid)
	return err
}

func scancomment(row RowLike) *Comment {
	comment := new(Comment)
	var dt, j string
	err := row.Scan(&comment.ID, &comment.PostID, &comment.PersonID, &comment.ParentID, &comment.Xid, &comment.APid, &dt, &comment.Subject, &comment.Content, &j, &comment.PersonName, &comment.PersonDisplay, &comment.PostXID)
	if err == sql.ErrNoRows {
		return nil
	}
	if err == nil {
		var meta commentMeta
		err = unjsonify(j, &meta)
		if err == nil {
			comment.Format = meta.Format
			comment.Tags = meta.Tags
			comment.Date, _ = time.Parse(dbtimeformat, dt)
		}
	}
	if err != nil {
		elog.Printf("error scanning comment: %s", err)
		return nil
	}
	return comment
}

func getcomment(commentid int64) *Comment {
	dbQuery("getcomment %d", commentid)
	row := stmtGetComment.QueryRow(commentid)
	return scancomment(row)
}

var commentInvalidator gencache.Invalidator[string]
var commentCache = gencache.New(gencache.Options[string, *Comment]{
	Fill: func(ref string) (*Comment, bool) {
		comment := dbFindComment(ref)
		return comment, true
	},
	Invalidator: &commentInvalidator, Limit: 600})

func findcomment(ref string) *Comment {
	comment, _ := commentCache.Get(ref)
	return comment
}

func dbFindComment(ref string) *Comment {
	dbQuery("findcomment %s", ref)
	var comment *Comment
	if strings.HasPrefix(ref, "https://") {
		row := stmtFindCommentByAPid.QueryRow(ref)
		comment = scancomment(row)
	} else {
		row := stmtFindComment.QueryRow(ref)
		comment = scancomment(row)
	}
	if comment != nil {
		reformatcomment(comment)
	}
	return comment
}

func getcomments(postid int64) []*Comment {
	dbQuery("getcomments %d", postid)
	var comments []*Comment
	rows, err := stmtGetComments.Query(postid)
	if err != nil {
		elog.Printf("error getting comments: %s", err)
		return comments
	}
	defer rows.Close()
	for rows.Next() {
		comment := scancomment(rows)
		if comment == nil {
			continue
		}
		comments = append(comments, comment)
	}
	return comments
}

func threadcomments(comments []*Comment) []*Comment {
	var thread []*Comment
	idmap := make(map[int64]*Comment)
	for _, c := range comments {
		idmap[c.ID] = c
	}
	for _, c := range comments {
		arr := &thread
		if p := idmap[c.ParentID]; p != nil {
			arr = &p.Kids
		}
		if l := len(*arr); l > 0 {
			(*arr)[l-1].Next = c
		}
		*arr = append(*arr, c)
	}
	return thread
}

func dbSaveChat(chat *Chat) error {
	if chat.Xid == "" {
		chat.Xid = generateXid()
	}
	if chat.APid == "" {
		chat.APid = fmt.Sprintf("https://%s%s", serverName, chat.Rel())
	}
	dt := chat.Date.UTC().Format(dbtimeformat)
	res, err := stmtSaveChat.Exec(chat.GroupID, chat.PersonID, chat.Xid, chat.APid, dt, chat.Content)
	if err == nil {
		chat.ID, _ = res.LastInsertId()
	}
	return err
}

func scanchat(row RowLike) *Chat {
	chat := new(Chat)
	var dt string
	err := row.Scan(&chat.ID, &chat.GroupID, &chat.PersonID, &chat.Xid, &chat.APid, &dt, &chat.Content, &chat.PersonName, &chat.GroupName)
	if err == sql.ErrNoRows {
		return nil
	}
	if err == nil {
		chat.Date, _ = time.Parse(dbtimeformat, dt)
	}
	if err != nil {
		elog.Printf("error scanning chat: %s", err)
		return nil
	}
	return chat
}

func findchat(ref string) *Chat {
	dbQuery("findchat %s", ref)
	row := stmtFindChat.QueryRow(ref)
	chat := scanchat(row)
	if chat != nil {
		reformatchat(chat)
	}
	return chat
}

func getchats(groupid int64) []*Chat {
	dbQuery("getchats %d", groupid)
	var chats []*Chat
	rows, err := stmtGetChats.Query(groupid)
	if err != nil {
		elog.Printf("error getting chats: %s", err)
		return chats
	}
	defer rows.Close()
	for rows.Next() {
		chat := scanchat(rows)
		if chat == nil {
			continue
		}
		chats = append(chats, chat)
	}
	return chats
}

func dbSaveSub(userid, groupid, personid, postid int64) error {
	_, err := stmtSaveSub.Exec(userid, groupid, personid, postid)
	usersubsCache.Clear(userid)
	return err
}

func dbDeleteSub(userid, groupid, personid, postid int64) error {
	_, err := stmtDeleteSub.Exec(userid, groupid, personid, postid)
	usersubsCache.Clear(userid)
	return err
}

type UserSubs struct {
	GroupSubs map[int64]bool
}

func (subs *UserSubs) HasGroup(groupid int64) bool {
	return subs.GroupSubs[groupid]
}

var usersubsCache = gencache.New(gencache.Options[int64, *UserSubs]{
	Fill: func(userid int64) (*UserSubs, bool) {
		subs := new(UserSubs)
		subs.GroupSubs = make(map[int64]bool)
		dbQuery("getsubsforuser %d", userid)
		rows, err := stmtGetSubsForUser.Query(userid)
		if err != nil {
			elog.Printf("error querying user subs: %s", err)
			return subs, true
		}
		defer rows.Close()
		for rows.Next() {
			var groupid, personid, postid int64
			err = rows.Scan(&groupid, &personid, &postid)
			if err != nil {
				elog.Printf("error scanning sub: %s", err)
				continue
			}
			if groupid != 0 {
				subs.GroupSubs[groupid] = true
			}
		}
		return subs, true
	}, Limit: 1000})

func (group *Group) IsSubbed(userid int64) bool {
	subs, ok := usersubsCache.Get(userid)
	return ok && subs.GroupSubs[group.ID]
}

func findsub(userid, groupid, personid, postid int64) bool {
	if userid == 0 {
		return false
	}
	dbQuery("findsub %d %d %d %d", userid, groupid, personid, postid)
	row := stmtFindSub.QueryRow(userid, groupid, personid, postid)
	var subid int64
	err := row.Scan(&subid)
	if err != nil {
		return false
	}
	return true
}

func scansubs(rows *sql.Rows, err error) []int64 {
	var subs []int64
	if err != nil {
		elog.Printf("error querying subs: %s", err)
		return subs
	}
	defer rows.Close()
	for rows.Next() {
		var sub int64
		err = rows.Scan(&sub)
		if err != nil {
			elog.Printf("error scanning sub: %s", err)
			continue
		}
		subs = append(subs, sub)
	}
	return subs
}

func (group *Group) Subs() []int64 {
	dbQuery("getgroupsubs %d", group.ID)
	rows, err := stmtGetSubsToGroup.Query(group.ID)
	return scansubs(rows, err)
}

func (person *Person) IsSubbed(userid int64) bool {
	return findsub(userid, 0, person.ID, 0)
}

func (post *Post) IsSubbed(userid int64) bool {
	return findsub(userid, 0, 0, post.ID)
}

func (post *Post) Subs() []int64 {
	dbQuery("getpostsubs %d", post.ID)
	rows, err := stmtGetSubsToPost.Query(post.ID)
	return scansubs(rows, err)
}

func (post *Post) IsShared(personid int64) bool {
	dbQuery("postshared %d %d", personid, post.ID)
	row := stmtFindShare.QueryRow(personid, post.ID)
	var shareid int64
	err := row.Scan(&shareid)
	if err != nil {
		return false
	}
	return true
}

func dbDeletePostAttachments(postid int64) error {
	_, err := stmtDeletePostAttachments.Exec(postid)
	return err
}

func dbSaveAttachment(att *Attachment) error {
	var err error
	if att.ID == 0 {
		var res sql.Result
		res, err = stmtSaveAttachment.Exec(att.PostID, att.Xid, att.Media, att.Desc)
		if err == nil {
			att.ID, _ = res.LastInsertId()
		}
	} else {
		_, err = stmtUpdateAttachment.Exec(att.Desc, att.ID)
	}
	return err
}

func scanattachment(row RowLike) *Attachment {
	att := new(Attachment)
	err := row.Scan(&att.ID, &att.PostID, &att.Xid, &att.Media, &att.Desc)
	if err == sql.ErrNoRows {
		return nil
	}
	if err != nil {
		elog.Printf("error scanning att: %s", err)
		return nil
	}
	return att
}

func getattachments(postid int64) []*Attachment {
	dbQuery("getattachments %d", postid)
	var atts []*Attachment
	rows, err := stmtGetAttachments.Query(postid)
	if err != nil {
		elog.Printf("error getting atts: %s", err)
		return atts
	}
	defer rows.Close()
	for rows.Next() {
		att := scanattachment(rows)
		if att == nil {
			continue
		}
		atts = append(atts, att)
	}
	return atts
}

func getmanyattachments(posts []*Post) {
	m := make(map[int64]*Post)
	var ids []string
	for _, post := range posts {
		ids = append(ids, strconv.FormatInt(post.ID, 10))
		m[post.ID] = post
	}
	q := fmt.Sprintf("%s where postid in (%s)", sqlGetAttachments, strings.Join(ids, ","))
	db := opendatabase()
	rows, err := db.Query(q)
	if err != nil {
		elog.Printf("error getting atts: %s", err)
		return
	}
	defer rows.Close()
	for rows.Next() {
		att := scanattachment(rows)
		if att == nil {
			continue
		}
		post := m[att.PostID]
		post.Attachments = append(post.Attachments, att)
	}
}

func dbSaveSeckey(owner, seckey string) error {
	_, err := stmtSaveSeckey.Exec(owner, seckey)
	return err
}

func dbSavePubkey(url, owner, pubkey string) error {
	dt := time.Now().UTC().Format(dbtimeformat)
	_, err := stmtSavePubkey.Exec(url, owner, pubkey, dt)
	return err
}

func dbFindPubkey(url string) string {
	dbQuery("findpubkey %s", url)
	row := stmtFindPubkey.QueryRow(url)
	var pubkey string
	err := row.Scan(&pubkey)
	if err != nil && err != sql.ErrNoRows {
		elog.Printf("error scanning pubkey: %s", err)
	}
	return pubkey
}

func dbSaveShare(personid, postid int64) error {
	_, err := stmtSaveShare.Exec(personid, postid)
	return err
}

func dbDeleteShare(personid, postid int64) error {
	_, err := stmtDeleteShare.Exec(personid, postid)
	return err
}

func dbSaveFollow(folxid, who, what string) error {
	_, err := stmtSaveFollow.Exec(folxid, who, what)
	return err
}

func dbDeleteFollow(who, what string) error {
	_, err := stmtDeleteFollow.Exec(who, what)
	return err
}

func dbClearFollows(rcpt string) {
	instance := originate(rcpt)
	db := opendatabase()
	q := "%" + instance + "%"
	rows, err := db.Query("select who from follows where who like ?", q)
	if err != nil {
		elog.Printf("error finding follows to clear: %s", err)
		return
	}
	defer rows.Close()
	var toclear []string
	for rows.Next() {
		var who string
		err = rows.Scan(&who)
		if err != nil {
			elog.Printf("error scanning follow to clear: %s", err)
			continue
		}
		if instance == originate(who) {
			toclear = append(toclear, who)
		}
	}
	rows.Close()
	for _, who := range toclear {
		_, err = db.Exec("delete from follows where who = ?", who)
		if err != nil {
			elog.Printf("error clearing follow: %s", err)
		}
	}
}

func findfollow(who, what string) string {
	dbQuery("findfollow %s %s", who, what)
	row := stmtFindFollow.QueryRow(who, what)
	var folxid string
	err := row.Scan(&folxid)
	if err != nil && err != sql.ErrNoRows {
		elog.Printf("error scanning follow: %s", err)
	}
	return folxid
}

func findfollowers(what string) []string {
	dbQuery("findfollowers %s", what)
	var whos []string
	rows, err := stmtFindFollowers.Query(what)
	if err != nil {
		elog.Printf("error getting followers: %s", err)
		return whos
	}
	defer rows.Close()
	for rows.Next() {
		var who string
		err = rows.Scan(&who)
		if err != nil {
			elog.Printf("error scanning follower: %s", err)
			continue
		}
		whos = append(whos, who)
	}
	return whos
}

func findfollowing(who string) []string {
	dbQuery("findfollowing %s", who)
	var whats []string
	rows, err := stmtFindFollowing.Query(who)
	if err != nil {
		elog.Printf("error getting following: %s", err)
		return whats
	}
	defer rows.Close()
	for rows.Next() {
		var what string
		err = rows.Scan(&what)
		if err != nil {
			elog.Printf("error scanning follow: %s", err)
			continue
		}
		whats = append(whats, what)
	}
	return whats
}

type reportMeta struct {
	Link  string `json:",omitempty"`
	Why   string `json:",omitempty"`
	Notes string `json:",omitempty"`
	Blame string `json:",omitempty"`
}

func dbSaveReport(report *Report) error {
	meta := reportMeta{
		Link:  report.Link,
		Why:   report.Why,
		Notes: report.Notes,
		Blame: report.Blame,
	}
	j, err := jsonify(&meta)
	if err == nil {
		dt := report.Date.UTC().Format(dbtimeformat)
		if report.ID == 0 {
			var res sql.Result
			res, err = stmtSaveReport.Exec(report.Who, report.What, report.Where, dt, j)
			if err == nil {
				report.ID, _ = res.LastInsertId()
			}
		} else {
			_, err = stmtUpdateReport.Exec(report.Who, report.What, report.Where, dt, j, report.ID)
		}
	}
	return err
}

func scanreport(row RowLike) *Report {
	report := new(Report)
	var dt, j string
	err := row.Scan(&report.ID, &report.Who, &report.What, &report.Where, &dt, &j)
	if err == sql.ErrNoRows {
		return nil
	}
	if err == nil {
		var meta reportMeta
		err = unjsonify(j, &meta)
		if err == nil {
			report.Link = meta.Link
			report.Why = meta.Why
			report.Notes = meta.Notes
			report.Blame = meta.Blame
			report.Date, _ = time.Parse(dbtimeformat, dt)
		}
	}
	if err != nil {
		elog.Printf("error scanning report: %s", err)
		return nil
	}
	return report
}

func getreports() []*Report {
	dbQuery("getreports")
	rows, err := stmtGetReports.Query()
	return scanreports(rows, err)
}

func getreportsbywhere(where string) []*Report {
	dbQuery("getreportswhere %s", where)
	rows, err := stmtGetReportsByWhere.Query(where)
	return scanreports(rows, err)
}

func scanreports(rows *sql.Rows, err error) []*Report {
	var reports []*Report
	if err != nil {
		elog.Printf("error getting reports: %s", err)
		return reports
	}
	defer rows.Close()
	for rows.Next() {
		report := scanreport(rows)
		if report == nil {
			continue
		}
		reports = append(reports, report)
	}
	return reports
}
