//
// Copyright (c) 2023 Ted Unangst <tedu@tedunangst.com>
//
// Permission to use, copy, modify, and distribute this software for any
// purpose with or without fee is hereby granted, provided that the above
// copyright notice and this permission notice appear in all copies.
//
// THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
// WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
// ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
// WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
// ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
// OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

package main

import (
	"bytes"
	"fmt"
	"io"
	"net/http"
	"net/url"
	"strings"
	"time"

	"github.com/gorilla/mux"
	"humungus.tedunangst.com/r/azorius/slices"
	"humungus.tedunangst.com/r/webs/gencache"
	"humungus.tedunangst.com/r/webs/httpsig"
	"humungus.tedunangst.com/r/webs/junk"
)

func webinbox(w http.ResponseWriter, r *http.Request) {
	if !enableFedi {
		http.NotFound(w, r)
		return
	}
	if isBannedRequest(r) {
		http.NotFound(w, r)
		return
	}

	if name := mux.Vars(r)["name"]; name != "" {
		user := finduser(name)
		if user == nil {
			http.NotFound(w, r)
			return
		}
	}
	if xid := mux.Vars(r)["xid"]; xid != "" {
		group := findgroup(xid)
		if group == nil || group.IsRemote() {
			http.NotFound(w, r)
			return
		}
	}

	var buf bytes.Buffer
	limiter := io.LimitReader(r.Body, 1*1024*1024)
	io.Copy(&buf, limiter)
	payload := buf.Bytes()
	j, err := junk.FromBytes(payload)
	if err != nil {
		ilog.Printf("bad payload: %s", err)
		return
	}

	if ignoreJunk(j) {
		return
	}
	keyname, err := httpsig.VerifyRequest(r, payload, findpubkey)
	if err != nil {
		if shouldReverify(keyname) {
			keyname, err = httpsig.VerifyRequest(r, payload, findpubkey)
		}
	}
	if err != nil {
		ilog.Printf("signature failed: %s", err)
		w.WriteHeader(http.StatusUnauthorized)
		io.WriteString(w, "valid signature required\n")
		return
	}
	origin := Origin{Domain: originate(keyname), Json: j}
	id, _ := j.GetString("id")
	actor, _ := j.GetString("actor")
	if !origin.Match(id) || !origin.Match(actor) {
		dlog.Printf("impersonated object %s from %s", id, origin.Domain)
		return
	}
	if isBannedPerson(actor) {
		dlog.Printf("rejected object %s from banned actor %s", id, actor)
		return
	}
	typ := getType(j)
	dlog.Printf("got an inbox: %s to %s from %s", typ, r.URL.Path, actor)
	switch typ {
	case "Announce":
		origin.Group = findgroup(actor)
		apHandleAnnounce(&origin, j, false)
	case "Create":
		apHandleCreate(&origin, j, false)
	case "Follow":
		apHandleFollow(j)
	case "Accept":
		apHandleAccept(j)
	case "Reject":
		apHandleReject(j)
	case "Update":
		apHandleUpdate(&origin, j)
	case "Undo":
		apHandleUndo(origin.Domain, j)
	case "Delete":
		apHandleDelete(&origin, j)
	case "Flag":
		apHandleFlag(j)
	case "ChatMessage":
		apHandleChatMessage(&origin, j)
	case "Ping":
		to, _ := j.GetString("to")
		apHandlePing(actor, to, id)
	case "Read":
	default:
		dlog.Printf("unknown activity in inbox: %s %s", typ, j.ToString())
	}
}

var outboxCache = gencache.New(gencache.Options[string, []byte]{
	Fill: func(apid string) ([]byte, bool) {
		var posts []*Post
		isGroup := false
		if person := findperson(apid); person != nil && person.IsLocal() {
			posts = getpostsbyperson(person.ID)
		} else if group := findgroup(apid); group != nil && group.IsLocal() {
			posts = getpostsforgroup(group, false)
			isGroup = true
		} else {
			return nil, false
		}
		reformatposts(posts)
		getmanyattachments(posts)
		var items []junk.Junk
		for _, post := range posts {
			person := findperson(post.PersonName)
			group := findgroup(post.GroupName)
			jp := jsonPost(post)
			jc := jsonCreate(person.APid, group.APid, jp)
			if isGroup {
				ja := jsonAnnounce(group.APid, group.APid, jc)
				items = append(items, ja)
			} else {
				items = append(items, jc)
			}
		}
		if len(items) == 0 {
			items = make([]junk.Junk, 0)
		}
		if len(items) > 20 {
			items = items[:20]
		}

		j := newActivity("OrderedCollection", apid+"/outbox")
		j["attributedTo"] = apid
		j["totalItems"] = len(items)
		j["orderedItems"] = items
		return j.ToBytes(), true
	},
	Invalidators: []*gencache.Invalidator[string]{&personInvalidator, &groupInvalidator},
	Duration:     60 * time.Second,
	Limit:        20})

func weboutbox(w http.ResponseWriter, r *http.Request) {
	if !enableFedi {
		http.NotFound(w, r)
		return
	}
	if isBannedRequest(r) {
		http.NotFound(w, r)
		return
	}
	apid := "https://" + serverName + r.URL.String()
	apid = strings.TrimSuffix(apid, "/outbox")

	j, _ := outboxCache.Get(apid)
	if len(j) == 0 {
		http.NotFound(w, r)
		return
	}
	w.Header().Set("Content-Type", apContentType)
	w.Write(j)
}

func webfeatured(w http.ResponseWriter, r *http.Request) {
	if !enableFedi {
		http.NotFound(w, r)
		return
	}
	if isBannedRequest(r) {
		http.NotFound(w, r)
		return
	}

	xid := mux.Vars(r)["xid"]
	group := findgroup(xid)
	if group == nil || group.IsRemote() {
		http.NotFound(w, r)
		return
	}
	apid := group.APid
	posts := getfeaturedposts(group)
	// reformatposts and getattachments already done
	var items []junk.Junk
	for _, post := range posts {
		jp := jsonPost(post)
		items = append(items, jp)
	}
	if len(items) == 0 {
		items = make([]junk.Junk, 0)
	}
	if len(items) > 20 {
		items = items[:20]
	}

	j := newActivity("OrderedCollection", apid+"/featured")
	j["attributedTo"] = apid
	j["totalItems"] = len(items)
	j["orderedItems"] = items

	w.Header().Set("Content-Type", apContentType)
	j.Write(w)
}

func webpostreplies(w http.ResponseWriter, r *http.Request) {
	if !enableFedi {
		http.NotFound(w, r)
		return
	}
	if isBannedRequest(r) {
		http.NotFound(w, r)
		return
	}

	xid := mux.Vars(r)["xid"]
	post := findpost(xid)
	if post == nil {
		http.NotFound(w, r)
		return
	}
	if post.IsRemote() {
		http.NotFound(w, r)
		return
	}
	comments := getcomments(post.ID)
	var links []string
	for _, comment := range comments {
		if comment.APid == "." {
			continue
		}
		links = append(links, comment.APid)
	}
	slices.Reverse(links)
	j := newActivity("OrderedCollection", post.APid+"/replies")
	j["attributedTo"] = post.APid
	j["totalItems"] = len(links)
	j["orderedItems"] = links
	w.Header().Set("Content-Type", apContentType)
	j.Write(w)
}

func apHandleUndo(origin string, j junk.Junk) {
	actor, _ := j.GetString("actor")
	obj, _ := j.GetMap("object")
	if len(obj) == 0 {
		dlog.Printf("not sure what to undo: %s", j.ToString())
		return
	}
	typ := getType(obj)
	switch typ {
	case "Announce":
		target, _ := obj.GetString("object")
		if target != "" {
			post := findpost(target)
			person := findperson(actor)
			if post != nil && person != nil && post.IsShared(person.ID) {
				err := dbDeleteShare(person.ID, post.ID)
				if err != nil {
					elog.Printf("error deleting share: %s", err)
				}
			}
			return
		}
		dlog.Printf("not ready to undo that announce yet: %s", obj.ToString())
	case "Follow":
		target, _ := obj.GetString("object")
		if target != "" {
			err := dbDeleteFollow(actor, target)
			if err != nil {
				elog.Printf("error deleting follow: %s", err)
			}
			return
		}
		dlog.Printf("can't undo follow: %s", obj.ToString())
	default:
		dlog.Printf("can't undo yet: %s", obj.ToString())
	}
}

func apHandleFollow(j junk.Junk) {
	folxid, _ := j.GetString("id")
	who, _ := j.GetString("actor")
	what, _ := j.GetString("object")
	if !isMine(what) {
		return
	}
	group := findgroup(what)
	user := findperson(what)
	if group == nil && user == nil {
		return
	}
	err := dbSaveFollow(folxid, who, what)
	if err != nil {
		elog.Printf("error saving follow: %s", err)
		return
	}
	apSendAccept(folxid, who, what)
}

func apSendAccept(folxid string, who, what string) {
	j := newActivity("Accept", what+"/accept/"+generateXid())
	j["actor"] = what
	j["to"] = who
	j["object"] = jsonFollow(folxid, who, what)

	go deliver(what, who, j.ToBytes())
}

func jsonFollow(folxid, who, what string) junk.Junk {
	j := newActivity("Follow", folxid)
	j["actor"] = who
	j["to"] = what
	j["object"] = what
	return j
}

func apSendFollow(folxid, who, what string) {
	j := jsonFollow(folxid, who, what)
	go deliver(who, what, j.ToBytes())
}

func jsonUndo(who string, what junk.Junk) junk.Junk {
	j := newActivity("Undo", who+"/undo/"+generateXid())
	j["actor"] = who
	j["to"] = apPublic
	j["object"] = what
	j["published"] = time.Now().UTC().Format(time.RFC3339)
	return j
}

func apSendUnfollow(who, what string) {
	folxid := findfollow(who, what)
	jf := jsonFollow(folxid, who, what)
	ju := jsonUndo(who, jf)
	go deliver(who, what, ju.ToBytes())
}

func apHandleAccept(j junk.Junk) {

}

func apHandleReject(j junk.Junk) {

}

func findInbox(rcpt string, shared bool) string {
	if person := findperson(rcpt); person != nil {
		if shared && person.ShareBox != "" {
			return person.ShareBox
		}
		return person.Inbox
	}
	if group := findgroup(rcpt); group != nil {
		if shared && group.ShareBox != "" {
			return group.ShareBox
		}
		return group.Inbox
	}
	if person := personForAPid(rcpt); person != nil {
		if shared && person.ShareBox != "" {
			return person.ShareBox
		}
		return person.Inbox
	}
	return ""
}

func collectBoxes(rcpts []string) map[string]bool {
	boxes := make(map[string]bool)
	for _, rcpt := range rcpts {
		if rcpt == apPublic {
			continue
		}
		if isMine(rcpt) {
			continue
		}
		inbox := findInbox(rcpt, true)
		if inbox != "" {
			boxes["%"+inbox] = true
			continue
		}
		boxes[rcpt] = true
	}
	return boxes
}

func apSendGroupUpdate(group *Group, changed bool, oldfeatured []string) {
	if changed {
		jp := jsonGroup(group)
		ju := jsonUpdate(group.APid, group.APid, jp)
		apSendGroupAnnounce(group, ju)
	}
	for _, o := range oldfeatured {
		if !slices.Contains(group.Featured, o) {
			apSendRemoveFeatured(group, o)
		}
	}
	for _, f := range group.Featured {
		if !slices.Contains(oldfeatured, f) {
			apSendAddFeatured(group, f)
		}
	}
}

func apSendAddFeatured(group *Group, id string) {
	j := jsonAddFeatured(group.APid, group.APid, id)
	apSendGroupAnnounce(group, j)
}

func apSendRemoveFeatured(group *Group, id string) {
	j := jsonRemoveFeatured(group.APid, group.APid, id)
	apSendGroupAnnounce(group, j)
}

func apSendPostUpdate(person *Person, post *Post) {
	group := findgroup(post.GroupName)
	jp := jsonPost(post)
	ju := jsonUpdate(person.APid, group.APid, jp)
	apBroadcast(person, ju, group)
}

func apSendPost(person *Person, post *Post) {
	group := findgroup(post.GroupName)
	jp := jsonPost(post)
	jc := jsonCreate(person.APid, group.APid, jp)
	apBroadcast(person, jc, group)
}

func apSendDeletePost(person string, post *Post) {
	group := findgroup(post.GroupName)
	jd := jsonDelete(person, group.APid, post.APid)
	apBroadcastAny(person, jd, group)
}

func apSendDeletePerson(person string) {
	jd := jsonDelete(person, "", person)
	whos := findfollowers(person)
	whats := findfollowing(person)
	rcpts := append(whos, whats...)
	apSendToRcpts(person, rcpts, jd.ToBytes())
}

func apSendChat(person *Person, group *Group, chat *Chat) {
	jm := jsonChat(chat)
	jc := jsonCreate(person.APid, group.APid, jm)
	if group.IsLocal() {
		apSendGroupAnnounce(group, jc)
	} else {
		go deliver(person.APid, group.APid, jc.ToBytes())
	}
}

func apSendComment(person *Person, group *Group, comment *Comment) {
	jcom := jsonComment(comment)
	jc := jsonCreate(person.APid, group.APid, jcom)
	apBroadcast(person, jc, group)
}

func apSendCommentUpdate(person *Person, group *Group, comment *Comment) {
	jcom := jsonComment(comment)
	ju := jsonUpdate(person.APid, group.APid, jcom)
	apBroadcast(person, ju, group)
}

func apSendCommentDelete(person string, group *Group, comment *Comment) {
	jd := jsonDelete(person, group.APid, comment.APid)
	apBroadcastAny(person, jd, group)
}

func apAnnouncePost(person *Person, post *Post) {
	ja := jsonAnnounce(person.APid, "", post.APid)
	apSendMsgByPerson(person.APid, nil, ja.ToBytes())
}

func apBroadcast(person *Person, pmsg junk.Junk, group *Group) {
	apBroadcastAny(person.APid, pmsg, group)
}

func apCollectCC(obj junk.Junk) []string {
	if j, ok := obj.GetMap("object"); ok {
		obj = j
	}
	return collectAddrs(nil, obj, "cc")
}

func apBroadcastAny(person string, pmsg junk.Junk, group *Group) {
	if isMine(person) {
		cc := apCollectCC(pmsg)
		apSendMsgByPerson(person, cc, pmsg.ToBytes())
	}
	if group.IsLocal() {
		apSendGroupAnnounce(group, pmsg)
	}
}

func apSendMsgByPerson(person string, cc []string, msg []byte) {
	rcpts := findfollowers(person)
	rcpts = append(rcpts, cc...)
	apSendToRcpts(person, rcpts, msg)
}

func apSendGroupAnnounce(group *Group, j junk.Junk) {
	rcpts := findfollowers(group.APid)
	rcpts = append(rcpts, group.APid)
	ja := jsonAnnounce(group.APid, group.APid, j)
	apSendToRcpts(group.APid, rcpts, ja.ToBytes())
}

func apSendToRcpts(from string, rcpts []string, msg []byte) {
	boxes := collectBoxes(rcpts)
	for inbox := range boxes {
		go deliver(from, inbox, msg)
	}
}

var fingerCache = gencache.New(gencache.Options[string, []byte]{
	Fill: func(orig string) ([]byte, bool) {
		if strings.HasPrefix(orig, "acct:") {
			orig = orig[5:]
		}
		name := orig
		idx := strings.LastIndexByte(name, '/')
		if idx != -1 {
			name = name[idx+1:]
			if fmt.Sprintf("https://%s/u/%s", serverName, name) != orig {
				ilog.Printf("foreign request rejected")
				name = ""
			}
		} else {
			idx = strings.IndexByte(name, '@')
			if idx != -1 {
				name = name[:idx]
				if !(name+"@"+serverName == orig) {
					ilog.Printf("foreign request rejected")
					name = ""
				}
			}
		}
		var apid string
		if user := finduser(name); user != nil {
			apid = user.Person.APid
		} else if group := findgroup(name); group != nil && group.IsLocal() {
			apid = group.APid
		} else {
			return nil, false
		}

		j := junk.New()
		j["subject"] = fmt.Sprintf("acct:%s@%s", name, serverName)
		j["aliases"] = []string{apid}
		l := junk.New()
		l["rel"] = "self"
		l["type"] = `application/activity+json`
		l["href"] = apid
		j["links"] = []junk.Junk{l}
		return j.ToBytes(), true
	},
	Duration: 60 * time.Second,
	Limit:    60})

func webfinger(w http.ResponseWriter, r *http.Request) {
	res := r.FormValue("resource")

	j, _ := fingerCache.Get(res)
	if len(j) == 0 {
		http.NotFound(w, r)
		return
	}
	w.Header().Set("Content-Type", "application/jrd+json")
	w.Write(j)
}

func resolvefinger(handle string) string {
	if handle[0] == '@' || handle[1] == '!' {
		handle = handle[1:]
	}
	parts := strings.Split(handle, "@")
	if len(parts) != 2 {
		return ""
	}
	url := fmt.Sprintf("https://%s/.well-known/webfinger?resource=acct:%s", parts[1], handle)
	j, err := junk.Get(url, junk.GetArgs{
		Accept:  "application/jrd+json",
		Timeout: fastTimeout,
	})
	if err != nil {
		dlog.Printf("error getting web finger: %s", err)
		return ""
	}
	links, _ := j.GetArray("links")
	for _, li := range links {
		l, ok := li.(junk.Junk)
		if !ok {
			continue
		}
		href, _ := l.GetString("href")
		rel, _ := l.GetString("rel")
		t, _ := l.GetString("type")
		if rel == "self" && isAPtype(t) {
			return href
		}
	}
	return ""
}

func ignoreJunk(j junk.Junk) bool {
	t := getType(j)
	if t == "Delete" {
		a, _ := j.GetString("actor")
		o, _ := j.GetString("object")
		if a == o && dbFindPerson(a) == nil {
			return true
		}
	}
	if t == "Announce" {
		if obj, ok := j.GetMap("object"); ok {
			j = obj
			t = getType(j)
		}
	}
	if t == "Undo" {
		if obj, ok := j.GetMap("object"); ok {
			j = obj
			t = getType(j)
		}
	}
	if t == "Like" || t == "Dislike" {
		return true
	}
	return false
}

var pubkeyCache = gencache.New(gencache.Options[string, httpsig.PublicKey]{
	Fill: func(keyname string) (httpsig.PublicKey, bool) {
		data := dbFindPubkey(keyname)
		if data == "" {
			j, err := fetchActivity(keyname, fastTimeout)
			if err != nil {
				dlog.Printf("unable to fetch needed pubkey: %s", err)
				return httpsig.PublicKey{}, false
			}
			typ := getType(j)
			if typ == "Person" || typ == "Service" {
				apHandlePerson(originate(keyname), j)
			} else if typ == "Group" {
				apHandleGroup(originate(keyname), j, false)
			}
			keyobj, ok := j.GetMap("publicKey")
			if ok {
				savepubkey(originate(keyname), keyobj, "")
			}
			data = dbFindPubkey(keyname)
		}
		if data == "" {
			dlog.Printf("no key found for %s", keyname)
			return httpsig.PublicKey{}, false
		}
		_, key, err := httpsig.DecodeKey(data)
		if err != nil {
			dlog.Printf("error decoding pubkey for %s: %s", keyname, err)
			return httpsig.PublicKey{}, false
		}
		return key, true
	},
	Limit: 600})

func findpubkey(keyname string) (httpsig.PublicKey, error) {
	var err error
	key, ok := pubkeyCache.Get(keyname)
	if !ok {
		err = fmt.Errorf("unable to find pubkey: %s", keyname)
	}
	return key, err
}

func savepubkey(origin string, keyobj junk.Junk, owner string) {
	pubkey, ok := keyobj.GetString("publicKeyPem")
	keyid, _ := keyobj.GetString("id")
	u, err := url.Parse(keyid)
	if err != nil {
		dlog.Printf("error parsing keyid: %s", err)
		return
	}
	u.Path = "/" + strings.TrimLeft(u.Path, "/")
	keyid = u.String()
	if owner == "" {
		owner, _ = keyobj.GetString("owner")
	}
	if owner == "" {
		dlog.Printf("no owner for key: %s", keyid)
		return
	}
	if originate(keyid) != origin || originate(owner) != origin {
		dlog.Printf("impersonated key %s for %s from %s", keyid, owner, origin)
		ok = false
	}
	if ok {
		err := dbSavePubkey(keyid, owner, pubkey)
		if err != nil {
			elog.Printf("error saving pubkey: %s", err)
		}
	}
}

func shouldReverify(keyname string) bool {
	if isMine(keyname) {
		return false
	}
	db := opendatabase()
	when := time.Now().Add(-60 * time.Minute).UTC().Format(dbtimeformat)
	res, err := db.Exec("delete from pubkeys where url = ? and dt < ?", keyname, when)
	if err != nil {
		elog.Printf("error clearing old pubkey: %s", err)
		return false
	}
	num, _ := res.RowsAffected()
	if num > 0 {
		dlog.Printf("deleted old pubkey %s", keyname)
		pubkeyCache.Clear(keyname)
		return true
	}
	return false
}

type keyInfo struct {
	name string
	key  httpsig.PrivateKey
}

var seckeyCache = gencache.New(gencache.Options[string, keyInfo]{
	Fill: func(from string) (keyInfo, bool) {
		row := stmtFindSeckey.QueryRow(from)
		var seckey string
		err := row.Scan(&seckey)
		if err != nil {
			elog.Printf("lost seckey for %s: %s", from, err)
			return keyInfo{}, false
		}
		key, _, err := httpsig.DecodeKey(seckey)
		if err != nil {
			elog.Printf("error decoding seckey: %s", err)
			return keyInfo{}, false
		}
		return keyInfo{from + "#key", key}, true
	},
	Limit: 300})

func getseckey(from string) (string, httpsig.PrivateKey) {
	ki, _ := seckeyCache.Get(from)
	return ki.name, ki.key
}

func apHandlePing(from, to, id string) {
	j := newActivity("Pong", serverPrefix+"/pong/"+generateXid())
	j["actor"] = to
	j["to"] = from
	j["object"] = id

	go deliver(to, from, j.ToBytes())
}
