//
// Copyright (c) 2023 Ted Unangst <tedu@tedunangst.com>
//
// Permission to use, copy, modify, and distribute this software for any
// purpose with or without fee is hereby granted, provided that the above
// copyright notice and this permission notice appear in all copies.
//
// THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
// WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
// ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
// WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
// ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
// OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

package main

import (
	"html/template"
	"net/url"
	"strings"

	"golang.org/x/net/html"
	"humungus.tedunangst.com/r/webs/htfilter"
	"humungus.tedunangst.com/r/webs/mz"
	"humungus.tedunangst.com/r/webs/templates"
)

var allowedclasses = make(map[string]bool)

func init() {
	allowedclasses["kw"] = true
	allowedclasses["bi"] = true
	allowedclasses["st"] = true
	allowedclasses["nm"] = true
	allowedclasses["tp"] = true
	allowedclasses["op"] = true
	allowedclasses["cm"] = true
	allowedclasses["al"] = true
	allowedclasses["dl"] = true
}

func markdown(input string) template.HTML {
	var marker mz.Marker
	return scrubhtml(marker.Mark(input), "")
}
func markdownShort(input string) template.HTML {
	var marker mz.Marker
	marker.Short = true
	return scrubhtml(marker.Mark(input), "")
}
func markdownimgs(input string) template.HTML {
	var marker mz.Marker
	marker.AllowImages = true
	imgfn := getimgproxy(true)
	return scrubhtmlimgs(marker.Mark(input), "", imgfn)
}

func scrubhtml(input, baseurl string) template.HTML {
	return scrubhtmlimgs(input, baseurl, nil)
}
func scrubhtmlimgs(input, baseurl string, imgfn func(node *html.Node) string) template.HTML {
	var htf htfilter.Filter
	htf.SpanClasses = allowedclasses
	htf.Imager = imgfn
	if baseurl != "" {
		htf.BaseURL, _ = url.Parse(baseurl)
	}
	html, _ := htf.String(input)
	return html
}

func getimgproxy(permstore bool) func(node *html.Node) string {
	return func(node *html.Node) string {
		src := htfilter.GetAttr(node, "src")
		alt := htfilter.GetAttr(node, "alt")
		att, err := attachmentfromurl(src, permstore)
		if err != nil {
			return string(templates.Sprintf(`&lt;img alt="%s" src="<a href="%s">%s</a>"&gt;`, alt, src, src))
		}
		return string(templates.Sprintf(`<img title="%s" alt="%s" src="/file/%s">`, alt, alt, att.Xid))
	}
}

func reformatgroups(groups []*Group) {
	for _, g := range groups {
		if g.HTDesc != "" {
			panic("already formatted group")
		}
		switch g.Format {
		case "markdown":
			g.HTDesc = markdownimgs(g.Desc)
		case "html":
			imgfn := getimgproxy(false)
			g.HTDesc = scrubhtmlimgs(g.Desc, g.APid, imgfn)
		}
	}
}

func reformatgroup(group *Group) {
	reformatgroups([]*Group{group})
}

func reformatposts(posts []*Post) {
	for _, p := range posts {
		if p.HTContent != "" {
			panic("already formatted post")
		}
		switch p.Format {
		case "markdown":
			p.HTContent = markdownimgs(p.Content)
		case "html":
			imgfn := getimgproxy(false)
			p.HTContent = scrubhtmlimgs(p.Content, p.APid, imgfn)
		}
	}
}

func reformatpost(post *Post) {
	reformatposts([]*Post{post})
}

func reformatcomments(comments []*Comment) {
	for _, c := range comments {
		if c.HTContent != "" {
			panic("already formatted comment")
		}
		switch c.Format {
		case "markdown":
			c.HTContent = markdown(c.Content)
		case "html":
			c.HTContent = scrubhtml(c.Content, c.APid)
		}
	}
}

func reformatcomment(comment *Comment) {
	reformatcomments([]*Comment{comment})
}

func reformatchats(chats []*Chat) {
	for _, c := range chats {
		if c.HTContent != "" {
			panic("already formatted comment")
		}
		c.HTContent = scrubhtml(c.Content, c.APid)
	}
}

func reformatchat(chat *Chat) {
	reformatchats([]*Chat{chat})
}

func reformatperson(person *Person) {
	if person.HTAbout != "" {
		panic("already formatted person")
	}
	switch person.Format {
	case "markdown":
		person.HTAbout = markdown(person.About)
	case "html":
		person.HTAbout = scrubhtml(person.About, person.APid)
	}
}

func normalizeLink(link string, origin string) string {
	if link == "" {
		return ""
	}
	u, err := url.Parse(link)
	if err != nil {
		return ""
	}
	if u.Host == origin && strings.HasPrefix(u.Path, "/pictrs/") {
		return ""
	}
	if u.Scheme == "" {
		u.Scheme = "https"
	}
	switch u.Host {
	case "www.youtube.com":
		if strings.HasPrefix(u.Path, "/embed") {
			vid := u.Path[strings.LastIndexByte(u.Path, '/')+1:]
			u.RawQuery = "v=" + vid
			u.Path = "/watch"
		}
	case "m.youtube.com":
		u.Host = "www.youtube.com"
	case "youtu.be":
		u.Host = "www.youtube.com"
		u.RawQuery = "v=" + u.Path[1:]
		u.Path = "/watch"
	}
	return u.String()
}

func normalizeTitle(title string) string {
	if strings.ToUpper(title) == title && strings.IndexByte(title, ' ') != -1 {
		title = strings.Title(strings.ToLower(title))
	}
	return title
}
