//
// Copyright (c) 2023 Ted Unangst <tedu@tedunangst.com>
//
// Permission to use, copy, modify, and distribute this software for any
// purpose with or without fee is hereby granted, provided that the above
// copyright notice and this permission notice appear in all copies.
//
// THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
// WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
// ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
// WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
// ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
// OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

package main

import (
	"flag"
	"fmt"
	golog "log"
	"log/syslog"
	"os"
	"runtime/pprof"
	"strconv"
	"time"

	"humungus.tedunangst.com/r/webs/log"
)

var softwareVersion = "0.3.2"

var serverName string
var serverPrefix string
var dataDir = "."
var viewDir = "."
var iconName = "icon.png"
var serverMsg string
var aboutMsg string
var siteNotif string
var loginMsg string
var signupMsg string

var cpuprofile = flag.String("cpuprofile", "", "write cpu profile to file")
var memprofile = flag.String("memprofile", "", "write memory profile to this file")
var memprofilefd *os.File

func reexecArgs(cmd string) []string {
	args := []string{"-datadir", dataDir}
	args = append(args, log.Args()...)
	args = append(args, cmd)
	return args
}

var elog, ilog, dlog *golog.Logger

func ElaborateUnitTests() {
}

func errx(msg string, args ...interface{}) {
	fmt.Fprintf(os.Stderr, msg+"\n", args...)
	os.Exit(1)
}

func main() {
	flag.StringVar(&dataDir, "datadir", dataDir, "data directory")
	flag.StringVar(&viewDir, "viewdir", viewDir, "view directory")
	flag.Parse()

	if *cpuprofile != "" {
		f, err := os.Create(*cpuprofile)
		if err != nil {
			errx("can't open cpu profile: %s", err)
		}
		pprof.StartCPUProfile(f)
	}
	if *memprofile != "" {
		f, err := os.Create(*memprofile)
		if err != nil {
			errx("can't open mem profile: %s", err)
		}
		memprofilefd = f
	}

	log.Init(log.Options{Progname: "azorius", Facility: syslog.LOG_NEWS})
	elog = log.E
	ilog = log.I
	dlog = log.D

	if os.Geteuid() == 0 {
		elog.Fatalf("do not run azorius as root")
	}

	args := flag.Args()
	cmd := "run"
	if len(args) > 0 {
		cmd = args[0]
		args = args[1:]
	}
	switch cmd {
	case "init":
		initdb()
	case "upgrade":
		upgradedb()
	case "version":
		fmt.Println(softwareVersion)
		os.Exit(0)
	}
	db := opendatabase()
	dbversion := 0
	getconfig("dbversion", &dbversion)
	if dbversion != myVersion {
		elog.Fatal("incorrect database version. run upgrade.")
	}
	prepareStatements(db)

	getconfig("servername", &serverName)
	serverPrefix = fmt.Sprintf("https://%s/", serverName)
	userAgent = userAgent + serverName
	getconfig("devel", &develMode)
	if develMode {
		disableTLS()
	}
	getconfig("fasttimeout", &fastTimeout)
	fastTimeout *= time.Second
	getconfig("slowtimeout", &slowTimeout)
	slowTimeout *= time.Second
	loadAdminSettings()

	switch cmd {
	case "trash":
		filescleanup()
	case "devel":
		if len(args) != 1 {
			errx("need an argument: devel (on|off)")
		}
		switch args[0] {
		case "on":
			setconfig("devel", 1)
		case "off":
			setconfig("devel", 0)
		default:
			errx("argument must be on or off")
		}
	case "setconfig":
		if len(args) != 2 {
			errx("need an argument: setconfig key val")
		}
		var val interface{}
		var err error
		if val, err = strconv.Atoi(args[1]); err != nil {
			val = args[1]
		}
		setconfig(args[0], val)
	case "adduser":
		adduser()
	case "chpass":
		if len(args) < 1 {
			errx("usage: azorius chpass username")
		}
		chpass(args[0])
	case "export":
		if len(args) < 2 {
			errx("usage: azorius export what where")
		}
		exportdata(args[0], args[1])
	case "import":
		if len(args) < 1 {
			errx("usage: azorius import where")
		}
		importdata(args[0])
	case "run":
		webserve()
	case "backend":
		backendServer()
	case "test":
		ElaborateUnitTests()
	default:
		errx("unknown command")
	}
}
