//
// Copyright (c) 2023 Ted Unangst <tedu@tedunangst.com>
//
// Permission to use, copy, modify, and distribute this software for any
// purpose with or without fee is hereby granted, provided that the above
// copyright notice and this permission notice appear in all copies.
//
// THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
// WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
// ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
// WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
// ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
// OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

package main

import (
	"bytes"
	"context"
	"crypto/tls"
	"fmt"
	"io"
	"net/http"
	"regexp"
	"sync"
	"time"

	"humungus.tedunangst.com/r/webs/httpsig"
	"humungus.tedunangst.com/r/webs/junk"
)

var userAgent = "azorius/1.0; "

var fastTimeout time.Duration = 5
var slowTimeout time.Duration = 30
var fetchUser = "root"
var onceSetFetch sync.Once

var httpClient = http.Client{}

func disableTLS() {
	httpClient.Transport = &http.Transport{
		TLSClientConfig: &tls.Config{
			InsecureSkipVerify: true,
		},
	}
}

func fetchActivity(url string, timeout time.Duration) (junk.Junk, error) {
	if isBannedPerson(url) {
		dlog.Printf("skipping banned url: %s", url)
		return nil, fmt.Errorf("url is on the banlist")
	}
	if hastombstone(url) {
		dlog.Printf("skipping dead url: %s", url)
		return nil, fmt.Errorf("url has a tombstone")
	}
	dlog.Printf("fetching activity: %s", url)
	sign := func(req *http.Request) error {
		onceSetFetch.Do(func() {
			fetchUser = dbGetUser(1).Person.APid
		})
		keyname, key := getseckey(fetchUser)
		httpsig.SignRequest(keyname, key, req, nil)
		return nil
	}
	if develMode {
		sign = nil
	}
	client := httpClient
	client.CheckRedirect = func(req *http.Request, via []*http.Request) error {
		if len(via) >= 5 {
			return fmt.Errorf("stopped after 5 redirects")
		}
		if sign != nil {
			sign(req)
		}
		return nil
	}
	j, err := junk.Get(url, junk.GetArgs{
		Accept:  apAcceptType,
		Agent:   userAgent,
		Timeout: timeout,
		Client:  &client,
		Fixup:   sign,
		Limit:   1 * 1024 * 1024,
	})
	if err != nil && err.Error() == "http get status: 410" {
		err := dbSaveTombstone(url)
		if err != nil {
			elog.Printf("error saving tombstone for %s: %s", url, err)
		}
	}
	return j, err
}

func fetchData(url string) ([]byte, error) {
	dlog.Printf("fetching data: %s", url)
	req, err := http.NewRequest("GET", url, nil)
	if err != nil {
		ilog.Printf("error fetching %s: %s", url, err)
		return nil, err
	}
	req.Header.Set("User-Agent", userAgent)
	ctx, cancel := context.WithTimeout(context.Background(), slowTimeout)
	defer cancel()
	req = req.WithContext(ctx)
	resp, err := httpClient.Do(req)
	if err != nil {
		ilog.Printf("error fetching %s: %s", url, err)
		return nil, err
	}
	defer resp.Body.Close()
	switch resp.StatusCode {
	case 200:
	case 201:
	case 202:
	default:
		return nil, fmt.Errorf("http get not 200: %d %s", resp.StatusCode, url)
	}
	var buf bytes.Buffer
	limiter := io.LimitReader(resp.Body, 10*1024*1024)
	io.Copy(&buf, limiter)
	return buf.Bytes(), nil
}

func PostMsg(keyname string, key httpsig.PrivateKey, url string, msg []byte) error {
	dlog.Printf("posting to %s", url)
	req, err := http.NewRequest("POST", url, bytes.NewReader(msg))
	if err != nil {
		return err
	}
	req.Header.Set("User-Agent", userAgent)
	req.Header.Set("Content-Type", apContentType)
	httpsig.SignRequest(keyname, key, req, msg)
	ctx, cancel := context.WithTimeout(context.Background(), 2*slowTimeout)
	defer cancel()
	req = req.WithContext(ctx)
	resp, err := httpClient.Do(req)
	if err != nil {
		return err
	}
	resp.Body.Close()

	switch resp.StatusCode {
	case 200:
	case 201:
	case 202:
	default:
		return fmt.Errorf("http post status: %d", resp.StatusCode)
	}
	ilog.Printf("successful post: %s %d", url, resp.StatusCode)
	return nil
}

var re_keyholder = regexp.MustCompile(`keyId="([^"]+)"`)

func requestOrigin(r *http.Request) string {
	if sig := r.Header.Get("Signature"); sig != "" {
		if m := re_keyholder.FindStringSubmatch(sig); len(m) == 2 {
			who := originate(m[1])
			return who
		}
	}
	return ""
}
