//
// Copyright (c) 2023 Ted Unangst <tedu@tedunangst.com>
//
// Permission to use, copy, modify, and distribute this software for any
// purpose with or without fee is hereby granted, provided that the above
// copyright notice and this permission notice appear in all copies.
//
// THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
// WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
// ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
// WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
// ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
// OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

package main

import (
	"fmt"
	"net"
	"os"
	"os/signal"
	"runtime/debug"
	"runtime/pprof"
	"sync"
	"syscall"
	"time"
)

var develMode = false

func init() {
	if softwareVersion != "develop" {
		return
	}
	bi, ok := debug.ReadBuildInfo()
	if !ok {
		return
	}
	var vcs, rev, mod string
	for _, bs := range bi.Settings {
		if bs.Key == "vcs" {
			vcs = "/" + bs.Value
		}
		if bs.Key == "vcs.revision" {
			rev = bs.Value
			if len(rev) > 12 {
				rev = rev[:12]
			}
			rev = "-" + rev
		}
		if bs.Key == "vcs.modified" && bs.Value == "true" {
			mod = "+"
		}
	}
	softwareVersion += vcs + rev + mod
}

var endoftheworld = make(chan bool)
var readyalready = make(chan bool)
var workinprogress = 0
var requestWG sync.WaitGroup
var listenSocket net.Listener

func enditall() {
	sig := make(chan os.Signal, 1)
	signal.Notify(sig, syscall.SIGINT, syscall.SIGTERM, syscall.SIGQUIT)
	<-sig
	ilog.Printf("stopping...")
	listenSocket.Close()
	for i := 0; i < workinprogress; i++ {
		endoftheworld <- true
	}
	if *cpuprofile != "" {
		pprof.StopCPUProfile()
	}
	if *memprofile != "" {
		pprof.WriteHeapProfile(memprofilefd)
		memprofilefd.Close()
	}
	ilog.Printf("waiting...")
	go func() {
		time.Sleep(10 * time.Second)
		elog.Printf("timed out waiting for requests to finish")
		os.Exit(0)
	}()
	for i := 0; i < workinprogress; i++ {
		<-readyalready
	}
	requestWG.Wait()
	ilog.Printf("apocalypse")
	os.Exit(0)
}

func openListener() (net.Listener, error) {
	var listenAddr string
	err := getconfig("listenaddr", &listenAddr)
	if err != nil {
		return nil, err
	}
	if listenAddr == "" {
		return nil, fmt.Errorf("must have listenaddr")
	}
	proto := "tcp"
	if listenAddr[0] == '/' {
		proto = "unix"
		err := os.Remove(listenAddr)
		if err != nil && !os.IsNotExist(err) {
			elog.Printf("unable to unlink socket: %s", err)
		}
	}
	listener, err := net.Listen(proto, listenAddr)
	if err != nil {
		return nil, err
	}
	if proto == "unix" {
		os.Chmod(listenAddr, 0777)
	}
	listenSocket = listener
	return listener, nil
}

func bgmonitor() {
	for {
		time.Sleep(50 * time.Minute)
	}
}

type Sensors struct {
	Memory float64
	Uptime float64
	CPU    float64
}

var boottime = time.Now()

func getSensors() Sensors {
	var usage syscall.Rusage
	syscall.Getrusage(syscall.RUSAGE_SELF, &usage)

	now := time.Now()

	var sensors Sensors
	sensors.Memory = float64(usage.Maxrss) / 1024.0
	sensors.Uptime = now.Sub(boottime).Seconds()
	sensors.CPU = time.Duration(usage.Utime.Nano()).Seconds()

	return sensors
}

func setFileLimits() error {
	var limit syscall.Rlimit
	limit.Cur = 512
	limit.Max = 512
	return syscall.Setrlimit(syscall.RLIMIT_NOFILE, &limit)
}
func setMemLimits() error {
	var limit syscall.Rlimit
	limit.Cur = 2 * 1024 * 1024 * 1024
	limit.Max = 2 * 1024 * 1024 * 1024
	return syscall.Setrlimit(syscall.RLIMIT_DATA, &limit)
}
