//
// Copyright (c) 2019 Ted Unangst <tedu@tedunangst.com>
//
// Permission to use, copy, modify, and distribute this software for any
// purpose with or without fee is hereby granted, provided that the above
// copyright notice and this permission notice appear in all copies.
//
// THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
// WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
// ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
// WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
// ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
// OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

package main

/*
#include <termios.h>

void
termecho(int on)
{
	struct termios t;
	tcgetattr(1, &t);
	if (on)
		t.c_lflag |= ECHO;
	else
		t.c_lflag &= ~ECHO;
	tcsetattr(1, TCSADRAIN, &t);
}
*/
import "C"

import (
	"bufio"
	"crypto/rand"
	"crypto/rsa"
	"database/sql"
	"fmt"
	"os"
	"os/signal"
	"regexp"
	"strings"

	"golang.org/x/crypto/bcrypt"
	_ "humungus.tedunangst.com/r/go-sqlite3"
	"humungus.tedunangst.com/r/webs/httpsig"
	"humungus.tedunangst.com/r/webs/login"
)

var re_plainname = regexp.MustCompile("^[[:alnum:]_-]+$")

var alreadyopendb *sql.DB
var stmtConfig *sql.Stmt

func initdb() {
	os.Mkdir(dataDir+"/attachments", 0755)
	os.Mkdir(dataDir+"/tempstore", 0755)
	dbname := dataDir + "/azorius.db"
	_, err := os.Stat(dbname)
	if err == nil {
		elog.Fatalf("%s already exists", dbname)
	}
	db, err := sql.Open("sqlite3", dbname)
	if err != nil {
		elog.Fatal(err)
	}
	alreadyopendb = db
	defer func() {
		os.Remove(dbname)
		os.Exit(1)
	}()
	c := make(chan os.Signal, 1)
	signal.Notify(c, os.Interrupt)
	go func() {
		<-c
		C.termecho(1)
		fmt.Printf("\n")
		os.Remove(dbname)
		os.Exit(1)
	}()

	_, err = db.Exec("PRAGMA journal_mode=WAL")
	if err != nil {
		elog.Print(err)
		return
	}
	for _, line := range strings.Split(sqlSchema, ";") {
		_, err = db.Exec(line)
		if err != nil {
			elog.Print(err)
			return
		}
	}
	prepareStatements(db)

	r := bufio.NewReader(os.Stdin)

	fmt.Printf("listen address: ")
	addr, err := r.ReadString('\n')
	if err != nil {
		elog.Print(err)
		return
	}
	addr = addr[:len(addr)-1]
	if len(addr) < 1 {
		elog.Print("that's way too short")
		return
	}
	setconfig("listenaddr", addr)
	fmt.Printf("server name: ")
	addr, err = r.ReadString('\n')
	if err != nil {
		elog.Print(err)
		return
	}
	serverName = addr[:len(addr)-1]
	if len(addr) < 1 {
		elog.Print("that's way too short")
		return
	}
	setconfig("servername", serverName)

	err = createuserinteractive(db, r, true)
	if err != nil {
		elog.Print(err)
		return
	}
	_, err = db.Exec("insert into persons values (-1, '[deleted]', '[deleted]', '.', '{}')")
	if err != nil {
		elog.Print(err)
		return
	}

	var randbytes [16]byte
	rand.Read(randbytes[:])
	key := fmt.Sprintf("%x", randbytes)
	setconfig("csrfkey", key)
	setconfig("dbversion", myVersion)

	setconfig("servermsg", "Things happen.")
	setconfig("aboutmsg", "### What is azorius?\nIt's pretty freaking alpha is what it is!")
	setconfig("loginmsg", "**login**")
	setconfig("devel", 0)

	db.Close()
	fmt.Printf("done.\n")
	os.Exit(0)
}

func adduser() {
	db := opendatabase()
	defer func() {
		os.Exit(1)
	}()
	c := make(chan os.Signal, 1)
	signal.Notify(c, os.Interrupt)
	go func() {
		<-c
		C.termecho(1)
		fmt.Printf("\n")
		os.Exit(1)
	}()

	r := bufio.NewReader(os.Stdin)

	err := createuserinteractive(db, r, false)
	if err != nil {
		elog.Print(err)
		return
	}
	os.Exit(0)
}

func chpass(username string) {
	user := finduser(username)
	if user == nil {
		elog.Fatal("no user")
	}
	defer func() {
		os.Exit(1)
	}()
	c := make(chan os.Signal, 1)
	signal.Notify(c, os.Interrupt)
	go func() {
		<-c
		C.termecho(1)
		fmt.Printf("\n")
		os.Exit(1)
	}()

	db := opendatabase()
	login.Init(login.InitArgs{Db: db, Logger: ilog})

	r := bufio.NewReader(os.Stdin)

	pass, err := askpassword(r)
	if err != nil {
		elog.Print(err)
		return
	}
	err = login.SetPassword(user.ID, pass)
	if err != nil {
		elog.Print(err)
		return
	}
	userid := user.ID
	_, err = db.Exec("delete from auth where userid = ?", userid)
	if err != nil {
		elog.Print(err)
		return
	}

	fmt.Printf("done\n")
	os.Exit(0)
}

func askpassword(r *bufio.Reader) (string, error) {
	C.termecho(0)
	fmt.Printf("password: ")
	pass, err := r.ReadString('\n')
	C.termecho(1)
	fmt.Printf("\n")
	if err != nil {
		return "", err
	}
	pass = pass[:len(pass)-1]
	if len(pass) < 6 {
		return "", fmt.Errorf("that's way too short")
	}
	return pass, nil
}

func createuserinteractive(db *sql.DB, r *bufio.Reader, isAdmin bool) error {
	fmt.Printf("username: ")
	name, err := r.ReadString('\n')
	if err != nil {
		return err
	}
	name = name[:len(name)-1]
	if len(name) < 1 {
		return fmt.Errorf("that's way too short")
	}
	if !re_plainname.MatchString(name) {
		return fmt.Errorf("alphanumeric only please")
	}
	if u2 := finduser(name); u2 != nil {
		return fmt.Errorf("user already exists")
	}
	pass, err := askpassword(r)
	if err != nil {
		return err
	}
	flags := flagIsApproved
	if isAdmin {
		flags |= flagIsAdmin
	}
	email := ""
	about := "what about me?"
	return createuser(name, pass, email, about, flags)
}

func generateKeys() (string, string, error) {
	k, err := rsa.GenerateKey(rand.Reader, 2048)
	if err != nil {
		return "", "", err
	}
	pubkey, err := httpsig.EncodeKey(&k.PublicKey)
	if err != nil {
		return "", "", err
	}
	seckey, err := httpsig.EncodeKey(k)
	if err != nil {
		return "", "", err
	}
	return pubkey, seckey, nil
}

func createuser(name, pass, email, about string, flags int64) error {
	hash, err := bcrypt.GenerateFromPassword([]byte(pass), 12)
	if err != nil {
		return err
	}
	person := Person{
		Name:     name,
		Display:  name,
		About:    about,
		Format:   "markdown",
		Inbox:    fmt.Sprintf("https://%s/u/%s/inbox", serverName, name),
		ShareBox: fmt.Sprintf("https://%s/inbox", serverName),
	}
	err = dbSavePerson(&person)
	if err != nil {
		return err
	}
	apid := person.APid
	pubkey, seckey, err := generateKeys()
	if err == nil {
		err = dbSavePubkey(apid+"#key", apid, pubkey)
	}
	if err == nil {
		err = dbSaveSeckey(apid, seckey)
	}
	if err != nil {
		return err
	}
	user := User{
		Name:   name,
		Hash:   hash,
		Email:  email,
		Flags:  flags,
		Person: &person,
	}
	err = dbSaveUser(&user)
	if err != nil {
		return err
	}
	return nil
}

func opendatabase() *sql.DB {
	if alreadyopendb != nil {
		return alreadyopendb
	}
	dbname := dataDir + "/azorius.db"
	_, err := os.Stat(dbname)
	if os.IsNotExist(err) {
		elog.Fatalf("database %s does not exit", dbname)
	}
	if err != nil {
		elog.Fatalf("unable to open database: %s", err)
	}
	db, err := sql.Open("sqlite3", dbname)
	if err != nil {
		elog.Fatalf("unable to open database: %s", err)
	}
	stmtConfig, err = db.Prepare("select value from config where key = ?")
	if err != nil {
		elog.Fatal(err)
	}
	alreadyopendb = db
	return db
}

func getconfig(key string, value interface{}) error {
	m, ok := value.(*map[string]bool)
	if ok {
		rows, err := stmtConfig.Query(key)
		if err != nil {
			return err
		}
		defer rows.Close()
		for rows.Next() {
			var s string
			err = rows.Scan(&s)
			if err != nil {
				return err
			}
			(*m)[s] = true
		}
		return nil
	}
	row := stmtConfig.QueryRow(key)
	err := row.Scan(value)
	if err == sql.ErrNoRows {
		err = nil
	}
	return err
}

func setconfig(key string, val interface{}) error {
	db := opendatabase()
	db.Exec("delete from config where key = ?", key)
	_, err := db.Exec("insert into config (key, value) values (?, ?)", key, val)
	return err
}
