//
// Copyright (c) 2023 Ted Unangst <tedu@tedunangst.com>
//
// Permission to use, copy, modify, and distribute this software for any
// purpose with or without fee is hereby granted, provided that the above
// copyright notice and this permission notice appear in all copies.
//
// THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
// WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
// ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
// WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
// ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
// OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

package main

import (
	"fmt"
	"html/template"
	"strings"
	"time"

	"humungus.tedunangst.com/r/webs/junk"
	"humungus.tedunangst.com/r/webs/templates"
)

type User struct {
	ID     int64
	Name   string
	Hash   []byte
	Flags  int64
	Email  string
	Person *Person
}

func (*User) MarshalJSON() ([]byte, error) {
	panic("cannot json")
}

const (
	flagIsAdmin    int64 = 1
	flagIsApproved int64 = 2
)

func (user *User) IsAdmin() bool {
	return user.Flags&flagIsAdmin != 0
}

func (user *User) IsApproved() bool {
	return user.Flags&flagIsApproved != 0
}

type Person struct {
	ID       int64
	APid     string
	Name     string
	Display  string
	About    string
	Format   string
	HTAbout  template.HTML
	Inbox    string
	ShareBox string
}

func (*Person) MarshalJSON() ([]byte, error) {
	panic("cannot json")
}

func (person *Person) Rel() string {
	return "/u/" + person.Name
}

func (person *Person) IsRemote() bool {
	return !person.IsLocal()
}
func (person *Person) IsLocal() bool {
	return strings.HasPrefix(person.APid, serverPrefix)
}

type Group struct {
	ID        int64
	APid      string
	Name      string
	Display   string
	Desc      string
	Format    string
	HTDesc    template.HTML
	Mods      []string
	Featured  []string
	Flags     int64
	Inbox     string
	ShareBox  string
	Outbox    string
	Refreshed time.Time
}

func (*Group) MarshalJSON() ([]byte, error) {
	panic("cannot json")
}

const (
	flagOnFront int64 = 1
	// isApproved = 2
	flagHasChat  = 4
	flagPicFocus = 8
	flagModPosts = 16
)

func (group *Group) Rel() string {
	return "/g/" + group.Name
}
func (group *Group) OnFront() bool {
	return group.Flags&flagOnFront != 0
}
func (group *Group) IsApproved() bool {
	return group.Flags&flagIsApproved != 0
}
func (group *Group) HasChat() bool {
	return group.Flags&flagHasChat != 0
}
func (group *Group) PicFocus() bool {
	return group.Flags&flagPicFocus != 0
}
func (group *Group) ModPosts() bool {
	return group.Flags&flagModPosts != 0
}

func (group *Group) IsRemote() bool {
	return !group.IsLocal()
}
func (group *Group) IsLocal() bool {
	return strings.HasPrefix(group.APid, serverPrefix)
}

type Post struct {
	ID          int64
	GroupID     int64
	GroupName   string
	PersonID    int64
	PersonName  string
	Xid         string
	APid        string
	Date        time.Time
	Pending     bool
	Title       string
	Link        string
	Site        string
	Content     string
	Format      string
	Context     string
	Tags        []string
	HTContent   template.HTML
	Attachments []*Attachment
	NumComments int64
}

func (*Post) MarshalJSON() ([]byte, error) {
	panic("cannot json")
}

func (post *Post) Rel() string {
	return fmt.Sprintf("/g/%s/p/%s", post.GroupName, post.Xid)
}

func (post *Post) IsRemote() bool {
	return !post.IsLocal()
}
func (post *Post) IsLocal() bool {
	return strings.HasPrefix(post.APid, serverPrefix)
}

type Attachment struct {
	ID     int64
	PostID int64
	Xid    string
	Media  string
	Hash   string
	Desc   string
}

func (*Attachment) MarshalJSON() ([]byte, error) {
	panic("cannot json")
}

func (att *Attachment) Url() string {
	return fmt.Sprintf("https://%s/file/%s", serverName, att.Xid)
}

type Comment struct {
	ID            int64
	PostID        int64
	PostXID       string
	PersonID      int64
	PersonName    string
	PersonDisplay string
	ParentID      int64
	Xid           string
	APid          string
	Date          time.Time
	Subject       string
	Content       string
	Format        string
	Tags          []string
	HTContent     template.HTML
	Kids          []*Comment
	Next          *Comment
}

func (*Comment) MarshalJSON() ([]byte, error) {
	panic("cannot json")
}

func (comment *Comment) Rel() string {
	return "/c/" + comment.Xid
}

func (comment *Comment) IsLocal() bool {
	return strings.HasPrefix(comment.APid, serverPrefix)
}
func (comment *Comment) IsRemote() bool {
	return !comment.IsLocal()
}

type Chat struct {
	ID         int64
	GroupID    int64
	GroupName  string
	PersonID   int64
	PersonName string
	Xid        string
	APid       string
	Date       time.Time
	Content    string
	HTContent  template.HTML
}

func (chat *Chat) MarshalJSON() ([]byte, error) {
	panic("cannot json")
}
func (chat *Chat) Rel() string {
	return fmt.Sprintf("/g/%s/chat/%s", chat.GroupName, chat.Xid)
}
func (chat *Chat) IsRemote() bool {
	return !strings.HasPrefix(chat.APid, serverPrefix)
}

func (chat *Chat) ShortName() template.HTML {
	name := chat.PersonName
	if len(name) > 12 {
		name = name[0:10] + ".."
	}
	return templates.Sprintf(`<span title="%s">%s</span>`, chat.PersonName, name)
}

func (chat *Chat) Timestamp() string {
	if bgnow.Sub(chat.Date) > 24*time.Hour {
		return chat.Date.Format("01/02")
	}
	return chat.Date.Format("15:04")
}

type Report struct {
	ID    int64
	Who   string
	What  string
	Link  string
	Where string
	Why   string
	Notes string
	Blame string
	Date  time.Time
}

func (*Report) MarshalJSON() ([]byte, error) {
	panic("cannot json")
}

type Notif struct {
	ID     int64
	UserID int64
	Date   time.Time
	Flags  int64
	Msg    string
	Url    string
}

func (*Notif) MarshalJSON() ([]byte, error) {
	panic("cannot json")
}

type Origin struct {
	Domain string
	Group  *Group
	Json   junk.Junk
}

func (origin *Origin) Match(url string) bool {
	return origin.Domain == originate(url)
}

func (origin *Origin) String() string {
	groupname := ""
	if origin.Group != nil {
		groupname = origin.Group.Name
	}
	return fmt.Sprintf("domain: %s group: %s json: %s", origin.Domain, groupname, origin.Json.ToString())
}
