//
// Copyright (c) 2023 Ted Unangst <tedu@tedunangst.com>
//
// Permission to use, copy, modify, and distribute this software for any
// purpose with or without fee is hereby granted, provided that the above
// copyright notice and this permission notice appear in all copies.
//
// THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
// WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
// ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
// WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
// ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
// OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

package main

import (
	"bytes"
	"crypto/sha512"
	"errors"
	"fmt"
	"html/template"
	"io"
	notrand "math/rand"
	"net"
	"net/http"
	"os"
	"sort"
	"strings"
	"time"

	"github.com/gorilla/mux"
	"humungus.tedunangst.com/r/webs/login"
	"humungus.tedunangst.com/r/webs/templates"
)

var readviews *templates.Template

func personforuser(r *http.Request) *Person {
	userinfo := login.GetUserInfo(r)
	return findperson(userinfo.Username)
}

func getInfo(r *http.Request) map[string]interface{} {
	templinfo := make(map[string]interface{})
	templinfo["StyleParam"] = getassetparam(viewDir + "/views/style.css")
	templinfo["LocalStyleParam"] = getassetparam(dataDir + "/views/local.css")
	templinfo["JSParam"] = getassetparam(viewDir + "/views/script.js")
	templinfo["LocalJSParam"] = getassetparam(dataDir + "/views/local.js")
	templinfo["ServerName"] = serverName
	templinfo["IconName"] = iconName
	templinfo["SiteAlert"] = HTsiteAlert
	if u := login.GetUserInfo(r); u != nil {
		templinfo["UserInfo"] = finduser(u.Username)
		templinfo["NumNotifs"] = numNotifs(u.UserID)
	}
	return templinfo
}

func httpGone(w http.ResponseWriter, r *http.Request) {
	http.Error(w, "410 Gone", http.StatusGone)
}

func oghome() template.HTML {
	imgurl := serverPrefix + "icon.png"
	short := aboutMsg
	if len(short) > 160 {
		short = short[0:160] + "..."
	}
	title := serverName
	return templates.Sprintf(
		`<meta property="og:title" content="%s" />
<meta property="og:type" content="website" />
<meta property="og:url" content="%s" />
<meta property="og:image" content="%s" />
<meta property="og:description" content="%s" />`,
		title, serverPrefix, imgurl, short)
}

func showfront(w http.ResponseWriter, r *http.Request) {
	posts := getpostsforfront()
	templinfo := getInfo(r)
	templinfo["OGSummary"] = oghome()
	templinfo["Posts"] = posts
	u := login.GetUserInfo(r)
	if u == nil && !develMode {
		w.Header().Set("Cache-Control", "max-age=60")
	}
	err := readviews.Execute(w, "homepage.html", templinfo)
	if err != nil {
		elog.Print(err)
	}
}

func showhome(w http.ResponseWriter, r *http.Request) {
	u := login.GetUserInfo(r)
	if u == nil {
		showfront(w, r)
		return
	}
	posts := getpostsforuser(u.UserID)
	if u != nil {
		user := finduser(u.Username)
		posts = filterPostsForUser(user, posts)
	}
	templinfo := getInfo(r)
	templinfo["OGSummary"] = oghome()
	templinfo["Posts"] = posts
	err := readviews.Execute(w, "homepage.html", templinfo)
	if err != nil {
		elog.Print(err)
	}
}

func showgroups(w http.ResponseWriter, r *http.Request) {
	u := login.GetUserInfo(r)
	groups := getgroups()
	var userid int64
	if u != nil {
		userid = u.UserID
		sort.Slice(groups, func(i, j int) bool {
			a, b := groups[i], groups[j]
			if a.IsSubbed(userid) == b.IsSubbed(userid) {
				if a.IsLocal() == b.IsLocal() {
					return groups[i].Name < groups[j].Name
				}
				return groups[i].IsLocal()
			}
			return a.IsSubbed(userid)
		})
	}
	templinfo := getInfo(r)
	templinfo["Groups"] = groups
	templinfo["UserID"] = userid
	templinfo["SaveCSRF"] = login.GetCSRF("savepost", r)
	if u == nil && !develMode {
		w.Header().Set("Cache-Control", "max-age=60")
	}
	err := readviews.Execute(w, "groups.html", templinfo)
	if err != nil {
		elog.Print(err)
	}
}

func shownewgroup(w http.ResponseWriter, r *http.Request) {
	templinfo := getInfo(r)
	templinfo["SaveCSRF"] = login.GetCSRF("savepost", r)
	err := readviews.Execute(w, "newgroup.html", templinfo)
	if err != nil {
		elog.Print(err)
	}
}

func oggroup(group *Group) template.HTML {
	imgurl := serverPrefix + "icon.png"
	short := group.Desc
	if len(short) > 160 {
		short = short[0:160] + "..."
	}
	title := group.Display
	return templates.Sprintf(
		`<meta property="og:title" content="%s" />
<meta property="og:type" content="website" />
<meta property="og:url" content="%s" />
<meta property="og:image" content="%s" />
<meta property="og:description" content="%s" />`,
		title, group.APid, imgurl, short)
}

func showonegroup(w http.ResponseWriter, r *http.Request) {
	name := mux.Vars(r)["name"]
	group := findgroup(name)
	if group == nil {
		url := "https://" + serverName + r.URL.String()
		if hastombstone(url) {
			httpGone(w, r)
		} else {
			http.NotFound(w, r)
		}
		return
	}
	if isAPrequest(r) {
		if isBannedRequest(r) {
			http.NotFound(w, r)
			return
		}
		if group.IsRemote() {
			http.NotFound(w, r)
		} else {
			apWriteGroup(group, w)
		}
		return
	}
	pending := false
	u := login.GetUserInfo(r)
	var user *User
	if u != nil {
		user = finduser(u.Username)
		pending = group.IsMod(user)
	}
	posts := getpostsforgroup(group, pending)
	if user != nil {
		posts = filterPostsForUser(user, posts)
	}
	if group.PicFocus() {
		getmanyattachments(posts)
	}
	featured := getfeaturedposts(group)
	templinfo := getInfo(r)
	if group.IsLocal() {
		templinfo["APAltLink"] = templates.Sprintf(`<link rel="alternate" type="application/activity+json" href="%s">`, group.APid)
	} else {
		templinfo["CanonLink"] = templates.Sprintf(`<link rel="canonical" href="%s">`, group.APid)
	}
	templinfo["OGSummary"] = oggroup(group)
	templinfo["Group"] = group
	templinfo["Featured"] = featured
	templinfo["Posts"] = posts
	templinfo["SaveCSRF"] = login.GetCSRF("savepost", r)
	if u == nil && !develMode {
		w.Header().Set("Cache-Control", "max-age=60")
	}
	err := readviews.Execute(w, "onegroup.html", templinfo)
	if err != nil {
		elog.Print(err)
	}
}

func shownewpost(w http.ResponseWriter, r *http.Request) {
	name := mux.Vars(r)["name"]
	group := findgroup(name)
	if group == nil {
		http.NotFound(w, r)
		return
	}
	templinfo := getInfo(r)
	templinfo["Group"] = group
	templinfo["SaveCSRF"] = login.GetCSRF("savepost", r)
	err := readviews.Execute(w, "newpost.html", templinfo)
	if err != nil {
		elog.Print(err)
	}
}

func showgroupchat(w http.ResponseWriter, r *http.Request) {
	name := mux.Vars(r)["name"]
	group := findgroup(name)
	if group == nil || !group.HasChat() {
		http.NotFound(w, r)
		return
	}
	chats := getchats(group.ID)
	reformatchats(chats)
	templinfo := getInfo(r)
	templinfo["Group"] = group
	templinfo["Chats"] = chats
	templinfo["SaveCSRF"] = login.GetCSRF("savepost", r)
	err := readviews.Execute(w, "groupchat.html", templinfo)
	if err != nil {
		elog.Print(err)
	}
}

func showgroupreports(w http.ResponseWriter, r *http.Request) {
	name := mux.Vars(r)["name"]
	group := findgroup(name)
	if group == nil {
		http.NotFound(w, r)
		return
	}
	user := finduser(login.GetUserInfo(r).Username)
	if !group.IsMod(user) {
		errorpage(w, r, "can't see that")
		return
	}

	reports := getreportsbywhere(group.APid)
	templinfo := getInfo(r)
	templinfo["Group"] = group
	templinfo["Reports"] = reports
	templinfo["SaveCSRF"] = login.GetCSRF("savepost", r)
	err := readviews.Execute(w, "groupreports.html", templinfo)
	if err != nil {
		elog.Print(err)
	}
}

func editonegroup(w http.ResponseWriter, r *http.Request) {
	name := mux.Vars(r)["name"]
	group := findgroup(name)
	if group == nil {
		http.NotFound(w, r)
		return
	}
	templinfo := getInfo(r)
	templinfo["Group"] = group
	templinfo["SaveCSRF"] = login.GetCSRF("savepost", r)
	err := readviews.Execute(w, "editgroup.html", templinfo)
	if err != nil {
		elog.Print(err)
	}
}

func featuredAPids(group *Group, featured string) []string {
	var apids []string
	featured = strings.Replace(featured, "\r", "", -1)
	featlist := strings.Split(featured, "\n")
	for _, f := range featlist {
		f = strings.TrimSpace(f)
		post := findpost(f)
		if post != nil && post.GroupName == group.Name {
			apids = append(apids, post.APid)
		}
	}
	return apids
}

func websavegroup(w http.ResponseWriter, r *http.Request) {
	name := r.FormValue("name")
	display := r.FormValue("display")
	desc := r.FormValue("desc")

	user := finduser(login.GetUserInfo(r).Username)
	update := r.FormValue("update") == "1"
	haschat := r.FormValue("haschat") == "1"
	modposts := r.FormValue("modposts") == "1"
	picfocus := r.FormValue("picfocus") == "1"
	var oldfeatured []string
	changed := false

	mtxSaveGroup.Lock()
	defer mtxSaveGroup.Unlock()
	mtxSavePerson.Lock()
	defer mtxSavePerson.Unlock()
	group := dbFindGroup(name)
	if update {
		featured := r.FormValue("featured")
		if group == nil {
			errorpage(w, r, "group not found")
			return
		}
		if !group.IsMod(user) {
			errorpage(w, r, "can't edit that")
			return
		}
		if group.IsLocal() {
			if group.Display != display {
				changed = true
				group.Display = display
			}
			if group.Desc != desc {
				changed = true
				group.Desc = desc
			}
			if modposts != group.ModPosts() {
				changed = true
				if modposts {
					group.Flags |= flagModPosts
				} else {
					group.Flags &= ^flagModPosts
				}
			}
			if haschat != group.HasChat() {
				changed = true
				if haschat {
					group.Flags |= flagHasChat
				} else {
					group.Flags &= ^flagHasChat
				}
			}
			oldfeatured = group.Featured
			group.Featured = featuredAPids(group, featured)
		}
		if picfocus != group.PicFocus() {
			changed = true
			if picfocus {
				group.Flags |= flagPicFocus
			} else {
				group.Flags &= ^flagPicFocus
			}
		}
	} else {
		if !re_plainname.MatchString(name) {
			errorpage(w, r, "need a plainer name")
			return
		}
		if group != nil {
			errorpage(w, r, "already have one")
			return
		}
		if dbFindUser(name) != nil {
			errorpage(w, r, "already have one")
			return
		}
		if hastombstone(name) {
			errorpage(w, r, "no resurrection")
			return
		}
		if display == "" {
			display = name
		}
		var flags int64
		if haschat {
			flags |= flagHasChat
		}
		group = &Group{
			Name:     name,
			Display:  display,
			Desc:     desc,
			Format:   "markdown",
			Mods:     []string{user.Name},
			Inbox:    fmt.Sprintf("https://%s/g/%s/inbox", serverName, name),
			ShareBox: fmt.Sprintf("https://%s/inbox", serverName),
			Flags:    flags,
		}
	}
	err := dbSaveGroup(group)
	if err != nil {
		elog.Printf("error saving group: %s", err)
		errorpage(w, r, "bad times")
		return
	}
	if !update {
		apid := group.APid
		pubkey, seckey, err := generateKeys()
		if err == nil {
			err = dbSavePubkey(apid+"#key", apid, pubkey)
		}
		if err == nil {
			err = dbSaveSeckey(apid, seckey)
		}
		if err != nil {
			elog.Printf("error generating keys: %s", err)
			errorpage(w, r, "bad times")
			return
		}
		notifNewGroup(group)
	}
	group = findgroup(name)

	if update && group.IsLocal() {
		apSendGroupUpdate(group, changed, oldfeatured)
	}
	http.Redirect(w, r, group.Rel(), http.StatusSeeOther)
}

func ogsummary(post *Post) template.HTML {
	imgurl := serverPrefix + "icon.png"
	for _, att := range post.Attachments {
		if strings.HasPrefix(att.Media, "image") {
			imgurl = att.Url()
			break
		}
	}
	short := post.Content
	if len(short) > 160 {
		short = short[0:160] + "..."
	}
	poster := findperson(post.PersonName)
	title := post.Title
	return templates.Sprintf(
		`<meta property="og:title" content="%s" />
<meta property="og:type" content="article" />
<meta property="article:author" content="%s" />
<meta property="og:url" content="%s" />
<meta property="og:image" content="%s" />
<meta property="og:description" content="%s" />`,
		title, poster.APid, post.APid, imgurl, short)
}

func showonepost(w http.ResponseWriter, r *http.Request) {
	xid := mux.Vars(r)["xid"]
	post := findpost(xid)
	if post == nil {
		url := "https://" + serverName + r.URL.String()
		if hastombstone(url) {
			httpGone(w, r)
		} else {
			http.NotFound(w, r)
		}
		return
	}
	if name := mux.Vars(r)["name"]; name != "" && name != post.GroupName {
		http.NotFound(w, r)
		return
	}
	if isAPrequest(r) {
		if isBannedRequest(r) {
			http.NotFound(w, r)
			return
		}
		if post.IsRemote() {
			http.NotFound(w, r)
		} else {
			apWritePost(post, w)
		}
		return
	}
	comments := getcomments(post.ID)
	u := login.GetUserInfo(r)
	if u != nil {
		user := finduser(u.Username)
		comments = filterCommentsForUser(user, comments)
	}
	reformatcomments(comments)
	comments = threadcomments(comments)
	templinfo := getPostInfo(post, r)
	if post.IsLocal() {
		templinfo["APAltLink"] = templates.Sprintf(`<link rel="alternate" type="application/activity+json" href="%s">`, post.APid)
	} else {
		templinfo["CanonLink"] = templates.Sprintf(`<link rel="canonical" href="%s">`, post.APid)
	}
	templinfo["OGSummary"] = ogsummary(post)
	templinfo["Comments"] = comments
	if u == nil && !develMode {
		w.Header().Set("Cache-Control", "max-age=60")
	}
	err := readviews.Execute(w, "onepost.html", templinfo)
	if err != nil {
		elog.Print(err)
	}
}

func getPostInfo(post *Post, r *http.Request) map[string]interface{} {
	templinfo := getInfo(r)
	templinfo["Post"] = post
	templinfo["Group"] = findgroup(post.GroupName)
	templinfo["SaveCSRF"] = login.GetCSRF("savepost", r)
	return templinfo
}

func editonepost(w http.ResponseWriter, r *http.Request) {
	xid := mux.Vars(r)["xid"]
	post := findpost(xid)
	if post == nil {
		http.NotFound(w, r)
		return
	}
	templinfo := getInfo(r)
	templinfo["Post"] = post
	templinfo["SaveCSRF"] = login.GetCSRF("savepost", r)
	err := readviews.Execute(w, "editpost.html", templinfo)
	if err != nil {
		elog.Print(err)
	}
}

func approveonepost(w http.ResponseWriter, r *http.Request) {
	xid := mux.Vars(r)["xid"]
	post := findpost(xid)
	if post == nil {
		http.NotFound(w, r)
		return
	}
	user := finduser(login.GetUserInfo(r).Username)
	if !post.CanApprove(user) {
		http.NotFound(w, r)
		return
	}
	post.Pending = false
	err := dbSavePost(post)
	if err != nil {
		elog.Printf("error saving post: %s", err)
		errorpage(w, r, "didn't work")
		return
	}
	poster := findperson(post.PersonName)
	apSendPost(poster, post)
	http.Redirect(w, r, post.Rel(), http.StatusSeeOther)
}

func showonecomment(w http.ResponseWriter, r *http.Request) {
	xid := mux.Vars(r)["xid"]
	comment := findcomment(xid)
	if comment == nil {
		url := "https://" + serverName + r.URL.String()
		if hastombstone(url) {
			httpGone(w, r)
		} else {
			http.NotFound(w, r)
		}
		return
	}
	if isAPrequest(r) {
		if isBannedRequest(r) {
			http.NotFound(w, r)
			return
		}
		if comment.IsRemote() {
			http.NotFound(w, r)
		} else {
			apWriteComment(comment, w)
		}
		return
	}
	post := findpost(comment.PostXID)
	templinfo := getInfo(r)
	if comment.IsLocal() {
		templinfo["APAltLink"] = templates.Sprintf(`<link rel="alternate" type="application/activity+json" href="%s">`, comment.APid)
	} else {
		templinfo["CanonLink"] = templates.Sprintf(`<link rel="canonical" href="%s">`, comment.APid)
	}
	templinfo["Comment"] = comment
	templinfo["Post"] = post
	u := login.GetUserInfo(r)
	if u == nil && !develMode {
		w.Header().Set("Cache-Control", "max-age=60")
	}
	err := readviews.Execute(w, "onecomment.html", templinfo)
	if err != nil {
		elog.Print(err)
	}
}

func editonecomment(w http.ResponseWriter, r *http.Request) {
	xid := mux.Vars(r)["xid"]
	comment := findcomment(xid)
	if comment == nil {
		http.NotFound(w, r)
		return
	}
	templinfo := getInfo(r)
	templinfo["Comment"] = comment
	templinfo["SaveCSRF"] = login.GetCSRF("savepost", r)
	err := readviews.Execute(w, "editcomment.html", templinfo)
	if err != nil {
		elog.Print(err)
	}
}

func showonechat(w http.ResponseWriter, r *http.Request) {
	name := mux.Vars(r)["name"]
	group := findgroup(name)
	xid := mux.Vars(r)["xid"]
	chat := findchat(xid)
	if group == nil || chat == nil || chat.GroupName != group.Name || !group.HasChat() {
		http.NotFound(w, r)
		return
	}
	if isAPrequest(r) {
		if isBannedRequest(r) {
			http.NotFound(w, r)
			return
		}
		if chat.IsRemote() {
			http.NotFound(w, r)
		} else {
			apWriteChat(chat, w)
		}
		return
	}
}

func showoneperson(w http.ResponseWriter, r *http.Request) {
	name := mux.Vars(r)["name"]
	person := findperson(name)
	if person == nil {
		url := "https://" + serverName + r.URL.String()
		if hastombstone(url) {
			httpGone(w, r)
		} else {
			http.NotFound(w, r)
		}
		return
	}
	if isAPrequest(r) {
		if isBannedRequest(r) {
			http.NotFound(w, r)
			return
		}
		if person.IsRemote() {
			http.NotFound(w, r)
		} else {
			apWritePerson(person, w)
		}
		return
	}
	u := login.GetUserInfo(r)
	if u == nil && person.IsRemote() {
		http.NotFound(w, r)
		return
	}
	posts := getpostsbyperson(person.ID)
	templinfo := getInfo(r)
	if person.IsLocal() {
		templinfo["APAltLink"] = templates.Sprintf(`<link rel="alternate" type="application/activity+json" href="%s">`, person.APid)
	} else {
		templinfo["CanonLink"] = templates.Sprintf(`<link rel="canonical" href="%s">`, person.APid)
	}
	templinfo["Person"] = person
	templinfo["Posts"] = posts
	templinfo["SaveCSRF"] = login.GetCSRF("savepost", r)
	if u == nil && !develMode {
		w.Header().Set("Cache-Control", "max-age=60")
	}
	err := readviews.Execute(w, "oneperson.html", templinfo)
	if err != nil {
		elog.Print(err)
	}
}

func webshare(w http.ResponseWriter, r *http.Request) {
	postxid := r.FormValue("postxid")
	post := findpost(postxid)
	if post == nil {
		errorpage(w, r, "which post?")
		return
	}
	user := finduser(login.GetUserInfo(r).Username)
	err := dbSaveShare(user.Person.ID, post.ID)
	if err != nil {
		elog.Printf("error saving share: %s", err)
		errorpage(w, r, "bad times")
		return
	}
	apAnnouncePost(user.Person, post)
	http.Redirect(w, r, post.Rel(), http.StatusSeeOther)
}

func websubscribe(w http.ResponseWriter, r *http.Request) {
	var groupid, personid, postid int64
	var group *Group
	var person *Person
	var post *Post
	if groupxid := r.FormValue("group"); groupxid != "" {
		group = findgroup(groupxid)
		if group == nil {
			errorpage(w, r, "which group?")
			return
		}
		groupid = group.ID
	} else if personxid := r.FormValue("person"); personxid != "" {
		person = findperson(personxid)
		if person == nil {
			errorpage(w, r, "which person?")
			return
		}
		personid = person.ID
	} else if postxid := r.FormValue("post"); postxid != "" {
		post = findpost(postxid)
		if post == nil {
			errorpage(w, r, "which post?")
			return
		}
		postid = post.ID
	} else {
		errorpage(w, r, "which what?")
		return
	}

	u := login.GetUserInfo(r)
	err := dbSaveSub(u.UserID, groupid, personid, postid)
	if err != nil {
		elog.Printf("error saving sub: %s", err)
		errorpage(w, r, "bad times")
		return
	}
	if group != nil {
		if group.IsRemote() {
			me := findperson(u.Username)
			folxid := me.APid + "/follow/" + generateXid()
			err = dbSaveFollow(folxid, me.APid, group.APid)
			if err != nil {
				elog.Printf("error saving follow: %s", err)
			} else {
				apSendFollow(folxid, me.APid, group.APid)
			}
		}
		http.Redirect(w, r, group.Rel(), http.StatusSeeOther)
	}
	if person != nil {
		if person.IsRemote() {
			me := findperson(u.Username)
			folxid := me.APid + "/follow/" + generateXid()
			err = dbSaveFollow(folxid, me.APid, person.APid)
			if err != nil {
				elog.Printf("error saving follow: %s", err)
			} else {
				apSendFollow(folxid, me.APid, person.APid)
			}
		}
		http.Redirect(w, r, person.Rel(), http.StatusSeeOther)
	}
	if post != nil {
		http.Redirect(w, r, post.Rel(), http.StatusSeeOther)
	}
}

func webunsubscribe(w http.ResponseWriter, r *http.Request) {
	var groupid, personid, postid int64
	var group *Group
	var person *Person
	var post *Post
	if groupxid := r.FormValue("group"); groupxid != "" {
		group = findgroup(groupxid)
		if group == nil {
			errorpage(w, r, "which group?")
			return
		}
		groupid = group.ID
	} else if personxid := r.FormValue("person"); personxid != "" {
		person = findperson(personxid)
		if person == nil {
			errorpage(w, r, "which person?")
			return
		}
		personid = person.ID
	} else if postxid := r.FormValue("post"); postxid != "" {
		post = findpost(postxid)
		if post == nil {
			errorpage(w, r, "which post?")
			return
		}
		postid = post.ID
	} else {
		errorpage(w, r, "which what?")
		return
	}
	u := login.GetUserInfo(r)
	err := dbDeleteSub(u.UserID, groupid, personid, postid)
	if err != nil {
		elog.Printf("error deleting sub: %s", err)
		errorpage(w, r, "bad times")
		return
	}
	if group != nil {
		if group.IsRemote() {
			me := findperson(u.Username)
			err = dbDeleteFollow(me.APid, group.APid)
			if err != nil {
				elog.Printf("error deleting follow: %s", err)
			}
			apSendUnfollow(me.APid, group.APid)
		}
		http.Redirect(w, r, group.Rel(), http.StatusSeeOther)
	}
	if person != nil {
		if person.IsRemote() {
			me := findperson(u.Username)
			err = dbDeleteFollow(me.APid, person.APid)
			if err != nil {
				elog.Printf("error deleting follow: %s", err)
			}
			apSendUnfollow(me.APid, person.APid)
		}
		http.Redirect(w, r, person.Rel(), http.StatusSeeOther)
	}
	if post != nil {
		http.Redirect(w, r, post.Rel(), http.StatusSeeOther)
	}
}

func websaveattachment(w http.ResponseWriter, r *http.Request) (*Attachment, error) {
	if !strings.HasPrefix(strings.ToLower(r.Header.Get("Content-Type")), "multipart/form-data") {
		return nil, nil
	}
	file, filehdr, err := r.FormFile("attachment")
	if err != nil {
		if err == http.ErrMissingFile {
			attxid := r.FormValue("attachxid")
			if attxid != "" {
				return attachmentfromxid(attxid)
			}
			return nil, nil
		}
		return nil, err
	}
	dlog.Printf("upload attachment: %s", filehdr.Filename)
	var buf bytes.Buffer
	io.Copy(&buf, file)
	file.Close()
	data := buf.Bytes()
	return attachmentfromdata(data, "", true)
}

func simpleName(name string) string {
	if idx := strings.IndexByte(name, '@'); idx > -1 {
		return name[0:idx]
	}
	return name
}

func tagSet(tags []string) []string {
	var set []string
	for _, tag := range tags {
		if tag == "" {
			continue
		}
		if tag[0] != '#' {
			tag = "#" + tag
		}
		set = append(set, tag)
	}
	sort.Strings(set)
	j := 0
	for i := 1; i < len(set); i++ {
		if set[j] == set[i] {
			continue
		}
		j++
		set[j] = set[i]
	}
	return set[0 : j+1]
}

func websaveperson(w http.ResponseWriter, r *http.Request) {
	name := r.FormValue("name")
	isDelete := r.FormValue("delete") == "1"

	mtxSavePerson.Lock()
	defer mtxSavePerson.Unlock()

	me := finduser(login.GetUserInfo(r).Username)
	if isDelete {
		person := findperson(name)
		if person == nil {
			errorpage(w, r, "can't find person")
			return
		}
		if !me.IsAdmin() {
			errorpage(w, r, "can't delete that")
			return
		}
		confirmed := r.FormValue("confirm") == "confirm"
		if !confirmed {
			errorpage(w, r, "not confirmed")
			return
		}
		ilog.Printf("deleting person: %s", name)
		var err error
		if person.IsLocal() {
			user := finduser(name)
			apSendDeletePerson(user.Person.APid)
			err = dbDeleteUser(user)
		} else {
			err = dbDeletePerson(person)
		}
		if err != nil {
			elog.Printf("error deleting person: %s", err)
		}
		http.Redirect(w, r, "/", http.StatusSeeOther)
		return
	}
	errorpage(w, r, "can only delete")
}

func websavepost(w http.ResponseWriter, r *http.Request) {
	title := r.FormValue("title")
	link := r.FormValue("link")
	content := r.FormValue("content")
	preview := r.FormValue("preview") == "preview"
	update := r.FormValue("update") == "1"
	isDelete := r.FormValue("delete") == "1"
	tags := strings.Split(r.FormValue("tags"), " ")

	att, err := websaveattachment(w, r)
	if err != nil {
		ilog.Printf("attachment failed: %s", err)
		errorpage(w, r, "trouble with the attachment")
		return
	}
	attdesc := r.FormValue("attachdesc")

	link = normalizeLink(link, "")
	title = normalizeTitle(title)

	user := finduser(login.GetUserInfo(r).Username)
	person := user.Person
	var post *Post
	if update || isDelete {
		postxid := r.FormValue("postxid")
		post = findpost(postxid)
		if post == nil {
			errorpage(w, r, "can't find post")
			return
		}
		group := findgroup(post.GroupName)
		if isDelete {
			if !post.CanMod(user) {
				errorpage(w, r, "can't delete that")
				return
			}
			confirmed := r.FormValue("confirm") == "confirm"
			if !confirmed {
				errorpage(w, r, "not confirmed")
				return
			}
			if post.IsLocal() || group.IsLocal() {
				apSendDeletePost(user.Person.APid, post)
			}
			ilog.Printf("deleting post: %s", postxid)
			err = dbDeletePost(post)
			if err != nil {
				elog.Printf("error deleting post: %s", err)
			}
			http.Redirect(w, r, group.Rel(), http.StatusSeeOther)
			return
		}
		if !post.CanEdit(user) {
			errorpage(w, r, "can't edit that")
			return
		}
		tags = append(tags, simpleName(group.Name))
		post = dbFindPost(postxid)
		post.Title = title
		post.Link = link
		post.Content = content
		post.Tags = tagSet(tags)
		for _, att := range post.Attachments {
			att.Desc = attdesc
			err = dbSaveAttachment(att)
			if err != nil {
				elog.Printf("error saving att: %s", err)
			}
		}
	} else {
		groupxid := r.FormValue("group")
		group := findgroup(groupxid)
		if group == nil {
			errorpage(w, r, "which group?")
			return
		}
		if link != "" {
			if xid := dbFindPostForLink(group.ID, link); xid != "" {
				redir := fmt.Sprintf("/g/%s/p/%s", group.Name, xid)
				http.Redirect(w, r, redir, http.StatusSeeOther)
				return
			}
		}
		tags = append(tags, simpleName(group.Name))
		post = &Post{
			GroupID:    group.ID,
			GroupName:  group.Name,
			PersonID:   person.ID,
			PersonName: person.Name,
			Date:       time.Now(),
			Title:      title,
			Link:       link,
			Content:    content,
			Format:     "markdown",
			Tags:       tagSet(tags),
		}
		if group.IsLocal() && group.ModPosts() {
			if !group.IsMod(user) {
				post.Pending = true
			}
		}
	}
	if preview {
		if att != nil {
			att.Desc = attdesc
			post.Attachments = append(post.Attachments, att)
		}
		post.HTContent = ""
		reformatpost(post)
		templinfo := getPostInfo(post, r)
		templinfo["Preview"] = true
		err := readviews.Execute(w, "editpost.html", templinfo)
		if err != nil {
			elog.Print(err)
		}
		return
	}
	err = dbSavePost(post)
	if err != nil {
		elog.Printf("error saving post: %s", err)
		errorpage(w, r, "bad times")
		return
	}
	post = findpost(post.Xid)

	if update && !post.Pending {
		apSendPostUpdate(person, post)
	} else {
		if att != nil {
			att.PostID = post.ID
			att.Desc = attdesc
			post.Attachments = append(post.Attachments, att)
			err = dbSaveAttachment(att)
			if err != nil {
				elog.Printf("error saving att: %s", err)
			}
		}
		group := findgroup(post.GroupName)
		notifNewPost(group, post)
		if !post.Pending {
			apSendPost(person, post)
		}
	}
	http.Redirect(w, r, post.Rel(), http.StatusSeeOther)
}

func webreply(w http.ResponseWriter, r *http.Request) {
	xid := r.FormValue("xid")
	parent := findcomment(xid)
	if parent == nil {
		http.NotFound(w, r)
		return
	}
	templinfo := getInfo(r)
	templinfo["Parent"] = parent
	templinfo["SaveCSRF"] = login.GetCSRF("savepost", r)
	err := readviews.Execute(w, "reply.html", templinfo)
	if err != nil {
		elog.Print(err)
	}
}

func websavecomment(w http.ResponseWriter, r *http.Request) {
	update := r.FormValue("update") == "1"
	isDelete := r.FormValue("delete") == "1"
	content := r.FormValue("content")

	user := finduser(login.GetUserInfo(r).Username)
	person := user.Person
	var comment, parent *Comment
	var post *Post
	if update || isDelete {
		commentxid := r.FormValue("commentxid")
		comment = findcomment(commentxid)
		if comment == nil {
			errorpage(w, r, "which comment?")
			return
		}
		post = findpost(comment.PostXID)
		if isDelete {
			if !comment.CanMod(user) {
				errorpage(w, r, "can't delete that")
				return
			}
			confirmed := r.FormValue("confirm") == "confirm"
			if !confirmed {
				errorpage(w, r, "not confirmed")
				return
			}
			group := findgroup(post.GroupName)
			if post.IsLocal() || group.IsLocal() {
				apSendCommentDelete(person.APid, group, comment)
			}
			ilog.Printf("deleting comment: %s", commentxid)
			err := dbDeleteComment(comment)
			if err != nil {
				elog.Printf("error deleting comment: %s", err)
			}
			http.Redirect(w, r, post.Rel(), http.StatusSeeOther)
			return
		}
		if !comment.CanEdit(user) {
			errorpage(w, r, "can't edit that")
			return
		}
		comment = dbFindComment(commentxid)
		comment.Content = content
	} else {
		parentxid := r.FormValue("parentxid")
		var parentid int64
		if parentxid != "" {
			parent = findcomment(parentxid)
			if parent == nil {
				errorpage(w, r, "parent is gone")
				return
			}
			parentid = parent.ID
			post = findpost(parent.PostXID)
		} else {
			postxid := r.FormValue("postxid")
			post = findpost(postxid)
		}
		if post == nil {
			errorpage(w, r, "which post?")
			return
		}
		comment = &Comment{
			PostID:     post.ID,
			PersonID:   person.ID,
			PersonName: person.Name,
			ParentID:   parentid,
			Date:       time.Now(),
			Subject:    "",
			Content:    content,
			Format:     "markdown",
		}
	}
	err := dbSaveComment(comment)
	if err != nil {
		elog.Printf("error saving comment: %s", err)
		errorpage(w, r, "bad times")
		return
	}
	comment = findcomment(comment.Xid)

	group := findgroup(post.GroupName)
	if update {
		apSendCommentUpdate(person, group, comment)
	} else {
		notifNewComment(post, parent, comment)
		apSendComment(person, group, comment)
	}
	http.Redirect(w, r, post.Rel(), http.StatusSeeOther)
}

func websavechat(w http.ResponseWriter, r *http.Request) {
	u := login.GetUserInfo(r)
	person := findperson(u.Username)
	group := findgroup(r.FormValue("group"))
	if group == nil || !group.HasChat() {
		errorpage(w, r, "which group?")
		return
	}

	content := r.FormValue("content")
	content = string(markdown(content))
	chat := &Chat{
		GroupID:    group.ID,
		GroupName:  group.Name,
		PersonID:   person.ID,
		PersonName: person.Name,
		Date:       time.Now(),
		Content:    content,
	}
	err := dbSaveChat(chat)
	if err != nil {
		elog.Printf("error saving chat: %s", err)
		errorpage(w, r, "bad times")
		return
	}
	reformatchat(chat)

	notifNewChat(u.UserID, group, chat)
	apSendChat(person, group, chat)
	http.Redirect(w, r, group.Rel()+"/chat", http.StatusSeeOther)
}

func shownotifs(w http.ResponseWriter, r *http.Request) {
	u := login.GetUserInfo(r)
	notifs := getnotifs(u.UserID)
	clearNotifs(u.UserID)

	templinfo := getInfo(r)
	templinfo["Notifs"] = notifs
	err := readviews.Execute(w, "notifs.html", templinfo)
	if err != nil {
		elog.Print(err)
	}
}
func showaccount(w http.ResponseWriter, r *http.Request) {
	u := login.GetUserInfo(r)
	user := finduser(u.Username)
	templinfo := getInfo(r)
	templinfo["User"] = user
	templinfo["SaveCSRF"] = login.GetCSRF("saveaccount", r)
	templinfo["LogoutCSRF"] = login.GetCSRF("logout", r)
	err := readviews.Execute(w, "account.html", templinfo)
	if err != nil {
		elog.Print(err)
	}
}
func websaveuseraccount(w http.ResponseWriter, r *http.Request) {
	about := strings.TrimSpace(r.FormValue("about"))
	email := r.FormValue("email")
	display := r.FormValue("display")
	u := login.GetUserInfo(r)
	user := dbFindUser(u.Username)
	user.Person.About = about
	user.Email = email
	user.Person.Display = display
	dbSaveUser(user)
	dbSavePerson(user.Person)
	http.Redirect(w, r, "/account", http.StatusSeeOther)
}
func websaveaccount(w http.ResponseWriter, r *http.Request) {
	pg := r.FormValue("pg")
	switch pg {
	case "filters":
		websaveuserfilters(w, r)
	case "account":
		websaveuseraccount(w, r)
	default:
		errorpage(w, r, "saving what now?")
	}
}

func websearch(w http.ResponseWriter, r *http.Request) {
	q := r.FormValue("q")
	if len(q) == 0 {
		templinfo := getInfo(r)
		err := readviews.Execute(w, "searchhelp.html", templinfo)
		if err != nil {
			elog.Print(err)
		}
		return
	}
	if q[0] == '@' || q[0] == '!' {
		q = resolvefinger(q)
		if q == "" {
			errorpage(w, r, "not found")
			return
		}
	}
	if strings.HasPrefix(q, "https://") {
		apSearch(w, r, q)
		return
	}
	posts := searchposts(q)
	templinfo := getInfo(r)
	templinfo["Posts"] = posts
	u := login.GetUserInfo(r)
	if u == nil && !develMode {
		w.Header().Set("Cache-Control", "max-age=60")
	}
	err := readviews.Execute(w, "homepage.html", templinfo)
	if err != nil {
		elog.Print(err)
	}
}

func errorpage(w http.ResponseWriter, r *http.Request, msg string) {
	templinfo := getInfo(r)
	templinfo["Error"] = msg
	w.WriteHeader(http.StatusBadRequest)
	err := readviews.Execute(w, "error.html", templinfo)
	if err != nil {
		elog.Print(err)
	}
}

func websignup(w http.ResponseWriter, r *http.Request) {
	if !allowSignup {
		errorpage(w, r, "no signups")
		return
	}
	name := r.FormValue("username")
	if !re_plainname.MatchString(name) {
		errorpage(w, r, "need a plainer name (alnum only)")
		return
	}
	mtxSaveGroup.Lock()
	defer mtxSaveGroup.Unlock()
	mtxSavePerson.Lock()
	defer mtxSavePerson.Unlock()
	if dbFindUser(name) != nil {
		errorpage(w, r, "already taken!")
		return
	}
	if dbFindGroup(name) != nil {
		errorpage(w, r, "already taken!")
		return
	}
	if hastombstone(name) {
		errorpage(w, r, "no resurrection")
		return
	}
	password := r.FormValue("password")
	password2 := r.FormValue("password2")
	if len(password) < 12 {
		errorpage(w, r, "password must be at least 12")
		return
	}
	if password != password2 {
		errorpage(w, r, "password must be the password")
		return
	}
	email := r.FormValue("email")
	about := strings.TrimSpace(r.FormValue("about"))
	if about == "" {
		about = "the default about is good enough for me"
	}
	flags := int64(0)
	if openSignups {
		flags |= flagIsApproved
	}
	err := createuser(name, password, email, about, flags)
	if err != nil {
		elog.Printf("error creating user: %s", err)
		errorpage(w, r, "bad times")
	}
	ilog.Printf("new signup: %s from %s", name, r.Header.Get("X-Forwarded-For"))
	user := finduser(name)
	notifNewUser(user)
	login.LoginFunc(w, r)
}

func dochpass(w http.ResponseWriter, r *http.Request) {
	err := login.ChangePassword(w, r)
	if err != nil {
		elog.Printf("error changing password: %s", err)
	}
	http.Redirect(w, r, "/account", http.StatusSeeOther)
}

func somedays() string {
	secs := 432000 + notrand.Int63n(432000)
	return fmt.Sprintf("%d", secs)
}

func serveviewasset(w http.ResponseWriter, r *http.Request) {
	serveasset(w, r, viewDir)
}
func servedataasset(w http.ResponseWriter, r *http.Request) {
	if r.URL.Path == "/favicon.ico" {
		r.URL.Path = "/icon.png"
	}
	serveasset(w, r, dataDir)
}
func serveasset(w http.ResponseWriter, r *http.Request, basedir string) {
	if !develMode {
		w.Header().Set("Cache-Control", "max-age=7776000")
	}
	http.ServeFile(w, r, basedir+"/views"+r.URL.Path)
}
func servehelp(w http.ResponseWriter, r *http.Request) {
	name := mux.Vars(r)["name"]
	data, err := os.ReadFile(viewDir + "/docs/" + name)
	if err != nil {
		http.NotFound(w, r)
		return
	}
	templinfo := getInfo(r)
	templinfo["Help"] = template.HTML(data)
	u := login.GetUserInfo(r)
	if u == nil && !develMode {
		w.Header().Set("Cache-Control", "max-age=3600")
	}
	err = readviews.Execute(w, "help.html", templinfo)
	if err != nil {
		elog.Print(err)
	}
}
func servehtml(w http.ResponseWriter, r *http.Request) {
	u := login.GetUserInfo(r)
	templinfo := getInfo(r)
	templinfo["AboutMsg"] = markdown(aboutMsg)
	templinfo["LoginMsg"] = markdown(loginMsg)
	templinfo["AllowSignup"] = allowSignup
	templinfo["SignupMsg"] = markdown(signupMsg)
	templinfo["Version"] = softwareVersion
	if r.URL.Path == "/about" {
		templinfo["Sensors"] = getSensors()
	}
	if u == nil && !develMode {
		w.Header().Set("Cache-Control", "max-age=60")
	}
	err := readviews.Execute(w, r.URL.Path[1:]+".html", templinfo)
	if err != nil {
		elog.Print(err)
	}
}

func servefile(w http.ResponseWriter, r *http.Request) {
	xid := mux.Vars(r)["xid"]
	file, ok := findfile(xid)
	if !ok {
		http.NotFound(w, r)
		return
	}
	w.Header().Set("Content-Type", file.media)
	w.Header().Set("X-Content-Type-Options", "nosniff")
	w.Header().Set("Cache-Control", "max-age="+somedays())
	filename := longtermname(xid)
	if !fileexists(filename) {
		filename = shorttermname(xid)
	}
	preview := r.FormValue("preview") == "1"
	if preview {
		filename = previewfile(filename)
	}
	http.ServeFile(w, r, filename)
}

func showrobots(w http.ResponseWriter, r *http.Request) {
	io.WriteString(w, "User-agent: *\n")
	io.WriteString(w, "Disallow: /help/\n")
}

func fiveoh(w http.ResponseWriter, r *http.Request) {
	if !develMode {
		return
	}
	fd, err := os.OpenFile("violations.json", os.O_CREATE|os.O_WRONLY|os.O_APPEND, 0666)
	if err != nil {
		elog.Printf("error opening violations! %s", err)
		return
	}
	defer fd.Close()
	io.Copy(fd, r.Body)
	fd.WriteString("\n")
}

func addcspheaders(next http.Handler) http.Handler {
	return http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
		requestWG.Add(1)
		defer requestWG.Done()
		policy := "default-src 'none'; script-src 'self'; connect-src 'self'; style-src 'self'; img-src 'self'; media-src 'self'"
		if develMode {
			policy += "; report-uri /csp-violation"
		}
		w.Header().Set("Content-Security-Policy", policy)
		next.ServeHTTP(w, r)
	})
}

func requestlogger(next http.Handler) http.Handler {
	return http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
		start := time.Now()
		next.ServeHTTP(w, r)
		stop := time.Now()
		dlog.Printf("request %s took: %s", r.URL.Path, stop.Sub(start))
	})
}

var savedassetparams = make(map[string]string)

func getassetparam(file string) string {
	if p, ok := savedassetparams[file]; ok {
		return p
	}
	data, err := os.ReadFile(file)
	if err != nil {
		return ""
	}
	hasher := sha512.New()
	hasher.Write(data)

	return fmt.Sprintf("?v=%.8x", hasher.Sum(nil))
}

func webserve() {
	db := opendatabase()
	login.Init(login.InitArgs{Db: db, Logger: ilog, Insecure: develMode, SameSiteStrict: !develMode})

	listener, err := openListener()
	if err != nil {
		elog.Fatal(err)
	}
	runBackendServer()
	initNotifs()
	go enditall()
	go deliveryManager()
	go bgmonitor()
	go refresher()
	debugtiming := false
	getconfig("debugtiming", &debugtiming)
	getconfig("debugqueries", &debugQueries)

	var toload []string
	dents, _ := os.ReadDir(viewDir + "/views")
	for _, dent := range dents {
		name := dent.Name()
		if strings.HasSuffix(name, ".html") {
			toload = append(toload, viewDir+"/views/"+name)
		}
	}
	readviews = templates.Load(develMode, toload...)

	if !develMode {
		assets := []string{
			viewDir + "/views/style.css",
			dataDir + "/views/local.css",
			viewDir + "/views/script.js",
			dataDir + "/views/local.js",
		}
		for _, s := range assets {
			savedassetparams[s] = getassetparam(s)
		}
	}

	err = setFileLimits()
	if err != nil {
		elog.Fatal(err)
	}
	securitizeweb()

	mux := mux.NewRouter()
	if debugtiming {
		mux.Use(requestlogger)
	}
	mux.Use(addcspheaders)
	mux.Use(login.Checker)

	posters := mux.Methods("POST").Subrouter()
	getters := mux.Methods("GET").Subrouter()

	getters.HandleFunc("/", showhome)
	getters.HandleFunc("/front", showfront)
	getters.HandleFunc("/groups", showgroups)
	getters.HandleFunc("/g/{name:[[:alnum:]@._-]+}", showonegroup)
	getters.HandleFunc("/g/{name:[[:alnum:]@._-]+}/chat", showgroupchat)
	getters.HandleFunc("/g/{name:[[:alnum:]@._-]+}/chat/{xid:[[:alnum:]_-]+}", showonechat)
	getters.HandleFunc("/g/{name:[[:alnum:]@._-]+}/p/{xid:[[:alnum:]_-]+}", showonepost)
	getters.HandleFunc("/p/{xid:[[:alnum:]_-]+}", showonepost)
	getters.HandleFunc("/u/{name:[[:alnum:]@._-]+}", showoneperson)
	getters.HandleFunc("/c/{xid:[[:alnum:]_-]+}", showonecomment)
	getters.HandleFunc("/file/{xid:[[:alnum:]._-]+}", servefile)

	getters.HandleFunc("/about", servehtml)
	getters.HandleFunc("/help/{name:[\\pL[:digit:]_.-]+}", servehelp)

	getters.HandleFunc("/robots.txt", showrobots)
	getters.HandleFunc("/.well-known/webfinger", webfinger)

	posters.HandleFunc("/inbox", webinbox)
	posters.HandleFunc("/g/{xid:[[:alnum:]_-]+}/inbox", webinbox)
	posters.HandleFunc("/u/{name:[[:alnum:]_-]+}/inbox", webinbox)
	getters.HandleFunc("/g/{xid:[[:alnum:]_-]+}/outbox", weboutbox)
	getters.HandleFunc("/g/{xid:[[:alnum:]_-]+}/featured", webfeatured)
	getters.HandleFunc("/u/{name:[[:alnum:]_-]+}/outbox", weboutbox)
	getters.HandleFunc("/g/{name:[[:alnum:]@._-]+}/p/{xid:[[:alnum:]_-]+}/replies", webpostreplies)
	getters.HandleFunc("/p/{xid:[[:alnum:]_-]+}/replies", webpostreplies)

	posters.HandleFunc("/dosignup", websignup)

	posters.HandleFunc("/csp-violation", fiveoh)

	getters.HandleFunc("/style.css", serveviewasset)
	getters.HandleFunc("/script.js", serveviewasset)
	getters.HandleFunc("/local.css", servedataasset)
	getters.HandleFunc("/local.js", servedataasset)
	getters.HandleFunc("/icon.png", servedataasset)
	getters.HandleFunc("/favicon.ico", servedataasset)

	getters.HandleFunc("/login", servehtml)
	posters.HandleFunc("/dologin", login.LoginFunc)
	getters.HandleFunc("/dologout", login.LogoutFunc)

	loggedin := mux.NewRoute().Subrouter()
	loggedin.Use(login.Required)
	loggedin.HandleFunc("/notifs", shownotifs)
	loggedin.HandleFunc("/account", showaccount)
	loggedin.HandleFunc("/account/filters", showuserfilters)
	loggedin.Handle("/saveaccount", login.CSRFWrapFunc("saveaccount", websaveaccount))
	loggedin.HandleFunc("/chpass", dochpass)
	loggedin.HandleFunc("/share", webshare)
	loggedin.HandleFunc("/subscribe", websubscribe)
	loggedin.HandleFunc("/unsubscribe", webunsubscribe)
	loggedin.HandleFunc("/search", websearch)
	loggedin.HandleFunc("/report", webreport)
	loggedin.HandleFunc("/savereport", websavereport)

	loggedin.HandleFunc("/newgroup", shownewgroup)
	loggedin.HandleFunc("/g/{name:[[:alnum:]@._-]+}/edit", editonegroup)
	loggedin.HandleFunc("/g/{name:[[:alnum:]@._-]+}/p/{xid:[[:alnum:]_-]+}/edit", editonepost)
	loggedin.HandleFunc("/g/{name:[[:alnum:]@._-]+}/p/{xid:[[:alnum:]_-]+}/approve", approveonepost)
	loggedin.HandleFunc("/p/{xid:[[:alnum:]_-]+}/edit", editonepost)
	loggedin.HandleFunc("/c/{xid:[[:alnum:]_-]+}/edit", editonecomment)
	loggedin.HandleFunc("/g/{name:[[:alnum:]@._-]+}/newpost", shownewpost)
	loggedin.HandleFunc("/g/{name:[[:alnum:]@._-]+}/reports", showgroupreports)

	approvedonly := loggedin.NewRoute().Subrouter()
	approvedonly.Use(approvedcheck)
	approvedonly.Handle("/savegroup", login.CSRFWrapFunc("savepost", websavegroup))
	approvedonly.Handle("/savepost", login.CSRFWrapFunc("savepost", websavepost))
	approvedonly.Handle("/savecomment", login.CSRFWrapFunc("savepost", websavecomment))
	approvedonly.Handle("/savechat", login.CSRFWrapFunc("savepost", websavechat))
	approvedonly.Handle("/saveperson", login.CSRFWrapFunc("savepost", websaveperson))
	approvedonly.HandleFunc("/reply", webreply)

	adminonly := loggedin.NewRoute().Subrouter()
	adminonly.Use(admincheck)
	adminonly.HandleFunc("/admin", showadmin)
	adminonly.HandleFunc("/admin/groups", showadmingroups)
	adminonly.HandleFunc("/admin/users", showadminusers)
	adminonly.HandleFunc("/admin/peers", showadminpeers)
	adminonly.HandleFunc("/admin/filters", showadminfilters)
	adminonly.HandleFunc("/admin/reports", showadminreports)
	adminonly.Handle("/saveadmin", login.CSRFWrapFunc("saveadmin", saveadmin))

	err = http.Serve(listener, mux)
	if err != nil && !errors.Is(err, net.ErrClosed) {
		elog.Printf("serve error: %s", err)
	}
	time.Sleep(15 * time.Second)
	elog.Printf("fell off the bottom")

}
